//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransCuda/cuda.h"

int
main(int argc,
     char **argv)
{
Platform_init(argc,argv);
Cuda_init();
int i,j,k,deviceCount;
const CudaDevice *devices=Cuda_devices(&deviceCount);
for(i=0;i<deviceCount;++i)
  {
  const CudaDevice *d=devices+i;
  Platform_msg("CUDA device %d:\n",CudaDevice_id(d));
  Platform_msg("  name: %s\n",CudaDevice_name(d));
  Platform_msg("  compute capability: %d.%d\n",
               CudaDevice_computeCapabilityMajor(d),
               CudaDevice_computeCapabilityMinor(d));
  Platform_msg("  total memory: %"FPTR"d\n",
               CudaDevice_totalMemory(d));
  Platform_msg("  max threads per block: %d\n",
               CudaDevice_maxThreadsPerBlock(d));
  Platform_msg("  max block dim: %d %d %d\n",
               CudaDevice_maxBlockDimX(d),
               CudaDevice_maxBlockDimY(d),
               CudaDevice_maxBlockDimZ(d));
  Platform_msg("  max grid dim: %d %d %d\n",
               CudaDevice_maxGridDimX(d),
               CudaDevice_maxGridDimY(d),
               CudaDevice_maxGridDimZ(d));
  Platform_msg("  max shared memory per block: %d\n",
               CudaDevice_maxSharedMemoryPerBlock(d));
  Platform_msg("  total constant memory: %d\n",
               CudaDevice_totalConstantMemory(d));
  Platform_msg("  warp size: %d\n",
               CudaDevice_warpSize(d));
  Platform_msg("  max registers per block: %d\n",
               CudaDevice_maxRegistersPerBlock(d));
  Platform_msg("  clock rate: %d kHz\n",
               CudaDevice_clockRate(d));
  Platform_msg("  gpu overlap: %d\n",
               CudaDevice_gpuOverlap(d));
  Platform_msg("  multiprocessor count: %d\n",
               CudaDevice_multiprocessorCount(d));
  Platform_msg("  core count: %d\n",
               CudaDevice_coreCount(d));
  Platform_msg("  kernel exec timeout: %d\n",
               CudaDevice_kernelExecTimeout(d));
  Platform_msg("  integrated: %d\n",
               CudaDevice_integrated(d));
  Platform_msg("  can map host memory: %d\n",
               CudaDevice_canMapHostMemory(d));
  Platform_msg("  concurrent kernels: %d\n",
               CudaDevice_concurrentKernels(d));
  Platform_msg("  ecc enabled: %d\n",
               CudaDevice_eccEnabled(d));
  Platform_msg("  pci bus id: %d\n",
               CudaDevice_pciBusId(d));
  Platform_msg("  pci device id: %d\n",
               CudaDevice_pciDeviceId(d));
  Platform_msg("  tcc driver: %d\n",
               CudaDevice_tccDriver(d));
  Platform_msg("  memory clock rate: %d kHz\n",
               CudaDevice_memoryClockRate(d));
  Platform_msg("  global memory bus width: %d\n",
               CudaDevice_globalMemoryBusWidth(d));
  Platform_msg("  l2 cache size: %d\n",
               CudaDevice_l2CacheSize(d));
  Platform_msg("  max threads per multiprocessor: %d\n",
               CudaDevice_maxThreadsPerMultiprocessor(d));
  Platform_msg("  unified addressing: %d\n",
               CudaDevice_unifiedAddressing(d));
  intptr_t bufferSize=-1;
  if(Platform_argc()>1)
    {
    sscanf(Platform_argv(1),"%"FPTR"d",&bufferSize);
    }
  if(bufferSize<0)
    {
    int extra=0x100;
    bufferSize=CudaDevice_maxGridDimX(d)*
               CudaDevice_maxBlockDimX(d)*sizeof(int);
    if(3*(bufferSize+extra)>CudaDevice_freeMemory(d))
      {
      bufferSize=(CudaDevice_freeMemory(d)/3)-extra;
      }
    }
  else
    {
    bufferSize=bufferSize*1024*1024;
    }
  if(bufferSize)
    {
    Platform_msg("  program:\n");
    const char * sourceCode=
      "\n"
      "extern \"C\" __global__\n"
      "void\n"
      "vectorAdd(int * __restrict__ out,\n"
      "          const int * __restrict__ in1,\n"
      "          const int * __restrict__ in2,\n"
      "          unsigned int arraySize)\n"
      "{\n"
      "unsigned int id=blockDim.x*blockIdx.x+threadIdx.x;\n"
      "if(id<arraySize)\n"
      "  {\n"
      "  out[id]=in1[id]+in2[id];\n"
      "  }\n"
      "}\n"
      "\n";
#if 0
    Platform_msg("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
                 "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    Platform_msg("%s",sourceCode);
    Platform_msg("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
                 "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
#endif
    CudaProgram p;
    CudaProgram_init(&p,d,"vectorAdd","--use_fast_math",sourceCode,true);
    if(CudaProgram_buildLog(&p))
      {
      Platform_msg("    build log: %s\n",
                   CudaProgram_buildLog(&p));
      }
    if(!CudaProgram_buildFailure(&p))
      {
      Platform_msg("    max threads per block: %d\n",
                   CudaProgram_maxThreadsPerBlock(&p));
      Platform_msg("    shared size bytes: %d\n",
                   CudaProgram_sharedSizeBytes(&p));
      Platform_msg("    const size bytes: %d\n",
                   CudaProgram_constSizeBytes(&p));
      Platform_msg("    local size bytes: %d\n",
                   CudaProgram_localSizeBytes(&p));
      Platform_msg("    num regs: %d\n",
                   CudaProgram_numRegs(&p));
      Platform_msg("    ptx version: %d\n",
                   CudaProgram_ptxVersion(&p));
      Platform_msg("    binary version: %d\n",
                   CudaProgram_binaryVersion(&p));
      Platform_msg("  execution:\n");
      Platform_msg("    buffer size: %"FPTR"d (%"FPTR"d MB)\n",
                   bufferSize,bufferSize/1024/1024);
      intptr_t arraySize=bufferSize/sizeof(int);
      int blockDim=CudaProgram_maxThreadsPerBlock(&p);
      Platform_msg("    block dim: %d\n",blockDim);
      int gridDim=0;
      while(gridDim*blockDim<arraySize)
        {
        ++gridDim;
        }
      Platform_msg("    grid dim: %d\n",gridDim);
      CudaMarker m0,m1,m2,m3;
      CudaMarker_init(&m0,d);
      CudaMarker_init(&m1,d);
      CudaMarker_init(&m2,d);
      CudaMarker_init(&m3,d);
      CudaBuffer outBuf,in1Buf,in2Buf;
      CudaBuffer_init(&outBuf,d,bufferSize);
      CudaBuffer_init(&in1Buf,d,bufferSize);
      CudaBuffer_init(&in2Buf,d,bufferSize);
      CudaLockedMem outMem,in1Mem,in2Mem;
      CudaLockedMem_init(&outMem,false,bufferSize);
      CudaLockedMem_init(&in1Mem,true,bufferSize);
      CudaLockedMem_init(&in2Mem,true,bufferSize);
      int *out=(int *)CudaLockedMem_data(&outMem);
      int *in1=(int *)CudaLockedMem_data(&in1Mem);
      int *in2=(int *)CudaLockedMem_data(&in2Mem);
      const void *args[]={ CudaBuffer_programArg(&outBuf),
                           CudaBuffer_programArg(&in1Buf),
                           CudaBuffer_programArg(&in2Buf),
                           &arraySize };
      int data1[17]={37,50,54,50,56,15,51,32,12,20,43,56,99,50,25,15,18};
      int data2[17]={35,51,54,58,55,17,20,48,73,12,44,55,15,58,75,18,14};
      for(j=0;j<5;++j)
        {
        for(k=0;k<arraySize;++k)
          {
          in1[k]=data1[(k+j)%17];
          in2[k]=data2[(k+j)%17];
          }
        CudaMarker_set(&m0);
        CudaBuffer_sendAll(&in1Buf,in1);
        CudaBuffer_sendAll(&in2Buf,in2);
        CudaMarker_set(&m1);
        for(int iter=0;iter<20;++iter)
          CudaProgram_launch(&p,gridDim,1,1,blockDim,1,1,0,args);
        CudaMarker_set(&m2);
        CudaBuffer_receiveAll(&outBuf,out);
        CudaMarker_set(&m3);
        CudaDevice_sync(d);
        Platform_msg("    ");
        for(k=0;k<20;++k) Platform_msg("%c",(char)out[k]);
        Platform_msg(" ... ");
        for(k=0;k<20;++k) Platform_msg("%c",(char)out[arraySize-20+k]);
        Platform_msg(" (%"F64"d+%"F64"d+%"F64"d us)\n",
                     CudaMarker_duration(&m1,&m0),
                     CudaMarker_duration(&m2,&m1),
                     CudaMarker_duration(&m3,&m2));
        }
      CudaLockedMem_destroy(&outMem);
      CudaLockedMem_destroy(&in1Mem);
      CudaLockedMem_destroy(&in2Mem);
      CudaBuffer_destroy(&outBuf);
      CudaBuffer_destroy(&in1Buf);
      CudaBuffer_destroy(&in2Buf);
      CudaMarker_destroy(&m0);
      CudaMarker_destroy(&m1);
      CudaMarker_destroy(&m2);
      CudaMarker_destroy(&m3);
      }
    CudaProgram_destroy(&p);
    }
  }
return 0;
}

DECLARE_WINDOWS_MAIN_FUNCTION

//----------------------------------------------------------------------------
