//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSPROG_RANDOM_H
#define TRANSPROG_RANDOM_H 1

#include "TransProg/math.h"

//  Adapted from MWC generator described in George Marsaglia's post
//  ``Random numbers for C: End, at last?'' on sci.stat.math,
//  sci.math, sci.math.num-analysis, sci.crypt, sci.physics.research,
//  comp.os.msdos.djgpp (Thu, 21 Jan 1999 03:08:52 GMT)

struct _RandomSeed
{
CACHE_ALIGNED
uint32_t s[2];
};

TRANSPROG_INLINE
uint32_t
Random_nextSeed(int cpuCtx)
{
extern TRANSPROG_API_VAR struct _RandomSeed _Random_seeds[];
struct _RandomSeed *seed=_Random_seeds+cpuCtx;
seed->s[0]=36969*(seed->s[0]&0x0000FFFFU)+(seed->s[0]>>16);
seed->s[1]=18000*(seed->s[1]&0x0000FFFFU)+(seed->s[1]>>16);
return (seed->s[0]<<16)+seed->s[1];
}

TRANSPROG_INLINE
uint32_t // random value in [0;limit[
Random_uint32(int cpuCtx,
              uint32_t limit)
{
return Random_nextSeed(cpuCtx)%limit;
}

TRANSPROG_INLINE
int32_t // random value in [0;limit[
Random_int32(int cpuCtx,
             int32_t limit)
{
return (int32_t)Random_uint32(cpuCtx,(uint32_t)limit);
}

TRANSPROG_INLINE
uint64_t // random value in [0;limit[
Random_uint64(int cpuCtx,
              uint64_t limit)
{
uint64_t high=Random_nextSeed(cpuCtx);
uint64_t low=Random_nextSeed(cpuCtx);
return ((high<<32)|low)%limit;
}

TRANSPROG_INLINE
int64_t // random value in [0;limit[
Random_int64(int cpuCtx,
             int64_t limit)
{
return (int64_t)Random_uint64(cpuCtx,(uint64_t)limit);
}

#if INT_MAX==INT32_MAX
# define Random_int Random_int32
#else
# error "weird size for int!"
#endif

#if INTPTR_MAX==INT32_MAX
# define Random_intptr Random_int32
#elif INTPTR_MAX==INT64_MAX
# define Random_intptr Random_int64
#else
# error "weird size for intptr_t!"
#endif

TRANSPROG_INLINE
real32_t // random value in [0.0f;1.0f[
Random_real32(int cpuCtx)
{
// nb: a 32-bit pattern will overflow a 23-bit significand for DP-IEEE754
const real32_t div=1.0f/8388608.0f;
return (Random_nextSeed(cpuCtx)&0x007FFFFFU)*div;
}

TRANSPROG_INLINE
real64_t // random value in [0.0;1.0[
Random_real64(int cpuCtx)
{
// nb: a 32-bit pattern won't overflow a 52-bit significand for DP-IEEE754
const real64_t div=1.0/4294967296.0;
return Random_nextSeed(cpuCtx)*div;
}

TRANSPROG_INLINE
real32_t // random value with mean 0.0f and standard deviation 1.0f
Random_normal32(int cpuCtx)
{
return cosf(M_2PIf*Random_real32(cpuCtx))*
       sqrtf(-2.0f*logf(1.0f-Random_real32(cpuCtx))); // Box-Muller method
}

TRANSPROG_INLINE
real64_t // random value with mean 0.0 and standard deviation 1.0
Random_normal64(int cpuCtx)
{
return cos(M_2PI*Random_real64(cpuCtx))*
       sqrt(-2.0*log(1.0-Random_real64(cpuCtx))); // Box-Muller method
}

#endif // TRANSPROG_RANDOM_H

//----------------------------------------------------------------------------
