//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSPROG_SCHEDULER_H
#define TRANSPROG_SCHEDULER_H 1

#include "TransProg/ptrVector.h"
#include "TransProg/synchro.h"

typedef struct _Scheduler Scheduler;
typedef struct _Activity Activity;

typedef enum { ACT_DEAD,
               ACT_RETRY,
               ACT_SKIP,
               ACT_DONE } ActivityResult;

typedef ActivityResult (*ActivityFunction)(Activity *,Scheduler *);

struct _Activity
{
ActivityFunction fnct;
NarrowCpuCtx cpuCtx; // minimize space
uint8_t retryCount;  // minimize space
uint8_t skipCount;   // minimize space
};

struct _SchedulerActivities
{
CACHE_ALIGNED
SpinSynchro sync;
PtrVector act;
};

struct _SchedulerContext
{
CACHE_ALIGNED
int bestCtx;
int bestAffinity;
int affinity[ASSUMED_MAX_CPU_COUNT];
intptr_t numaUsage[ASSUMED_MAX_NUMA_COUNT];
struct _SchedulerActivities activities[2];
};

struct _Scheduler
{
volatile int64_t step;
bool workStealing;
intptr_t prefetch;
struct _SchedulerContext ctx[ASSUMED_MAX_CPU_COUNT];
};

//----------------------------------------------------------------------------

TRANSPROG_API
void
Activity_init(Activity *a,
              ActivityFunction fnct,
              Scheduler *s,
              int cpuCtx);

TRANSPROG_API
void
Activity_initCopy(Activity *a,
                  Scheduler *s,
                  const Activity *ref);

TRANSPROG_API
void
Activity_destroy(Activity *a,
                 Scheduler *s);

TRANSPROG_INLINE
ActivityFunction
Activity_function(const Activity *a)
{
return a->fnct;
}

TRANSPROG_INLINE
int
Activity_cpuCtx(const Activity *a)
{
return a->cpuCtx;
}

TRANSPROG_INLINE
int
Activity_retryCount(const Activity *a)
{
return a->retryCount;
}

TRANSPROG_INLINE
int
Activity_skipCount(const Activity *a)
{
return a->skipCount;
}

TRANSPROG_INLINE
void
Activity_adjustAffinity(Activity *a,
                        Scheduler *s,
                        int otherMemCtx,
                        int weight)
{
ASSERT_CURRENT_CPUCTX(Activity_cpuCtx(a));
struct _SchedulerContext *c=s->ctx+Activity_cpuCtx(a);
int affinity=(c->affinity[otherMemCtx]+=weight);
if(affinity>c->bestAffinity)
  {
  c->bestAffinity=affinity;
  c->bestCtx=otherMemCtx;
  }
}

//----------------------------------------------------------------------------

TRANSPROG_API
void
Scheduler_init(Scheduler *s);

TRANSPROG_INLINE
Scheduler *
Scheduler_new(void)
{
Scheduler *s=Memory_alloc(1,Scheduler);
Scheduler_init(s);
return s;
}

TRANSPROG_API
void
Scheduler_destroy(Scheduler *s);

TRANSPROG_INLINE
void
Scheduler_delete(Scheduler *s)
{
Scheduler_destroy(s);
Memory_free(s);
}

TRANSPROG_INLINE
bool
Scheduler_workStealing(const Scheduler *s)
{
return s->workStealing;
}

TRANSPROG_INLINE
void
Scheduler_setWorkStealing(Scheduler *s,
                          bool workStealing)
{
s->workStealing=workStealing;
}

TRANSPROG_INLINE
intptr_t
Scheduler_prefetch(const Scheduler *s)
{
return s->prefetch;
}

TRANSPROG_INLINE
void
Scheduler_setPrefetch(Scheduler *s,
                      intptr_t prefetch)
{
s->prefetch=prefetch;
}

TRANSPROG_INLINE
int64_t
Scheduler_step(const Scheduler *s)
{
return s->step;
}

TRANSPROG_INLINE
void
Scheduler_advanceStep(Scheduler *s)
{
++s->step;
}

TRANSPROG_INLINE
int
Scheduler_fromState(const Scheduler *s)
{
return Scheduler_step(s)&1;
}

TRANSPROG_INLINE
int
Scheduler_toState(const Scheduler *s)
{
return 1-Scheduler_fromState(s);
}

TRANSPROG_API
void
Scheduler_activate(int cpuCtx,
                   void *scheduler);

TRANSPROG_INLINE
intptr_t * // 0 initialized storage for numa usage
Scheduler_initNumaUsage(Scheduler *s,
                        int cpuCtx)
{
intptr_t *numaUsage=s->ctx[cpuCtx].numaUsage;
memset(numaUsage,0,Platform_numaCtxCount()*sizeof(intptr_t));
return numaUsage;
}

#endif // TRANSPROG_SCHEDULER_H

//----------------------------------------------------------------------------
