//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSPROG_SPATIALUTILS_H
#define TRANSPROG_SPATIALUTILS_H 1

#include "TransProg/math.h"
#include "TransProg/memory.h"

typedef enum { MTX_R00=0,  MTX_R01=4,  MTX_R02=8,    MTX_TX=12,
               MTX_R10=1,  MTX_R11=5,  MTX_R12=9,    MTX_TY=13,
               MTX_R20=2,  MTX_R21=6,  MTX_R22=10,   MTX_TZ=14,
                MTX_Z0=3,   MTX_Z1=7,   MTX_Z2=11,  MTX_ONE=15
             } MatrixIndex; // COLUMN-major ordered (OpenGL compliant)

typedef enum { LOC_MIDDLE=0,
               LOC_XPOS=1<<0,
               LOC_XNEG=1<<1,
               LOC_YPOS=1<<2,
               LOC_YNEG=1<<3,
               LOC_ZPOS=1<<4,
               LOC_ZNEG=1<<5 } VertexLocation;

//----------------------------------------------------------------------------

typedef struct _SpatialBasis
{
real64_t matrix[16];
} SpatialBasis;

TRANSPROG_INLINE
void
SpatialBasis_init(SpatialBasis *sb)
{
extern TRANSPROG_API_VAR real64_t _SpatialUtils_idMtx[16];
memcpy(sb->matrix,_SpatialUtils_idMtx,16*sizeof(real64_t));
}

TRANSPROG_INLINE
SpatialBasis *
SpatialBasis_new(void)
{
SpatialBasis *sb=Memory_alloc(1,SpatialBasis);
SpatialBasis_init(sb);
return sb;
}

TRANSPROG_INLINE
void
SpatialBasis_destroy(SpatialBasis *sb)
{
(void)sb;
}

TRANSPROG_INLINE
void
SpatialBasis_delete(SpatialBasis *sb)
{
SpatialBasis_destroy(sb);
Memory_free(sb);
}

TRANSPROG_INLINE
void
SpatialBasis_toReal32(const SpatialBasis *sb,
                      real32_t *out_matrix)
{
const real64_t *mtx=sb->matrix;
out_matrix[MTX_R00]=(real32_t)mtx[MTX_R00];
out_matrix[MTX_R10]=(real32_t)mtx[MTX_R10];
out_matrix[MTX_R20]=(real32_t)mtx[MTX_R20];
out_matrix[MTX_Z0]=0.0f;
out_matrix[MTX_R01]=(real32_t)mtx[MTX_R01];
out_matrix[MTX_R11]=(real32_t)mtx[MTX_R11];
out_matrix[MTX_R21]=(real32_t)mtx[MTX_R21];
out_matrix[MTX_Z1]=0.0f;
out_matrix[MTX_R02]=(real32_t)mtx[MTX_R02];
out_matrix[MTX_R12]=(real32_t)mtx[MTX_R12];
out_matrix[MTX_R22]=(real32_t)mtx[MTX_R22];
out_matrix[MTX_Z2]=0.0f;
out_matrix[MTX_TX]=(real32_t)mtx[MTX_TX];
out_matrix[MTX_TY]=(real32_t)mtx[MTX_TY];
out_matrix[MTX_TZ]=(real32_t)mtx[MTX_TZ];
out_matrix[MTX_ONE]=1.0f;
}

TRANSPROG_API
bool // success
SpatialBasis_adjustOrthonormality(SpatialBasis *sb);

TRANSPROG_API
void
SpatialBasis_globalToLocalLocation(const SpatialBasis *sb,
                                   SpatialBasis *inout_location);

TRANSPROG_API
void
SpatialBasis_localToGlobalLocation(const SpatialBasis *sb,
                                   SpatialBasis *inout_location);

TRANSPROG_INLINE
void
SpatialBasis_globalToLocalPosition(const SpatialBasis *sb,
                                   real64_t *inout_x,
                                   real64_t *inout_y,
                                   real64_t *inout_z)
{
const real64_t *mtx=sb->matrix;
real64_t x=*inout_x-mtx[MTX_TX];
real64_t y=*inout_y-mtx[MTX_TY];
real64_t z=*inout_z-mtx[MTX_TZ];
*inout_x=mtx[MTX_R00]*x+mtx[MTX_R10]*y+mtx[MTX_R20]*z;
*inout_y=mtx[MTX_R01]*x+mtx[MTX_R11]*y+mtx[MTX_R21]*z;
*inout_z=mtx[MTX_R02]*x+mtx[MTX_R12]*y+mtx[MTX_R22]*z;
}

TRANSPROG_INLINE
void
SpatialBasis_globalToLocalPositions16(const SpatialBasis *sb,
                                      uint16_t *inout_xyz,
                                      int count)
{
const real64_t *mtx=sb->matrix;
count*=3;
for(int i=0;i<count;i+=3)
  {
  real64_t x=halfToReal64(inout_xyz[i+0])-mtx[MTX_TX];
  real64_t y=halfToReal64(inout_xyz[i+1])-mtx[MTX_TY];
  real64_t z=halfToReal64(inout_xyz[i+2])-mtx[MTX_TZ];
  inout_xyz[i+0]=real64ToHalf(mtx[MTX_R00]*x+mtx[MTX_R10]*y+mtx[MTX_R20]*z);
  inout_xyz[i+1]=real64ToHalf(mtx[MTX_R01]*x+mtx[MTX_R11]*y+mtx[MTX_R21]*z);
  inout_xyz[i+2]=real64ToHalf(mtx[MTX_R02]*x+mtx[MTX_R12]*y+mtx[MTX_R22]*z);
  }
}

TRANSPROG_INLINE
void
SpatialBasis_globalToLocalPositions32(const SpatialBasis *sb,
                                      real32_t *inout_xyz,
                                      int count)
{
const real64_t *mtx=sb->matrix;
count*=3;
for(int i=0;i<count;i+=3)
  {
  real64_t x=((real64_t)inout_xyz[i+0])-mtx[MTX_TX];
  real64_t y=((real64_t)inout_xyz[i+1])-mtx[MTX_TY];
  real64_t z=((real64_t)inout_xyz[i+2])-mtx[MTX_TZ];
  inout_xyz[i+0]=(real32_t)(mtx[MTX_R00]*x+mtx[MTX_R10]*y+mtx[MTX_R20]*z);
  inout_xyz[i+1]=(real32_t)(mtx[MTX_R01]*x+mtx[MTX_R11]*y+mtx[MTX_R21]*z);
  inout_xyz[i+2]=(real32_t)(mtx[MTX_R02]*x+mtx[MTX_R12]*y+mtx[MTX_R22]*z);
  }
}

TRANSPROG_INLINE
void
SpatialBasis_globalToLocalPositions64(const SpatialBasis *sb,
                                      real64_t *inout_xyz,
                                      int count)
{
const real64_t *mtx=sb->matrix;
count*=3;
for(int i=0;i<count;i+=3)
  {
  real64_t x=inout_xyz[i+0]-mtx[MTX_TX];
  real64_t y=inout_xyz[i+1]-mtx[MTX_TY];
  real64_t z=inout_xyz[i+2]-mtx[MTX_TZ];
  inout_xyz[i+0]=mtx[MTX_R00]*x+mtx[MTX_R10]*y+mtx[MTX_R20]*z;
  inout_xyz[i+1]=mtx[MTX_R01]*x+mtx[MTX_R11]*y+mtx[MTX_R21]*z;
  inout_xyz[i+2]=mtx[MTX_R02]*x+mtx[MTX_R12]*y+mtx[MTX_R22]*z;
  }
}

TRANSPROG_INLINE
void
SpatialBasis_localToGlobalPosition(const SpatialBasis *sb,
                                   real64_t *inout_x,
                                   real64_t *inout_y,
                                   real64_t *inout_z)
{
const real64_t *mtx=sb->matrix;
real64_t x=*inout_x;
real64_t y=*inout_y;
real64_t z=*inout_z;
*inout_x=mtx[MTX_R00]*x+mtx[MTX_R01]*y+mtx[MTX_R02]*z+mtx[MTX_TX];
*inout_y=mtx[MTX_R10]*x+mtx[MTX_R11]*y+mtx[MTX_R12]*z+mtx[MTX_TY];
*inout_z=mtx[MTX_R20]*x+mtx[MTX_R21]*y+mtx[MTX_R22]*z+mtx[MTX_TZ];
}

TRANSPROG_INLINE
void
SpatialBasis_localToGlobalPositions16(const SpatialBasis *sb,
                                      uint16_t *inout_xyz,
                                      int count)
{
const real64_t *mtx=sb->matrix;
count*=3;
for(int i=0;i<count;i+=3)
  {
  real64_t x=halfToReal64(inout_xyz[i+0]);
  real64_t y=halfToReal64(inout_xyz[i+1]);
  real64_t z=halfToReal64(inout_xyz[i+2]);
  inout_xyz[i+0]=real64ToHalf(mtx[MTX_R00]*x+mtx[MTX_R01]*y+mtx[MTX_R02]*z+
                             mtx[MTX_TX]);
  inout_xyz[i+1]=real64ToHalf(mtx[MTX_R10]*x+mtx[MTX_R11]*y+mtx[MTX_R12]*z+
                             mtx[MTX_TY]);
  inout_xyz[i+2]=real64ToHalf(mtx[MTX_R20]*x+mtx[MTX_R21]*y+mtx[MTX_R22]*z+
                             mtx[MTX_TZ]);
  }
}

TRANSPROG_INLINE
void
SpatialBasis_localToGlobalPositions32(const SpatialBasis *sb,
                                      real32_t *inout_xyz,
                                      int count)
{
const real64_t *mtx=sb->matrix;
count*=3;
for(int i=0;i<count;i+=3)
  {
  real64_t x=(real64_t)inout_xyz[i+0];
  real64_t y=(real64_t)inout_xyz[i+1];
  real64_t z=(real64_t)inout_xyz[i+2];
  inout_xyz[i+0]=(real32_t)(mtx[MTX_R00]*x+mtx[MTX_R01]*y+mtx[MTX_R02]*z+
                           mtx[MTX_TX]);
  inout_xyz[i+1]=(real32_t)(mtx[MTX_R10]*x+mtx[MTX_R11]*y+mtx[MTX_R12]*z+
                           mtx[MTX_TY]);
  inout_xyz[i+2]=(real32_t)(mtx[MTX_R20]*x+mtx[MTX_R21]*y+mtx[MTX_R22]*z+
                           mtx[MTX_TZ]);
  }
}

TRANSPROG_INLINE
void
SpatialBasis_localToGlobalPositions64(const SpatialBasis *sb,
                                      real64_t *inout_xyz,
                                      int count)
{
const real64_t *mtx=sb->matrix;
count*=3;
for(int i=0;i<count;i+=3)
  {
  real64_t x=inout_xyz[i+0];
  real64_t y=inout_xyz[i+1];
  real64_t z=inout_xyz[i+2];
  inout_xyz[i+0]=mtx[MTX_R00]*x+mtx[MTX_R01]*y+mtx[MTX_R02]*z+mtx[MTX_TX];
  inout_xyz[i+1]=mtx[MTX_R10]*x+mtx[MTX_R11]*y+mtx[MTX_R12]*z+mtx[MTX_TY];
  inout_xyz[i+2]=mtx[MTX_R20]*x+mtx[MTX_R21]*y+mtx[MTX_R22]*z+mtx[MTX_TZ];
  }
}

TRANSPROG_INLINE
void
SpatialBasis_globalToLocalVector(const SpatialBasis *sb,
                                 real64_t *inout_x,
                                 real64_t *inout_y,
                                 real64_t *inout_z)
{
const real64_t *mtx=sb->matrix;
real64_t x=*inout_x;
real64_t y=*inout_y;
real64_t z=*inout_z;
*inout_x=mtx[MTX_R00]*x+mtx[MTX_R10]*y+mtx[MTX_R20]*z;
*inout_y=mtx[MTX_R01]*x+mtx[MTX_R11]*y+mtx[MTX_R21]*z;
*inout_z=mtx[MTX_R02]*x+mtx[MTX_R12]*y+mtx[MTX_R22]*z;
}

TRANSPROG_INLINE
void
SpatialBasis_globalToLocalVectors16(const SpatialBasis *sb,
                                    uint16_t *inout_xyz,
                                    int count)
{
const real64_t *mtx=sb->matrix;
count*=3;
for(int i=0;i<count;i+=3)
  {
  real64_t x=halfToReal64(inout_xyz[i+0]);
  real64_t y=halfToReal64(inout_xyz[i+1]);
  real64_t z=halfToReal64(inout_xyz[i+2]);
  inout_xyz[i+0]=real64ToHalf(mtx[MTX_R00]*x+mtx[MTX_R10]*y+mtx[MTX_R20]*z);
  inout_xyz[i+1]=real64ToHalf(mtx[MTX_R01]*x+mtx[MTX_R11]*y+mtx[MTX_R21]*z);
  inout_xyz[i+2]=real64ToHalf(mtx[MTX_R02]*x+mtx[MTX_R12]*y+mtx[MTX_R22]*z);
  }
}

TRANSPROG_INLINE
void
SpatialBasis_globalToLocalVectors32(const SpatialBasis *sb,
                                    real32_t *inout_xyz,
                                    int count)
{
const real64_t *mtx=sb->matrix;
count*=3;
for(int i=0;i<count;i+=3)
  {
  real64_t x=(real64_t)inout_xyz[i+0];
  real64_t y=(real64_t)inout_xyz[i+1];
  real64_t z=(real64_t)inout_xyz[i+2];
  inout_xyz[i+0]=(real32_t)(mtx[MTX_R00]*x+mtx[MTX_R10]*y+mtx[MTX_R20]*z);
  inout_xyz[i+1]=(real32_t)(mtx[MTX_R01]*x+mtx[MTX_R11]*y+mtx[MTX_R21]*z);
  inout_xyz[i+2]=(real32_t)(mtx[MTX_R02]*x+mtx[MTX_R12]*y+mtx[MTX_R22]*z);
  }
}

TRANSPROG_INLINE
void
SpatialBasis_globalToLocalVectors64(const SpatialBasis *sb,
                                    real64_t *inout_xyz,
                                    int count)
{
const real64_t *mtx=sb->matrix;
count*=3;
for(int i=0;i<count;i+=3)
  {
  real64_t x=inout_xyz[i+0];
  real64_t y=inout_xyz[i+1];
  real64_t z=inout_xyz[i+2];
  inout_xyz[i+0]=mtx[MTX_R00]*x+mtx[MTX_R10]*y+mtx[MTX_R20]*z;
  inout_xyz[i+1]=mtx[MTX_R01]*x+mtx[MTX_R11]*y+mtx[MTX_R21]*z;
  inout_xyz[i+2]=mtx[MTX_R02]*x+mtx[MTX_R12]*y+mtx[MTX_R22]*z;
  }
}

TRANSPROG_INLINE
void
SpatialBasis_localToGlobalVector(const SpatialBasis *sb,
                                 real64_t *inout_x,
                                 real64_t *inout_y,
                                 real64_t *inout_z)
{
const real64_t *mtx=sb->matrix;
real64_t x=*inout_x;
real64_t y=*inout_y;
real64_t z=*inout_z;
*inout_x=mtx[MTX_R00]*x+mtx[MTX_R01]*y+mtx[MTX_R02]*z;
*inout_y=mtx[MTX_R10]*x+mtx[MTX_R11]*y+mtx[MTX_R12]*z;
*inout_z=mtx[MTX_R20]*x+mtx[MTX_R21]*y+mtx[MTX_R22]*z;
}

TRANSPROG_INLINE
void
SpatialBasis_localToGlobalVectors16(const SpatialBasis *sb,
                                    uint16_t *inout_xyz,
                                    int count)
{
const real64_t *mtx=sb->matrix;
count*=3;
for(int i=0;i<count;i+=3)
  {
  real64_t x=halfToReal64(inout_xyz[i+0]);
  real64_t y=halfToReal64(inout_xyz[i+1]);
  real64_t z=halfToReal64(inout_xyz[i+2]);
  inout_xyz[i+0]=real64ToHalf(mtx[MTX_R00]*x+mtx[MTX_R01]*y+mtx[MTX_R02]*z);
  inout_xyz[i+1]=real64ToHalf(mtx[MTX_R10]*x+mtx[MTX_R11]*y+mtx[MTX_R12]*z);
  inout_xyz[i+2]=real64ToHalf(mtx[MTX_R20]*x+mtx[MTX_R21]*y+mtx[MTX_R22]*z);
  }
}

TRANSPROG_INLINE
void
SpatialBasis_localToGlobalVectors32(const SpatialBasis *sb,
                                    real32_t *inout_xyz,
                                    int count)
{
const real64_t *mtx=sb->matrix;
count*=3;
for(int i=0;i<count;i+=3)
  {
  real64_t x=(real64_t)inout_xyz[i+0];
  real64_t y=(real64_t)inout_xyz[i+1];
  real64_t z=(real64_t)inout_xyz[i+2];
  inout_xyz[i+0]=(real32_t)(mtx[MTX_R00]*x+mtx[MTX_R01]*y+mtx[MTX_R02]*z);
  inout_xyz[i+1]=(real32_t)(mtx[MTX_R10]*x+mtx[MTX_R11]*y+mtx[MTX_R12]*z);
  inout_xyz[i+2]=(real32_t)(mtx[MTX_R20]*x+mtx[MTX_R21]*y+mtx[MTX_R22]*z);
  }
}

TRANSPROG_INLINE
void
SpatialBasis_localToGlobalVectors64(const SpatialBasis *sb,
                                    real64_t *inout_xyz,
                                    int count)
{
const real64_t *mtx=sb->matrix;
count*=3;
for(int i=0;i<count;i+=3)
  {
  real64_t x=inout_xyz[i+0];
  real64_t y=inout_xyz[i+1];
  real64_t z=inout_xyz[i+2];
  inout_xyz[i+0]=mtx[MTX_R00]*x+mtx[MTX_R01]*y+mtx[MTX_R02]*z;
  inout_xyz[i+1]=mtx[MTX_R10]*x+mtx[MTX_R11]*y+mtx[MTX_R12]*z;
  inout_xyz[i+2]=mtx[MTX_R20]*x+mtx[MTX_R21]*y+mtx[MTX_R22]*z;
  }
}

TRANSPROG_API
void
SpatialBasis_globalToLocalOrientation(const SpatialBasis *sb,
                                      real64_t *inout_roll,
                                      real64_t *inout_pitch,
                                      real64_t *inout_yaw);

TRANSPROG_API
void
SpatialBasis_localToGlobalOrientation(const SpatialBasis *sb,
                                      real64_t *inout_roll,
                                      real64_t *inout_pitch,
                                      real64_t *inout_yaw);

TRANSPROG_INLINE
void
SpatialBasis_getLocalX(const SpatialBasis *sb,
                       real64_t *out_x,
                       real64_t *out_y,
                       real64_t *out_z)
{
const real64_t *mtx=sb->matrix;
*out_x=mtx[MTX_R00];
*out_y=mtx[MTX_R10];
*out_z=mtx[MTX_R20];
}

TRANSPROG_INLINE
void
SpatialBasis_getLocalY(const SpatialBasis *sb,
                       real64_t *out_x,
                       real64_t *out_y,
                       real64_t *out_z)
{
const real64_t *mtx=sb->matrix;
*out_x=mtx[MTX_R01];
*out_y=mtx[MTX_R11];
*out_z=mtx[MTX_R21];
}

TRANSPROG_INLINE
void
SpatialBasis_getLocalZ(const SpatialBasis *sb,
                       real64_t *out_x,
                       real64_t *out_y,
                       real64_t *out_z)
{
const real64_t *mtx=sb->matrix;
*out_x=mtx[MTX_R02];
*out_y=mtx[MTX_R12];
*out_z=mtx[MTX_R22];
}

TRANSPROG_INLINE
void
SpatialBasis_setMatrix(SpatialBasis *sb,
                       const real64_t *matrix)
{
memcpy(sb->matrix,matrix,16*sizeof(real64_t));
}

TRANSPROG_API
void
SpatialBasis_move(SpatialBasis *sb,
                  const SpatialBasis *motion);

TRANSPROG_API
void
SpatialBasis_invert(SpatialBasis *sb);

TRANSPROG_INLINE
void
SpatialBasis_getPosition(const SpatialBasis *sb,
                         real64_t *pout_x,
                         real64_t *pout_y,
                         real64_t *pout_z)
{
const real64_t *mtx=sb->matrix;
*pout_x=mtx[MTX_TX];
*pout_y=mtx[MTX_TY];
*pout_z=mtx[MTX_TZ];
}

TRANSPROG_INLINE
void
SpatialBasis_setPosition(SpatialBasis *sb,
                         real64_t px,
                         real64_t py,
                         real64_t pz)
{
real64_t *mtx=sb->matrix;
mtx[MTX_TX]=px;
mtx[MTX_TY]=py;
mtx[MTX_TZ]=pz;
}

TRANSPROG_INLINE
void
SpatialBasis_translate(SpatialBasis *sb,
                       real64_t dx,
                       real64_t dy,
                       real64_t dz)
{
real64_t *mtx=sb->matrix;
mtx[MTX_TX]+=dx*mtx[MTX_R00]+dy*mtx[MTX_R01]+dz*mtx[MTX_R02];
mtx[MTX_TY]+=dx*mtx[MTX_R10]+dy*mtx[MTX_R11]+dz*mtx[MTX_R12];
mtx[MTX_TZ]+=dx*mtx[MTX_R20]+dy*mtx[MTX_R21]+dz*mtx[MTX_R22];
}

TRANSPROG_API
void
SpatialBasis_extractZyxOrientation(const SpatialBasis *sb,
                                   real64_t *out_roll,
                                   real64_t *out_pitch,
                                   real64_t *out_yaw);

TRANSPROG_API
void
SpatialBasis_extractXyzOrientation(const SpatialBasis *sb,
                                   real64_t *out_roll,
                                   real64_t *out_pitch,
                                   real64_t *out_yaw);

TRANSPROG_API
void
SpatialBasis_extractRotation(const SpatialBasis *sb,
                             real64_t *out_x,
                             real64_t *out_y,
                             real64_t *out_z,
                             real64_t *out_angle);

TRANSPROG_INLINE
void
SpatialBasis_cancelRotation(SpatialBasis *sb)
{
real64_t *mtx=sb->matrix;
mtx[MTX_R00]=mtx[MTX_R11]=mtx[MTX_R22]=1.0;
mtx[MTX_R01]=mtx[MTX_R02]=
mtx[MTX_R10]=mtx[MTX_R12]=
mtx[MTX_R20]=mtx[MTX_R21]=0.0;
}

TRANSPROG_API
void
SpatialBasis_setZyxOrientation(SpatialBasis *sb,
                               real64_t roll,
                               real64_t pitch,
                               real64_t yaw);

TRANSPROG_API
void
SpatialBasis_setXyzOrientation(SpatialBasis *sb,
                               real64_t roll,
                               real64_t pitch,
                               real64_t yaw);

TRANSPROG_API
void
SpatialBasis_setRotation(SpatialBasis *sb,
                         real64_t x,
                         real64_t y,
                         real64_t z,
                         real64_t angle);

TRANSPROG_API
void
SpatialBasis_rotate(SpatialBasis *sb,
                    real64_t x,
                    real64_t y,
                    real64_t z,
                    real64_t dAngle);

TRANSPROG_API
void
SpatialBasis_roll(SpatialBasis *sb,
                  real64_t dRoll);

TRANSPROG_API
void
SpatialBasis_pitch(SpatialBasis *sb,
                   real64_t dPitch);

TRANSPROG_API
void
SpatialBasis_yaw(SpatialBasis *sb,
                 real64_t dYaw);

TRANSPROG_INLINE
bool // point is visible
SpatialBasis_pointVisibility(const SpatialBasis *sb,
                             real64_t range,
                             real64_t yawVision,
                             real64_t pitchVision,
                             real64_t x,
                             real64_t y,
                             real64_t z,
                             real64_t *out_distance,
                             real64_t *out_yawAngle,
                             real64_t *out_pitchAngle)
{
SpatialBasis_globalToLocalPosition(sb,&x,&y,&z);
// FIXME: avoid computing everything if not necessary
cart3DToSpherical64(x,y,z,out_distance,out_yawAngle,out_pitchAngle);
return ((*out_distance<=range)&&
        (fabs(*out_yawAngle)<=0.5*yawVision)&&
        (fabs(*out_pitchAngle)<=0.5*pitchVision)) ? true : false;
}

TRANSPROG_INLINE
bool // location is visible
SpatialBasis_visibility(const SpatialBasis *sb,
                        real64_t range,
                        real64_t yawVision,
                        real64_t pitchVision,
                        const SpatialBasis *location,
                        real64_t *out_distance,
                        real64_t *out_yawAngle,
                        real64_t *out_pitchAngle)
{
return SpatialBasis_pointVisibility(sb,yawVision,pitchVision,range,
                                    location->matrix[MTX_TX],
                                    location->matrix[MTX_TY],
                                    location->matrix[MTX_TZ],
                                    out_distance,out_yawAngle,out_pitchAngle);
}

//----------------------------------------------------------------------------

typedef struct _NearestVisible
{
const SpatialBasis *viewPoint;
real64_t yawVision;
real64_t pitchVision;
real64_t bestDistance;
real64_t bestYaw;
real64_t bestPitch;
void *bestCandidate;
} NearestVisible;

TRANSPROG_INLINE
void
NearestVisible_start(NearestVisible *nv,
                     const SpatialBasis *viewPoint,
                     real64_t range,
                     real64_t yawVision,
                     real64_t pitchVision)
{
nv->viewPoint=viewPoint;
nv->yawVision=yawVision;
nv->pitchVision=pitchVision;
nv->bestDistance=range;
nv->bestYaw=0.0;
nv->bestPitch=0.0;
nv->bestCandidate=(void *)0;
}

TRANSPROG_INLINE
void
NearestVisible_addPoint(NearestVisible *nv,
                        real64_t x,
                        real64_t y,
                        real64_t z,
                        void *candidate)
{
real64_t distance,yaw,pitch;
if(SpatialBasis_pointVisibility(nv->viewPoint,
                                nv->bestDistance,
                                nv->yawVision,
                                nv->pitchVision,
                                x,y,z,&distance,&yaw,&pitch))
  {
  nv->bestDistance=distance;
  nv->bestYaw=yaw;
  nv->bestPitch=pitch;
  nv->bestCandidate=candidate;
  }
}

TRANSPROG_INLINE
void
NearestVisible_add(NearestVisible *nv,
                   const SpatialBasis *location,
                   void *candidate)
{
NearestVisible_addPoint(nv,
                        location->matrix[MTX_TX],
                        location->matrix[MTX_TY],
                        location->matrix[MTX_TZ],
                        candidate);
}

TRANSPROG_INLINE
void * // best candidate
NearestVisible_get(NearestVisible *nv,
                   real64_t *out_distance,
                   real64_t *out_yaw,
                   real64_t *out_pitch)
{
*out_distance=nv->bestDistance;
*out_yaw=nv->bestYaw;
*out_pitch=nv->bestPitch;
return nv->bestCandidate;
}

//----------------------------------------------------------------------------

typedef struct _SpatialTransform
{
real64_t matrix[16];
bool identity;
} SpatialTransform;

TRANSPROG_INLINE
void
SpatialTransform_init(SpatialTransform *st)
{
extern TRANSPROG_API_VAR real64_t _SpatialUtils_idMtx[16];
memcpy(st->matrix,_SpatialUtils_idMtx,16*sizeof(real64_t));
st->identity=true;
}

TRANSPROG_INLINE
SpatialTransform *
SpatialTransform_new(void)
{
SpatialTransform *st=Memory_alloc(1,SpatialTransform);
SpatialTransform_init(st);
return st;
}

TRANSPROG_INLINE
void
SpatialTransform_destroy(SpatialTransform *st)
{
(void)st;
}

TRANSPROG_INLINE
void
SpatialTransform_delete(SpatialTransform *st)
{
SpatialTransform_destroy(st);
Memory_free(st);
}

TRANSPROG_INLINE
void
SpatialTransform_toReal32(const SpatialTransform *st,
                          real32_t *out_matrix)
{
const real64_t *mtx=st->matrix;
out_matrix[MTX_R00]=(real32_t)mtx[MTX_R00];
out_matrix[MTX_R10]=(real32_t)mtx[MTX_R10];
out_matrix[MTX_R20]=(real32_t)mtx[MTX_R20];
out_matrix[MTX_Z0]=0.0f;
out_matrix[MTX_R01]=(real32_t)mtx[MTX_R01];
out_matrix[MTX_R11]=(real32_t)mtx[MTX_R11];
out_matrix[MTX_R21]=(real32_t)mtx[MTX_R21];
out_matrix[MTX_Z1]=0.0f;
out_matrix[MTX_R02]=(real32_t)mtx[MTX_R02];
out_matrix[MTX_R12]=(real32_t)mtx[MTX_R12];
out_matrix[MTX_R22]=(real32_t)mtx[MTX_R22];
out_matrix[MTX_Z2]=0.0f;
out_matrix[MTX_TX]=(real32_t)mtx[MTX_TX];
out_matrix[MTX_TY]=(real32_t)mtx[MTX_TY];
out_matrix[MTX_TZ]=(real32_t)mtx[MTX_TZ];
out_matrix[MTX_ONE]=1.0f;
}

TRANSPROG_INLINE
void
SpatialTransform_setMatrix(SpatialTransform *st,
                           const real64_t *matrix)
{
st->identity=false;
memcpy(st->matrix,matrix,16*sizeof(real64_t));
}

TRANSPROG_API
void
SpatialTransform_invert(SpatialTransform *st);

TRANSPROG_INLINE
void
SpatialTransform_applyToPosition(const SpatialTransform *st,
                                 real64_t *inout_x,
                                 real64_t *inout_y,
                                 real64_t *inout_z)
{
if(!st->identity)
  {
  real64_t x=*inout_x;
  real64_t y=*inout_y;
  real64_t z=*inout_z;
  const real64_t *mtx=st->matrix;
  *inout_x=mtx[MTX_R00]*x+mtx[MTX_R01]*y+mtx[MTX_R02]*z+mtx[MTX_TX];
  *inout_y=mtx[MTX_R10]*x+mtx[MTX_R11]*y+mtx[MTX_R12]*z+mtx[MTX_TY];
  *inout_z=mtx[MTX_R20]*x+mtx[MTX_R21]*y+mtx[MTX_R22]*z+mtx[MTX_TZ];
  }
}

TRANSPROG_INLINE
void
SpatialTransform_applyToPositions16(const SpatialTransform *st,
                                    uint16_t *inout_xyz,
                                    int count)
{
if(!st->identity)
  {
  const real64_t *mtx=st->matrix;
  count*=3;
  for(int i=0;i<count;i+=3)
    {
    real64_t x=halfToReal64(inout_xyz[i+0]);
    real64_t y=halfToReal64(inout_xyz[i+1]);
    real64_t z=halfToReal64(inout_xyz[i+2]);
    inout_xyz[i+0]=real64ToHalf(mtx[MTX_R00]*x+mtx[MTX_R01]*y+mtx[MTX_R02]*z+
                               mtx[MTX_TX]);
    inout_xyz[i+1]=real64ToHalf(mtx[MTX_R10]*x+mtx[MTX_R11]*y+mtx[MTX_R12]*z+
                               mtx[MTX_TY]);
    inout_xyz[i+2]=real64ToHalf(mtx[MTX_R20]*x+mtx[MTX_R21]*y+mtx[MTX_R22]*z+
                               mtx[MTX_TZ]);
    }
  }
}

TRANSPROG_INLINE
void
SpatialTransform_applyToPositions32(const SpatialTransform *st,
                                    real32_t *inout_xyz,
                                    int count)
{
if(!st->identity)
  {
  const real64_t *mtx=st->matrix;
  count*=3;
  for(int i=0;i<count;i+=3)
    {
    real64_t x=(real64_t)inout_xyz[i+0];
    real64_t y=(real64_t)inout_xyz[i+1];
    real64_t z=(real64_t)inout_xyz[i+2];
    inout_xyz[i+0]=(real32_t)(mtx[MTX_R00]*x+mtx[MTX_R01]*y+mtx[MTX_R02]*z+
                             mtx[MTX_TX]);
    inout_xyz[i+1]=(real32_t)(mtx[MTX_R10]*x+mtx[MTX_R11]*y+mtx[MTX_R12]*z+
                             mtx[MTX_TY]);
    inout_xyz[i+2]=(real32_t)(mtx[MTX_R20]*x+mtx[MTX_R21]*y+mtx[MTX_R22]*z+
                             mtx[MTX_TZ]);
    }
  }
}

TRANSPROG_INLINE
void
SpatialTransform_applyToPositions64(const SpatialTransform *st,
                                    real64_t *inout_xyz,
                                    int count)
{
if(!st->identity)
  {
  const real64_t *mtx=st->matrix;
  count*=3;
  for(int i=0;i<count;i+=3)
    {
    real64_t x=inout_xyz[i+0];
    real64_t y=inout_xyz[i+1];
    real64_t z=inout_xyz[i+2];
    inout_xyz[i+0]=mtx[MTX_R00]*x+mtx[MTX_R01]*y+mtx[MTX_R02]*z+mtx[MTX_TX];
    inout_xyz[i+1]=mtx[MTX_R10]*x+mtx[MTX_R11]*y+mtx[MTX_R12]*z+mtx[MTX_TY];
    inout_xyz[i+2]=mtx[MTX_R20]*x+mtx[MTX_R21]*y+mtx[MTX_R22]*z+mtx[MTX_TZ];
    }
  }
}

TRANSPROG_INLINE
void
SpatialTransform_applyToVector(const SpatialTransform *st,
                               real64_t *inout_x,
                               real64_t *inout_y,
                               real64_t *inout_z)
{
if(!st->identity)
  {
  real64_t x=*inout_x;
  real64_t y=*inout_y;
  real64_t z=*inout_z;
  const real64_t *mtx=st->matrix;
  *inout_x=mtx[MTX_R00]*x+mtx[MTX_R01]*y+mtx[MTX_R02]*z;
  *inout_y=mtx[MTX_R10]*x+mtx[MTX_R11]*y+mtx[MTX_R12]*z;
  *inout_z=mtx[MTX_R20]*x+mtx[MTX_R21]*y+mtx[MTX_R22]*z;
  }
}

TRANSPROG_INLINE
void
SpatialTransform_applyToVectors16(const SpatialTransform *st,
                                  uint16_t *inout_xyz,
                                  int count)
{
if(!st->identity)
  {
  const real64_t *mtx=st->matrix;
  count*=3;
  for(int i=0;i<count;i+=3)
    {
    real64_t x=halfToReal64(inout_xyz[i+0]);
    real64_t y=halfToReal64(inout_xyz[i+1]);
    real64_t z=halfToReal64(inout_xyz[i+2]);
    inout_xyz[i+0]=real64ToHalf(mtx[MTX_R00]*x+mtx[MTX_R01]*y+mtx[MTX_R02]*z);
    inout_xyz[i+1]=real64ToHalf(mtx[MTX_R10]*x+mtx[MTX_R11]*y+mtx[MTX_R12]*z);
    inout_xyz[i+2]=real64ToHalf(mtx[MTX_R20]*x+mtx[MTX_R21]*y+mtx[MTX_R22]*z);
    }
  }
}

TRANSPROG_INLINE
void
SpatialTransform_applyToVectors32(const SpatialTransform *st,
                                  real32_t *inout_xyz,
                                  int count)
{
if(!st->identity)
  {
  const real64_t *mtx=st->matrix;
  count*=3;
  for(int i=0;i<count;i+=3)
    {
    real64_t x=(real64_t)inout_xyz[i+0];
    real64_t y=(real64_t)inout_xyz[i+1];
    real64_t z=(real64_t)inout_xyz[i+2];
    inout_xyz[i+0]=(real32_t)(mtx[MTX_R00]*x+mtx[MTX_R01]*y+mtx[MTX_R02]*z);
    inout_xyz[i+1]=(real32_t)(mtx[MTX_R10]*x+mtx[MTX_R11]*y+mtx[MTX_R12]*z);
    inout_xyz[i+2]=(real32_t)(mtx[MTX_R20]*x+mtx[MTX_R21]*y+mtx[MTX_R22]*z);
    }
  }
}

TRANSPROG_INLINE
void
SpatialTransform_applyToVectors64(const SpatialTransform *st,
                                  real64_t *inout_xyz,
                                  int count)
{
if(!st->identity)
  {
  const real64_t *mtx=st->matrix;
  count*=3;
  for(int i=0;i<count;i+=3)
    {
    real64_t x=inout_xyz[i+0];
    real64_t y=inout_xyz[i+1];
    real64_t z=inout_xyz[i+2];
    inout_xyz[i+0]=mtx[MTX_R00]*x+mtx[MTX_R01]*y+mtx[MTX_R02]*z;
    inout_xyz[i+1]=mtx[MTX_R10]*x+mtx[MTX_R11]*y+mtx[MTX_R12]*z;
    inout_xyz[i+2]=mtx[MTX_R20]*x+mtx[MTX_R21]*y+mtx[MTX_R22]*z;
    }
  }
}

TRANSPROG_INLINE
bool // transformation as no effect
SpatialTransform_identity(const SpatialTransform *st,
                          bool checkCoefs)
{
if(checkCoefs)
  {
  extern TRANSPROG_API_VAR real64_t _SpatialUtils_idMtx[16];
  ((SpatialTransform *)st)->identity=
  memcmp(st->matrix,_SpatialUtils_idMtx,16*sizeof(real64_t)) ? false : true;
  }
return st->identity;
}

TRANSPROG_INLINE
void
SpatialTransform_setIdentity(SpatialTransform *st)
{
st->identity=true;
extern TRANSPROG_API_VAR real64_t _SpatialUtils_idMtx[16];
memcpy(st->matrix,_SpatialUtils_idMtx,16*sizeof(real64_t));
}

TRANSPROG_API
void
SpatialTransform_preTransform(SpatialTransform *st,
                              const SpatialTransform *other);

TRANSPROG_API
void
SpatialTransform_postTransform(SpatialTransform *st,
                               const SpatialTransform *other);

TRANSPROG_API
void
SpatialTransform_preMove(SpatialTransform *st,
                         const SpatialBasis *motion);

TRANSPROG_API
void
SpatialTransform_postMove(SpatialTransform *st,
                          const SpatialBasis *motion);

TRANSPROG_INLINE
void
SpatialTransform_preTranslate(SpatialTransform *st,
                              real64_t tx,
                              real64_t ty,
                              real64_t tz)
{
st->identity=false;
real64_t *mtx=st->matrix;
mtx[MTX_TX]+=tx;
mtx[MTX_TY]+=ty;
mtx[MTX_TZ]+=tz;
}

TRANSPROG_INLINE
void
SpatialTransform_postTranslate(SpatialTransform *st,
                               real64_t tx,
                               real64_t ty,
                               real64_t tz)
{
real64_t *mtx=st->matrix;
if(st->identity)
  {
  st->identity=false;
  mtx[MTX_TX]=tx;
  mtx[MTX_TY]=ty;
  mtx[MTX_TZ]=tz;
  }
else
  {
  mtx[MTX_TX]+=tx*mtx[MTX_R00]+ty*mtx[MTX_R01]+tz*mtx[MTX_R02];
  mtx[MTX_TY]+=tx*mtx[MTX_R10]+ty*mtx[MTX_R11]+tz*mtx[MTX_R12];
  mtx[MTX_TZ]+=tx*mtx[MTX_R20]+ty*mtx[MTX_R21]+tz*mtx[MTX_R22];
  }
}

TRANSPROG_API
void
SpatialTransform_preRotate(SpatialTransform *st,
                           real64_t x,
                           real64_t y,
                           real64_t z,
                           real64_t angle);

TRANSPROG_API
void
SpatialTransform_postRotate(SpatialTransform *st,
                            real64_t x,
                            real64_t y,
                            real64_t z,
                            real64_t angle);

TRANSPROG_API
void
SpatialTransform_preRoll(SpatialTransform *st,
                         real64_t roll);

TRANSPROG_API
void
SpatialTransform_postRoll(SpatialTransform *st,
                          real64_t roll);

TRANSPROG_API
void
SpatialTransform_prePitch(SpatialTransform *st,
                          real64_t pitch);

TRANSPROG_API
void
SpatialTransform_postPitch(SpatialTransform *st,
                           real64_t pitch);

TRANSPROG_API
void
SpatialTransform_preYaw(SpatialTransform *st,
                        real64_t yaw);

TRANSPROG_API
void
SpatialTransform_postYaw(SpatialTransform *st,
                         real64_t yaw);

TRANSPROG_INLINE
void
SpatialTransform_preScale(SpatialTransform *st,
                          real64_t sx,
                          real64_t sy,
                          real64_t sz)
{
if((sx>0.0)&&(sy>0.0)&&(sz>0.0))
  {
  st->identity=false;
  real64_t *mtx=st->matrix;
  mtx[MTX_R00]*=sx;
  mtx[MTX_R01]*=sx;
  mtx[MTX_R02]*=sx;
  mtx[MTX_TX]*=sx;
  mtx[MTX_R10]*=sy;
  mtx[MTX_R11]*=sy;
  mtx[MTX_R12]*=sy;
  mtx[MTX_TY]*=sy;
  mtx[MTX_R20]*=sz;
  mtx[MTX_R21]*=sz;
  mtx[MTX_R22]*=sz;
  mtx[MTX_TZ]*=sz;
  }
}

TRANSPROG_INLINE
void
SpatialTransform_postScale(SpatialTransform *st,
                           real64_t sx,
                           real64_t sy,
                           real64_t sz)
{
if((sx>0.0)&&(sy>0.0)&&(sz>0.0))
  {
  st->identity=false;
  real64_t *mtx=st->matrix;
  mtx[MTX_R00]*=sx;
  mtx[MTX_R10]*=sx;
  mtx[MTX_R20]*=sx;
  mtx[MTX_R01]*=sy;
  mtx[MTX_R11]*=sy;
  mtx[MTX_R21]*=sy;
  mtx[MTX_R02]*=sz;
  mtx[MTX_R12]*=sz;
  mtx[MTX_R22]*=sz;
  }
}

//----------------------------------------------------------------------------

typedef struct _BoundingBox
{
real64_t xmin;
real64_t xmax;
real64_t ymin;
real64_t ymax;
real64_t zmin;
real64_t zmax;
} BoundingBox;

TRANSPROG_INLINE
void
BoundingBox_init(BoundingBox *bb)
{
bb->xmin=bb->ymin=bb->zmin=DBL_MAX;
bb->xmax=bb->ymax=bb->zmax=-DBL_MAX;
}

TRANSPROG_INLINE
BoundingBox *
BoundingBox_new(void)
{
BoundingBox *bb=Memory_alloc(1,BoundingBox);
BoundingBox_init(bb);
return bb;
}

TRANSPROG_INLINE
void
BoundingBox_destroy(BoundingBox *bb)
{
(void)bb;
}

TRANSPROG_INLINE
void
BoundingBox_delete(BoundingBox *bb)
{
BoundingBox_destroy(bb);
Memory_free(bb);
}

TRANSPROG_INLINE
bool
BoundingBox_valid(const BoundingBox *bb)
{
return bb->xmin!=DBL_MAX;
}

TRANSPROG_INLINE
void
BoundingBox_invalidate(BoundingBox *bb)
{
bb->xmin=bb->ymin=bb->zmin=DBL_MAX;
bb->xmax=bb->ymax=bb->zmax=-DBL_MAX;
}

TRANSPROG_INLINE
real64_t
BoundingBox_maxX(const BoundingBox *bb)
{
return bb->xmax;
}

TRANSPROG_INLINE
real64_t
BoundingBox_minX(const BoundingBox *bb)
{
return bb->xmin;
}

TRANSPROG_INLINE
real64_t
BoundingBox_maxY(const BoundingBox *bb)
{
return bb->ymax;
}

TRANSPROG_INLINE
real64_t
BoundingBox_minY(const BoundingBox *bb)
{
return bb->ymin;
}

TRANSPROG_INLINE
real64_t
BoundingBox_maxZ(const BoundingBox *bb)
{
return bb->zmax;
}

TRANSPROG_INLINE
real64_t
BoundingBox_minZ(const BoundingBox *bb)
{
return bb->zmin;
}

TRANSPROG_INLINE
real64_t
BoundingBox_sizeX(const BoundingBox *bb)
{
return bb->xmax-bb->xmin;
}

TRANSPROG_INLINE
real64_t
BoundingBox_sizeY(const BoundingBox *bb)
{
return bb->ymax-bb->ymin;
}

TRANSPROG_INLINE
real64_t
BoundingBox_sizeZ(const BoundingBox *bb)
{
return bb->zmax-bb->zmin;
}

TRANSPROG_INLINE
real64_t
BoundingBox_maxSize(const BoundingBox *bb)
{
real64_t sx=BoundingBox_sizeX(bb);
real64_t sy=BoundingBox_sizeY(bb);
real64_t sz=BoundingBox_sizeZ(bb);
return sx>sy ? (sx>sz ? sx : sz) : (sy>sz ? sy : sz);
}

TRANSPROG_INLINE
real64_t
BoundingBox_minSize(const BoundingBox *bb)
{
real64_t sx=BoundingBox_sizeX(bb);
real64_t sy=BoundingBox_sizeY(bb);
real64_t sz=BoundingBox_sizeZ(bb);
return sx<sy ? (sx<sz ? sx : sz) : (sy<sz ? sy : sz);
}

TRANSPROG_INLINE
real64_t
BoundingBox_centerX(const BoundingBox *bb)
{
return 0.5*(bb->xmax+bb->xmin);
}

TRANSPROG_INLINE
real64_t
BoundingBox_centerY(const BoundingBox *bb)
{
return 0.5*(bb->ymax+bb->ymin);
}

TRANSPROG_INLINE
real64_t
BoundingBox_centerZ(const BoundingBox *bb)
{
return 0.5*(bb->zmax+bb->zmin);
}

TRANSPROG_INLINE
void
BoundingBox_addVertex(BoundingBox *bb,
                      real64_t x,
                      real64_t y,
                      real64_t z)
{
// FIXME: shouldn't we test validity to avoid testing for max after
//        finding a min?
if(x<bb->xmin) bb->xmin=x;
if(x>bb->xmax) bb->xmax=x;
if(y<bb->ymin) bb->ymin=y;
if(y>bb->ymax) bb->ymax=y;
if(z<bb->zmin) bb->zmin=z;
if(z>bb->zmax) bb->zmax=z;
}

TRANSPROG_INLINE
void
BoundingBox_addVertices16(BoundingBox *bb,
                          const uint16_t *xyz,
                          int count)
{
count*=3;
for(int i=0;i<count;i+=3)
  {
  BoundingBox_addVertex(bb,halfToReal64(xyz[i+0]),
                           halfToReal64(xyz[i+1]),
                           halfToReal64(xyz[i+2]));
  }
}

TRANSPROG_INLINE
void
BoundingBox_addVertices32(BoundingBox *bb,
                          const real32_t *xyz,
                          int count)
{
count*=3;
for(int i=0;i<count;i+=3)
  {
  BoundingBox_addVertex(bb,(real64_t)xyz[i+0],
                           (real64_t)xyz[i+1],
                           (real64_t)xyz[i+2]);
  }
}

TRANSPROG_INLINE
void
BoundingBox_addVertices64(BoundingBox *bb,
                          const real64_t *xyz,
                          int count)
{
count*=3;
for(int i=0;i<count;i+=3)
  {
  BoundingBox_addVertex(bb,xyz[i+0],xyz[i+1],xyz[i+2]);
  }
}

TRANSPROG_INLINE
void
BoundingBox_addBox(BoundingBox *bb,
                   const BoundingBox *other)
{
if(other->xmin<bb->xmin) bb->xmin=other->xmin;
if(other->xmax>bb->xmax) bb->xmax=other->xmax;
if(other->ymin<bb->ymin) bb->ymin=other->ymin;
if(other->ymax>bb->ymax) bb->ymax=other->ymax;
if(other->zmin<bb->zmin) bb->zmin=other->zmin;
if(other->zmax>bb->zmax) bb->zmax=other->zmax;
}

TRANSPROG_API
void
BoundingBox_addMovedBox(BoundingBox *bb,
                        const SpatialBasis *motion,
                        const BoundingBox *other);

TRANSPROG_API
void
BoundingBox_addTransformedBox(BoundingBox *bb,
                              const SpatialTransform *transform,
                              const BoundingBox *other);

TRANSPROG_INLINE
bool
BoundingBox_containsVertex(const BoundingBox *bb,
                           real64_t x,
                           real64_t y,
                           real64_t z)
{
return (x>=bb->xmin)&&(x<=bb->xmax)&&
       (y>=bb->ymin)&&(y<=bb->ymax)&&
       (z>=bb->zmin)&&(z<=bb->zmax) ? true : false;
}

TRANSPROG_INLINE
VertexLocation
BoundingBox_locateVertex(const BoundingBox *bb,
                         real64_t x,
                         real64_t y,
                         real64_t z)
{
int loc;
if(x>bb->xmax)      loc=LOC_XPOS;
else if(x<bb->xmin) loc=LOC_XNEG;
else                loc=LOC_MIDDLE;
if(y>bb->ymax)      loc|=LOC_YPOS;
else if(y<bb->ymin) loc|=LOC_YNEG;
if(z>bb->zmax)      loc|=LOC_ZPOS;
else if(z<bb->zmin) loc|=LOC_ZNEG;
return (VertexLocation)loc;
}

//----------------------------------------------------------------------------

typedef struct _Frustum
{
real64_t nearDistance;
real64_t farDistance;
real64_t leftSlope;
real64_t rightSlope;
real64_t upSlope;
real64_t downSlope;
} Frustum;

TRANSPROG_INLINE
void
Frustum_init(Frustum *f)
{
f->nearDistance=0.1;
f->farDistance=100.0;
f->leftSlope=1.0;
f->rightSlope=-1.0;
f->upSlope=1.0;
f->downSlope=-1.0;
}

TRANSPROG_INLINE
Frustum *
Frustum_new(void)
{
Frustum *f=Memory_alloc(1,Frustum);
Frustum_init(f);
return f;
}

TRANSPROG_INLINE
void
Frustum_destroy(Frustum *f)
{
(void)f;
}

TRANSPROG_INLINE
void
Frustum_delete(Frustum *f)
{
Frustum_destroy(f);
Memory_free(f);
}

TRANSPROG_INLINE
VertexLocation
Frustum_locateVertex(const Frustum *f,
                     real64_t x,
                     real64_t y,
                     real64_t z)
{
int loc;
if(x>f->farDistance)       loc=LOC_XPOS;
else if(x<f->nearDistance) loc=LOC_XNEG;
else                       loc=LOC_MIDDLE;
if(y>f->leftSlope*x)       loc|=LOC_YPOS;
if(y<f->rightSlope*x)      loc|=LOC_YNEG;
if(z>f->upSlope*x)         loc|=LOC_ZPOS;
if(z<f->downSlope*x)       loc|=LOC_ZNEG;
return (VertexLocation)loc;
}

TRANSPROG_API
bool // bounding box is visible
Frustum_boxVisibility(const Frustum *f,
                      const SpatialBasis *viewPoint,
                      const BoundingBox *box,
                      const SpatialBasis *location,
                      const SpatialTransform *transform);

//----------------------------------------------------------------------------

typedef struct _Perspective
{
Frustum frustum;
real64_t upDownFieldOfView;
real64_t aspectRatio;
real64_t leftRightSlopeOffset;
real64_t upDownSlopeOffset;
real64_t stereoEyeSeparation;
real64_t stereoViewDistance;
} Perspective;

TRANSPROG_API
void
Perspective_init(Perspective *p);

TRANSPROG_INLINE
Perspective *
Perspective_new(void)
{
Perspective *p=Memory_alloc(1,Perspective);
Perspective_init(p);
return p;
}

TRANSPROG_INLINE
void
Perspective_destroy(Perspective *p)
{
(void)p;
}

TRANSPROG_INLINE
void
Perspective_delete(Perspective *p)
{
Perspective_destroy(p);
Memory_free(p);
}

TRANSPROG_INLINE
real64_t
Perspective_upDownFieldOfView(const Perspective *p)
{
return p->upDownFieldOfView;
}

TRANSPROG_API
void
Perspective_setUpDownFieldOfView(Perspective *p,
                                 real64_t upDownFieldOfView);

TRANSPROG_INLINE
real64_t
Perspective_aspectRatio(const Perspective *p)
{
return p->aspectRatio;
}

TRANSPROG_API
void
Perspective_setAspectRatio(Perspective *p,
                           real64_t aspectRatio);

TRANSPROG_INLINE
const Frustum *
Perspective_frustum(const Perspective *p)
{
return &p->frustum;
}

TRANSPROG_INLINE
void
Perspective_setNearDistance(Perspective *p,
                            real64_t nearDistance)
{
if(nearDistance>=p->frustum.farDistance) // avoid crossing planes
  {
  real64_t factor=p->frustum.farDistance/p->frustum.nearDistance;
  p->frustum.farDistance=nearDistance*factor;
  }
p->frustum.nearDistance=nearDistance;
}

TRANSPROG_INLINE
void
Perspective_setFarDistance(Perspective *p,
                           real64_t farDistance)
{
if(farDistance<=p->frustum.nearDistance) // avoid crossing planes
  {
  real64_t factor=p->frustum.nearDistance/p->frustum.farDistance;
  p->frustum.nearDistance=farDistance*factor;
  }
p->frustum.farDistance=farDistance;
}


TRANSPROG_INLINE
real64_t
Perspective_leftRightSlopeOffset(const Perspective *p)
{
return p->leftRightSlopeOffset;
}

TRANSPROG_API
void
Perspective_setLeftRightSlopeOffset(Perspective *p,
                                    real64_t leftRightSlopeOffset);

TRANSPROG_INLINE
real64_t
Perspective_upDownSlopeOffset(const Perspective *p)
{
return p->upDownSlopeOffset;
}

TRANSPROG_API
void
Perspective_setUpDownSlopeOffset(Perspective *p,
                                 real64_t upDownSlopeOffset);

TRANSPROG_INLINE
real64_t
Perspective_stereoEyeSeparation(const Perspective *p)
{
return p->stereoEyeSeparation;
}

TRANSPROG_INLINE
void
Perspective_setStereoEyeSeparation(Perspective *p,
                                   real64_t stereoEyeSeparation)
{
p->stereoEyeSeparation=stereoEyeSeparation;
}

TRANSPROG_INLINE
real64_t
Perspective_stereoViewDistance(const Perspective *p)
{
return p->stereoViewDistance;
}

TRANSPROG_INLINE
void
Perspective_setStereoViewDistance(Perspective *p,
                                  real64_t stereoViewDistance)
{
p->stereoViewDistance=stereoViewDistance;
}

TRANSPROG_INLINE
VertexLocation
Perspective_locateVertex(const Perspective *p,
                         real64_t x,
                         real64_t y,
                         real64_t z)
{
return Frustum_locateVertex(&p->frustum,x,y,z);
}

TRANSPROG_INLINE
bool // bounding box is visible
Perspective_boxVisibility(const Perspective *p,
                          const SpatialBasis *viewPoint,
                          const BoundingBox *box,
                          const SpatialBasis *location,
                          const SpatialTransform *transform)
{
return Frustum_boxVisibility(&p->frustum,viewPoint,box,location,transform);
}

#endif // TRANSPROG_SPATIALUTILS_H

//----------------------------------------------------------------------------
