#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (C) 2008-2015 Fabrice HARROUET (ENIB)
#
# This file is part of TransProg.
# 
# TransProg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# TransProg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

import os
import sys
import traceback
import tempfile
import subprocess
import xml.dom.pulldom

# xml.dom.minidom loads the xml tree at once and thus can lead to memory
# exhaustion in case of large input file.
# Conversely, xml.dom.pulldom allows to consider each node one after
# another and does not require the entire xml tree to stand in memory
# at once.

class Structure:
  def __init__(self,**kw):
    for (k,v) in kw.items():
      setattr(self,k,v)

def pullNodes(xmlDoc,functions):
  for (evt,xmlNode) in xmlDoc:
    if evt==xml.dom.pulldom.START_ELEMENT:
      if xmlNode.localName=='cdecl':
        decl=pullCdecl(xmlDoc)
        if decl.swigKind=='variable' and decl.swigDecl.startswith('p.f('):
          decl.swigKind='function'
          decl.swigDecl=decl.swigDecl[2:]
        if decl.swigKind=='function' and decl.swigName and \
           decl.swigDecl.startswith('f(') and \
           'v(...)' not in [p.swigType for p in decl.swigParmlist]:
          count=0
          index=2
          while True:
            c=decl.swigDecl[index]
            index+=1
            if c=='(':
              count+=1
            elif c==')':
              if count:
                count-=1
              else:
                break
          decl.swigType=decl.swigDecl[index+1:]+decl.swigType
          functions.append(decl)
        # else:
        #   sys.stdout.write('unknown cdecl: %s %s\n'
        #                    % (decl.swigKind,decl.swigName))
      elif xmlNode.localName=='import':
        pullIgnored(xmlDoc)
      else:
        pullNodes(xmlDoc,functions)
    elif evt==xml.dom.pulldom.END_ELEMENT:
      break

def pullCdecl(xmlDoc):
  decl=Structure(swigKind='',
                 swigDecl='',
                 swigType='void',
                 swigName='',
                 swigParmlist=[])
  for (evt,xmlNode) in xmlDoc:
    if evt==xml.dom.pulldom.START_ELEMENT:
      if xmlNode.localName=='attributelist':
        for (evt2,xmlNode2) in xmlDoc:
          if evt2==xml.dom.pulldom.START_ELEMENT:
            if xmlNode2.localName=='attribute':
              xmlDoc.expandNode(xmlNode2)
              nameAttr=xmlNode2.getAttribute('name')
              if nameAttr=='name':
                decl.swigName=xmlNode2.getAttribute('value')
              elif nameAttr=='type':
                decl.swigType=xmlNode2.getAttribute('value')
              elif nameAttr=='kind':
                decl.swigKind=xmlNode2.getAttribute('value')
              elif nameAttr=='decl':
                decl.swigDecl=xmlNode2.getAttribute('value')
            elif xmlNode2.localName=='parmlist':
              for (evt3,xmlNode3) in xmlDoc:
                if evt3==xml.dom.pulldom.START_ELEMENT:
                  if xmlNode3.localName=='parm':
                    parm=Structure(swigName='',swigType='void')
                    for (evt4,xmlNode4) in xmlDoc:
                      if evt4==xml.dom.pulldom.START_ELEMENT:
                        if xmlNode4.localName=='attributelist':
                          for (evt5,xmlNode5) in xmlDoc:
                            if evt5==xml.dom.pulldom.START_ELEMENT:
                              if xmlNode5.localName=='attribute':
                                xmlDoc.expandNode(xmlNode5)
                                nameAttr=xmlNode5.getAttribute('name')
                                if nameAttr=='name':
                                  parm.swigName=xmlNode5.getAttribute('value')
                                elif nameAttr=='type':
                                  parm.swigType=xmlNode5.getAttribute('value')
                              else:
                                pullIgnored(xmlDoc)
                            elif evt5==xml.dom.pulldom.END_ELEMENT:
                              break
                        else:
                          pullIgnored(xmlDoc)
                      elif evt4==xml.dom.pulldom.END_ELEMENT:
                        break
                    if parm.swigType!='void':
                      decl.swigParmlist.append(parm)
                  else:
                    pullIgnored(xmlDoc)
                elif evt3==xml.dom.pulldom.END_ELEMENT:
                  break
            else:
              pullIgnored(xmlDoc)
          elif evt2==xml.dom.pulldom.END_ELEMENT:
            break
      else:
        pullIgnored(xmlDoc)
    elif evt==xml.dom.pulldom.END_ELEMENT:
      break
  return decl

def pullIgnored(xmlDoc):
  depth=0
  for (evt,xmlNode) in xmlDoc:
    if evt==xml.dom.pulldom.START_ELEMENT:
      depth+=1
    elif evt==xml.dom.pulldom.END_ELEMENT:
      if depth:
        depth-=1
      else:
        break

def showType(t):
  if t=='v(...)':
    return '...'
  t=[i for i in reversed(t.split('.'))]
  s=t[0]
  for o in t[1:]:
    if o=='p':
      if not s.endswith('*'):
        s+=' '
      s+='*'
    elif o.startswith('q(') and o.endswith(')'):
      if s.endswith('*'):
        for o in o[2:-1].split(','):
          s+=' %s'%o
      else:
        for o in o[2:-1].split(','):
          s='%s %s'%(o,s)
  return s

def showFunction(f):
  s='AsyncCall_typedef'
  if f.swigParmlist:
    s+='_%d'%len(f.swigParmlist)
  if f.swigType=='void':
    s+='_void('
  else:
    s+='(%s,'%showType(f.swigType)
  indent=len(s)
  s+='AsyncCall_%s'%f.swigName
  for p in f.swigParmlist:
    s+=',\n%s%s'%(indent*' ',showType(p.swigType))
  s+=')\n'
  return s

if len(sys.argv)<1:
  sys.stderr.write(
             'usage: %s file_name [-i include_dir] [-d macro[=value]]'+
             ' -h [helper_header] header_files ...\n' % sys.argv[0])
  sys.exit(1)

preamble='//'+'-'*76+r'''
// Copyright (C) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
// 
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

// This file is generated by %s
// Do not edit or your changes will be lost!
'''%os.path.basename(sys.argv[0])

swigMissing=''
try:
  subprocess.Popen(['swig','-version'],
                   stdout=subprocess.PIPE,
                   stderr=subprocess.PIPE).communicate()
except:
  swigMissing='Cannot find swig, skipping async-call generation'
  sys.stderr.write('\n%s\n\n'%swigMissing)

outputName=sys.argv[1]
macros={ '_REENTRANT' : None,
         'BYTE_ORDER' : 1234,
         '__attribute__(a)' : None,
         '__x86_64__' : None }
includeDirs=[]
headerFiles=[]
helperHeaders=[]
i=2
while i<len(sys.argv):
  if sys.argv[i]=='-i' and i<len(sys.argv)+1:
    i+=1
    if not sys.argv[i] in includeDirs:
      includeDirs.append(sys.argv[i])
  elif sys.argv[i]=='-d' and i<len(sys.argv)+1:
    i+=1
    macro=sys.argv[i]
    value=None
    affect=macro.find('=')
    if affect>0:
      value=macro[affect+1:]
      macro=macro[:affect]
    macros[macro]=value
  elif sys.argv[i]=='-h' and i<len(sys.argv)+1:
    i+=1
    if not sys.argv[i] in helperHeaders:
      helperHeaders.append(sys.argv[i])
  else:
    if not sys.argv[i] in headerFiles:
      headerFiles.append(sys.argv[i])
  i+=1
includeDirs=[os.path.normpath(i) for i in includeDirs]
outputFile=open(outputName,'w')
outputFile.write('%s\n'%preamble)
guardMacro=os.path.normpath(outputName).upper().replace(os.path.sep,'_') \
                                               .replace(os.path.extsep,'_') \
                                               .replace(':','_')
outputFile.write('#ifndef %s\n#define %s 1\n\n'%(guardMacro,guardMacro))
outputFile.write('#include "TransProg/asyncCall.h"\n\n')
(fd,swigInputName)=tempfile.mkstemp('.i','swig_',os.path.curdir)
os.close(fd)
(fd,swigOutputName)=tempfile.mkstemp('.xml','swig_',os.path.curdir)
os.close(fd)
for f in headerFiles:
  swigInput=open(swigInputName,'w')
  swigInput.write('%module TransProg_AsyncCall\n')
  for (m,v) in macros.items():
    if v:
      swigInput.write('%%define %s %s %%enddef\n'%(m,v))
    else:
      swigInput.write('%%define %s %%enddef\n'%m)
  for h in helperHeaders:
    if h!=f:
      swigInput.write('%%import "%s"\n'%h)
  swigInput.write('%%include "%s"\n'%f)
  swigInput.close()
  swigCmd=['swig','-xml','-cpperraswarn']
  for i in includeDirs:
    swigCmd.append('-I%s'%i)
  swigCmd+=['-o',swigOutputName,swigInputName]
  if swigMissing:
    failure=swigMissing
  else:
    failure=''
    try:
      proc=subprocess.Popen(swigCmd)
      proc.wait()
      if proc.returncode:
        failure='swig failure for file %s: %d'%(f,proc.returncode)
    except:
      traceback.print_exc()
      failure='swig failure for file %s'%f
    if failure:
      sys.stderr.write('\n%s\n\n'%failure)
  outputFile.write('//'+'-'*76+'\n\n')
  outputFile.write('#include "%s"\n\n'%f)
  if failure:
    outputFile.write('// %s\n\n'%failure)
  else:
    if 0:
      for l in open(swigOutputName): sys.stdout.write(l)
    functions=[]
    pullNodes(xml.dom.pulldom.parse(swigOutputName),functions)
    for f in functions:
      outputFile.write('%s\n'%showFunction(f))
outputFile.write('#endif // %s\n\n'%guardMacro)
outputFile.write('//'+'-'*76+'\n')
os.unlink(swigInputName)
os.unlink(swigOutputName)

#-----------------------------------------------------------------------------
