//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransCuda/cuda.h"

#include "TransProg/path.h"

typedef unsigned int GLenum; // FIXME: needed by cudaGL.h
typedef unsigned int GLuint; // FIXME: needed by cudaGL.h
#include <cudaGL.h>

#define OUT_EXT "cubin"

#define CUDA_CALL(function,args)                                 \
        do                                                       \
          {                                                      \
          CUresult cuErr=function args;                          \
          if(cuErr) CUDA_DROP_DEAD(#function"() failure",cuErr); \
          } while(0)

#define DEV_ATTR(value,attrib) \
        CUDA_CALL(cuDeviceGetAttribute,(&dev->value,attrib,dev->id))

#define FUNC_ATTR(value,attrib) \
        CUDA_CALL(cuFuncGetAttribute,(&program->value,attrib,program->kernel))

TRANSCUDA_API_VAR const void *_CudaBuffer_nullArg=(const void *)0;

typedef struct _CudaManager
{
bool initDone;
int initCpuCtx;
int deviceCount;
CudaDevice *devices;
const CudaDevice *glDevice;
} CudaManager;

static CudaManager mgr;

void
Cuda_init(void)
{
if(!mgr.initDone)
  {
  mgr.initDone=true;
  mgr.initCpuCtx=Platform_cpuCtx();
  int i;
  CUDA_CALL(cuInit,(0));
  CUDA_CALL(cuDeviceGetCount,(&mgr.deviceCount));
  mgr.glDevice=(const CudaDevice *)0;
  if(mgr.deviceCount)
    {
    unsigned int glDeviceCount=0;
    CUdevice *glDevices=Memory_alloc(mgr.deviceCount,CUdevice);
    cuGLGetDevices(&glDeviceCount,glDevices,mgr.deviceCount,
                   CU_GL_DEVICE_LIST_CURRENT_FRAME); // don't care for failure
    mgr.devices=Memory_alloc(mgr.deviceCount,CudaDevice);
    for(i=0;i<mgr.deviceCount;++i)
      {
      CudaDevice *dev=mgr.devices+i;
      CUDA_CALL(cuDeviceGet,(&dev->id,i));
      for(unsigned int j=0;j<glDeviceCount;++j)
        {
        if(glDevices[j]==dev->id)
          {
          mgr.glDevice=dev;
          break;
          }
        }
      if(dev==mgr.glDevice)
        {
        CUDA_CALL(cuGLCtxCreate,(&dev->context,
                                 CU_CTX_SCHED_SPIN|CU_CTX_MAP_HOST,dev->id));
        }
      else
        {
        CUDA_CALL(cuCtxCreate,(&dev->context,
                               CU_CTX_SCHED_SPIN|CU_CTX_MAP_HOST,dev->id));
        }
      CUDA_CALL(cuCtxSetCurrent,(dev->context));
      CUDA_CALL(cuStreamCreate,(&dev->stream,0));
      dev->peerMask=0;
      if(dev->id>=(int)(8*sizeof(dev->peerMask)))
        {
        DROP_DEAD("insufficient width for CudaDevice.peerMask\n");
        }
      dev->name=Memory_alloc(0x100,char);
      CUDA_CALL(cuDeviceGetName,(dev->name,0x100,dev->id));
      CUDA_CALL(cuDeviceComputeCapability,(&dev->computeCapabilityMajor,
                                           &dev->computeCapabilityMinor,
                                           dev->id));
      CUDA_CALL(cuDeviceTotalMem,(&dev->totalMemory,dev->id));
      DEV_ATTR(max_threads_per_block,
               CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_BLOCK);
      DEV_ATTR(max_block_dim_x,
               CU_DEVICE_ATTRIBUTE_MAX_BLOCK_DIM_X);
      DEV_ATTR(max_block_dim_y,
               CU_DEVICE_ATTRIBUTE_MAX_BLOCK_DIM_Y);
      DEV_ATTR(max_block_dim_z,
               CU_DEVICE_ATTRIBUTE_MAX_BLOCK_DIM_Z);
      DEV_ATTR(max_grid_dim_x,
               CU_DEVICE_ATTRIBUTE_MAX_GRID_DIM_X);
      DEV_ATTR(max_grid_dim_y,
               CU_DEVICE_ATTRIBUTE_MAX_GRID_DIM_Y);
      DEV_ATTR(max_grid_dim_z,
               CU_DEVICE_ATTRIBUTE_MAX_GRID_DIM_Z);
      DEV_ATTR(max_shared_memory_per_block,
               CU_DEVICE_ATTRIBUTE_MAX_SHARED_MEMORY_PER_BLOCK);
      DEV_ATTR(total_constant_memory,
               CU_DEVICE_ATTRIBUTE_TOTAL_CONSTANT_MEMORY);
      DEV_ATTR(warp_size,
               CU_DEVICE_ATTRIBUTE_WARP_SIZE);
      // DEV_ATTR(max_pitch,
      //          CU_DEVICE_ATTRIBUTE_MAX_PITCH);
      // DEV_ATTR(maximum_texture1d_width,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE1D_WIDTH);
      // DEV_ATTR(maximum_texture1d_linear_width,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE1D_LINEAR_WIDTH);
      // DEV_ATTR(maximum_texture2d_width,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE2D_WIDTH);
      // DEV_ATTR(maximum_texture2d_height,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE2D_HEIGHT);
      // DEV_ATTR(maximum_texture2d_linear_width,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE2D_LINEAR_WIDTH);
      // DEV_ATTR(maximum_texture2d_linear_height,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE2D_LINEAR_HEIGHT);
      // DEV_ATTR(maximum_texture2d_linear_pitch,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE2D_LINEAR_PITCH);
      // DEV_ATTR(maximum_texture3d_width,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE3D_WIDTH);
      // DEV_ATTR(maximum_texture3d_height,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE3D_HEIGHT);
      // DEV_ATTR(maximum_texture3d_depth,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE3D_DEPTH);
      // DEV_ATTR(maximum_texture3d_width_alternate,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE3D_WIDTH_ALTERNATE);
      // DEV_ATTR(maximum_texture3d_height_alternate,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE3D_HEIGHT_ALTERNATE);
      // DEV_ATTR(maximum_texture3d_depth_alternate,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE3D_DEPTH_ALTERNATE);
      // DEV_ATTR(maximum_texturecubemap_width,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURECUBEMAP_WIDTH);
      // DEV_ATTR(maximum_texture1d_layered_width,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE1D_LAYERED_WIDTH);
      // DEV_ATTR(maximum_texture1d_layered_layers,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE1D_LAYERED_LAYERS);
      // DEV_ATTR(maximum_texture2d_layered_width,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE2D_LAYERED_WIDTH);
      // DEV_ATTR(maximum_texture2d_layered_height,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE2D_LAYERED_HEIGHT);
      // DEV_ATTR(maximum_texture2d_layered_layers,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE2D_LAYERED_LAYERS);
      // DEV_ATTR(maximum_texturecubemap_layered_width,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURECUBEMAP_LAYERED_WIDTH);
      // DEV_ATTR(maximum_texturecubemap_layered_layers,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURECUBEMAP_LAYERED_LAYERS);
      // DEV_ATTR(maximum_surface1d_width,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_SURFACE1D_WIDTH);
      // DEV_ATTR(maximum_surface2d_width,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_SURFACE2D_WIDTH);
      // DEV_ATTR(maximum_surface2d_height,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_SURFACE2D_HEIGHT);
      // DEV_ATTR(maximum_surface3d_width,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_SURFACE3D_WIDTH);
      // DEV_ATTR(maximum_surface3d_height,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_SURFACE3D_HEIGHT);
      // DEV_ATTR(maximum_surface3d_depth,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_SURFACE3D_DEPTH);
      // DEV_ATTR(maximum_surface1d_layered_width,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_SURFACE1D_LAYERED_WIDTH);
      // DEV_ATTR(maximum_surface1d_layered_layers,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_SURFACE1D_LAYERED_LAYERS);
      // DEV_ATTR(maximum_surface2d_layered_width,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_SURFACE2D_LAYERED_WIDTH);
      // DEV_ATTR(maximum_surface2d_layered_height,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_SURFACE2D_LAYERED_HEIGHT);
      // DEV_ATTR(maximum_surface2d_layered_layers,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_SURFACE2D_LAYERED_LAYERS);
      // DEV_ATTR(maximum_surfacecubemap_width,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_SURFACECUBEMAP_WIDTH);
      // DEV_ATTR(maximum_surfacecubemap_layered_width,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_SURFACECUBEMAP_LAYERED_WIDTH);
      // DEV_ATTR(maximum_surfacecubemap_layered_layers,
      //          CU_DEVICE_ATTRIBUTE_MAXIMUM_SURFACECUBEMAP_LAYERED_LAYERS);
      DEV_ATTR(max_registers_per_block,
               CU_DEVICE_ATTRIBUTE_MAX_REGISTERS_PER_BLOCK);
      DEV_ATTR(clock_rate,
               CU_DEVICE_ATTRIBUTE_CLOCK_RATE);
      // DEV_ATTR(texture_alignment,
      //          CU_DEVICE_ATTRIBUTE_TEXTURE_ALIGNMENT);
      // DEV_ATTR(texture_pitch_alignment,
      //          CU_DEVICE_ATTRIBUTE_TEXTURE_PITCH_ALIGNMENT);
      DEV_ATTR(gpu_overlap,
               CU_DEVICE_ATTRIBUTE_GPU_OVERLAP);
      DEV_ATTR(multiprocessor_count,
               CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT);
      switch(dev->computeCapabilityMajor*10+dev->computeCapabilityMinor)
        {
        case 10:
        case 11:
        case 12:
        case 13:
          {
          dev->core_count=dev->multiprocessor_count*8;
          break;
          }
        case 20:
          {
          dev->core_count=dev->multiprocessor_count*32;
          break;
          }
        case 21:
          {
          dev->core_count=dev->multiprocessor_count*48;
          break;
          }
        case 30:
        case 35:
          {
          dev->core_count=dev->multiprocessor_count*192;
          break;
          }
        default:
          {
          dev->core_count=0;
          Platform_msg("warning: unknown compute capability"
                       " %d.%d for GPU device %s\n",
                       dev->computeCapabilityMajor,
                       dev->computeCapabilityMinor,
                       dev->name);
          }
        }
      DEV_ATTR(kernel_exec_timeout,
               CU_DEVICE_ATTRIBUTE_KERNEL_EXEC_TIMEOUT);
      DEV_ATTR(integrated,
               CU_DEVICE_ATTRIBUTE_INTEGRATED);
      DEV_ATTR(can_map_host_memory,
               CU_DEVICE_ATTRIBUTE_CAN_MAP_HOST_MEMORY);
      // DEV_ATTR(compute_mode,
      //          CU_DEVICE_ATTRIBUTE_COMPUTE_MODE);
      DEV_ATTR(concurrent_kernels,
               CU_DEVICE_ATTRIBUTE_CONCURRENT_KERNELS);
      DEV_ATTR(ecc_enabled,
               CU_DEVICE_ATTRIBUTE_ECC_ENABLED);
      DEV_ATTR(pci_bus_id,
               CU_DEVICE_ATTRIBUTE_PCI_BUS_ID);
      DEV_ATTR(pci_device_id,
               CU_DEVICE_ATTRIBUTE_PCI_DEVICE_ID);
      DEV_ATTR(tcc_driver,
               CU_DEVICE_ATTRIBUTE_TCC_DRIVER);
      DEV_ATTR(memory_clock_rate,
               CU_DEVICE_ATTRIBUTE_MEMORY_CLOCK_RATE);
      DEV_ATTR(global_memory_bus_width,
               CU_DEVICE_ATTRIBUTE_GLOBAL_MEMORY_BUS_WIDTH);
      DEV_ATTR(l2_cache_size,
               CU_DEVICE_ATTRIBUTE_L2_CACHE_SIZE);
      DEV_ATTR(max_threads_per_multiprocessor,
               CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_MULTIPROCESSOR);
      DEV_ATTR(unified_addressing,
               CU_DEVICE_ATTRIBUTE_UNIFIED_ADDRESSING);
      }
    Memory_free(glDevices);
    }
  else
    {
    mgr.devices=(CudaDevice *)0;
    }
  for(i=0;i<mgr.deviceCount;++i) // sort devices to ease peer access
    {
    CudaDevice *devA=mgr.devices+i;
    CudaDevice *bestDev=devA;
    for(int j=i+1;j<mgr.deviceCount;++j)
      {
      CudaDevice *devB=mgr.devices+j;
#define COMPARE_FIELD(f) if(devB->f>bestDev->f) { bestDev=devB; continue; } \
                         if(devB->f<bestDev->f) { continue; }
      COMPARE_FIELD(computeCapabilityMajor)
      COMPARE_FIELD(computeCapabilityMinor)
      COMPARE_FIELD(multiprocessor_count)
      COMPARE_FIELD(clock_rate)
      COMPARE_FIELD(totalMemory)
      COMPARE_FIELD(memory_clock_rate)
      }
    if(bestDev!=devA)
      {
      CudaDevice tmp=*devA;
      *devA=*bestDev;
      *bestDev=tmp;
      if(mgr.glDevice==devA)
        {
        mgr.glDevice=bestDev;
        }
      else if(mgr.glDevice==bestDev)
        {
        mgr.glDevice=devA;
        }
      }
    }
  for(i=0;i<mgr.deviceCount;++i)
    {
    CudaDevice *devA=mgr.devices+i;
    for(int j=i+1;j<mgr.deviceCount;++j)
      {
      CudaDevice *devB=mgr.devices+j;
      int canAccessPeer;
      CUDA_CALL(cuDeviceCanAccessPeer,(&canAccessPeer,devA->id,devB->id));
      if(canAccessPeer)
        {
        devA->peerMask|=1ULL<<devB->id;
        CUDA_CALL(cuCtxSetCurrent,(devA->context));
        CUDA_CALL(cuCtxEnablePeerAccess,(devB->context,0));
        }
      CUDA_CALL(cuDeviceCanAccessPeer,(&canAccessPeer,devB->id,devA->id));
      if(canAccessPeer)
        {
        devB->peerMask|=1ULL<<devA->id;
        CUDA_CALL(cuCtxSetCurrent,(devB->context));
        CUDA_CALL(cuCtxEnablePeerAccess,(devA->context,0));
        }
      }
    }
  }
}

void
Cuda_errorMsg(int errorCode,
              char *out_msg,
              int msgSize)
{
const char *errorMsg;
switch(errorCode)
  {
  case CUDA_SUCCESS:
    errorMsg="CUDA_SUCCESS"; break;
  case CUDA_ERROR_INVALID_VALUE:
    errorMsg="CUDA_ERROR_INVALID_VALUE"; break;
  case CUDA_ERROR_OUT_OF_MEMORY:
    errorMsg="CUDA_ERROR_OUT_OF_MEMORY"; break;
  case CUDA_ERROR_NOT_INITIALIZED:
    errorMsg="CUDA_ERROR_NOT_INITIALIZED"; break;
  case CUDA_ERROR_DEINITIALIZED:
    errorMsg="CUDA_ERROR_DEINITIALIZED"; break;
  case CUDA_ERROR_PROFILER_DISABLED:
    errorMsg="CUDA_ERROR_PROFILER_DISABLED"; break;
  case CUDA_ERROR_PROFILER_NOT_INITIALIZED:
    errorMsg="CUDA_ERROR_PROFILER_NOT_INITIALIZED"; break;
  case CUDA_ERROR_PROFILER_ALREADY_STARTED:
    errorMsg="CUDA_ERROR_PROFILER_ALREADY_STARTED"; break;
  case CUDA_ERROR_PROFILER_ALREADY_STOPPED:
    errorMsg="CUDA_ERROR_PROFILER_ALREADY_STOPPED"; break;
  case CUDA_ERROR_NO_DEVICE:
    errorMsg="CUDA_ERROR_NO_DEVICE"; break;
  case CUDA_ERROR_INVALID_DEVICE:
    errorMsg="CUDA_ERROR_INVALID_DEVICE"; break;
  case CUDA_ERROR_INVALID_IMAGE:
    errorMsg="CUDA_ERROR_INVALID_IMAGE"; break;
  case CUDA_ERROR_INVALID_CONTEXT:
    errorMsg="CUDA_ERROR_INVALID_CONTEXT"; break;
  case CUDA_ERROR_CONTEXT_ALREADY_CURRENT:
    errorMsg="CUDA_ERROR_CONTEXT_ALREADY_CURRENT"; break;
  case CUDA_ERROR_MAP_FAILED:
    errorMsg="CUDA_ERROR_MAP_FAILED"; break;
  case CUDA_ERROR_UNMAP_FAILED:
    errorMsg="CUDA_ERROR_UNMAP_FAILED"; break;
  case CUDA_ERROR_ARRAY_IS_MAPPED:
    errorMsg="CUDA_ERROR_ARRAY_IS_MAPPED"; break;
  case CUDA_ERROR_ALREADY_MAPPED:
    errorMsg="CUDA_ERROR_ALREADY_MAPPED"; break;
  case CUDA_ERROR_NO_BINARY_FOR_GPU:
    errorMsg="CUDA_ERROR_NO_BINARY_FOR_GPU"; break;
  case CUDA_ERROR_ALREADY_ACQUIRED:
    errorMsg="CUDA_ERROR_ALREADY_ACQUIRED"; break;
  case CUDA_ERROR_NOT_MAPPED:
    errorMsg="CUDA_ERROR_NOT_MAPPED"; break;
  case CUDA_ERROR_NOT_MAPPED_AS_ARRAY:
    errorMsg="CUDA_ERROR_NOT_MAPPED_AS_ARRAY"; break;
  case CUDA_ERROR_NOT_MAPPED_AS_POINTER:
    errorMsg="CUDA_ERROR_NOT_MAPPED_AS_POINTER"; break;
  case CUDA_ERROR_ECC_UNCORRECTABLE:
    errorMsg="CUDA_ERROR_ECC_UNCORRECTABLE"; break;
  case CUDA_ERROR_UNSUPPORTED_LIMIT:
    errorMsg="CUDA_ERROR_UNSUPPORTED_LIMIT"; break;
  case CUDA_ERROR_CONTEXT_ALREADY_IN_USE:
    errorMsg="CUDA_ERROR_CONTEXT_ALREADY_IN_USE"; break;
  case CUDA_ERROR_INVALID_SOURCE:
    errorMsg="CUDA_ERROR_INVALID_SOURCE"; break;
  case CUDA_ERROR_FILE_NOT_FOUND:
    errorMsg="CUDA_ERROR_FILE_NOT_FOUND"; break;
  case CUDA_ERROR_SHARED_OBJECT_SYMBOL_NOT_FOUND:
    errorMsg="CUDA_ERROR_SHARED_OBJECT_SYMBOL_NOT_FOUND"; break;
  case CUDA_ERROR_SHARED_OBJECT_INIT_FAILED:
    errorMsg="CUDA_ERROR_SHARED_OBJECT_INIT_FAILED"; break;
  case CUDA_ERROR_OPERATING_SYSTEM:
    errorMsg="CUDA_ERROR_OPERATING_SYSTEM"; break;
  case CUDA_ERROR_INVALID_HANDLE:
    errorMsg="CUDA_ERROR_INVALID_HANDLE"; break;
  case CUDA_ERROR_NOT_FOUND:
    errorMsg="CUDA_ERROR_NOT_FOUND"; break;
  case CUDA_ERROR_NOT_READY:
    errorMsg="CUDA_ERROR_NOT_READY"; break;
  case CUDA_ERROR_LAUNCH_FAILED:
    errorMsg="CUDA_ERROR_LAUNCH_FAILED"; break;
  case CUDA_ERROR_LAUNCH_OUT_OF_RESOURCES:
    errorMsg="CUDA_ERROR_LAUNCH_OUT_OF_RESOURCES"; break;
  case CUDA_ERROR_LAUNCH_TIMEOUT:
    errorMsg="CUDA_ERROR_LAUNCH_TIMEOUT"; break;
  case CUDA_ERROR_LAUNCH_INCOMPATIBLE_TEXTURING:
    errorMsg="CUDA_ERROR_LAUNCH_INCOMPATIBLE_TEXTURING"; break;
  case CUDA_ERROR_PEER_ACCESS_ALREADY_ENABLED:
    errorMsg="CUDA_ERROR_PEER_ACCESS_ALREADY_ENABLED"; break;
  case CUDA_ERROR_PEER_ACCESS_NOT_ENABLED:
    errorMsg="CUDA_ERROR_PEER_ACCESS_NOT_ENABLED"; break;
  case CUDA_ERROR_PRIMARY_CONTEXT_ACTIVE:
    errorMsg="CUDA_ERROR_PRIMARY_CONTEXT_ACTIVE"; break;
  case CUDA_ERROR_CONTEXT_IS_DESTROYED:
    errorMsg="CUDA_ERROR_CONTEXT_IS_DESTROYED"; break;
  case CUDA_ERROR_ASSERT:
    errorMsg="CUDA_ERROR_ASSERT"; break;
  case CUDA_ERROR_TOO_MANY_PEERS:
    errorMsg="CUDA_ERROR_TOO_MANY_PEERS"; break;
  case CUDA_ERROR_HOST_MEMORY_ALREADY_REGISTERED:
    errorMsg="CUDA_ERROR_HOST_MEMORY_ALREADY_REGISTERED"; break;
  case CUDA_ERROR_HOST_MEMORY_NOT_REGISTERED:
    errorMsg="CUDA_ERROR_HOST_MEMORY_NOT_REGISTERED"; break;
  case CUDA_ERROR_UNKNOWN:
    errorMsg="CUDA_ERROR_UNKNOWN"; break;
  default:
    errorMsg="unknown Cuda error";
  }
snprintf(out_msg,msgSize,"%s (%d)",errorMsg,errorCode);
out_msg[msgSize-1]='\0'; // ensure string termination
}

const CudaDevice *
Cuda_devices(int *out_deviceCount)
{
*out_deviceCount=mgr.deviceCount;
return mgr.devices;
}

const CudaDevice * // device used for OpenGL rendering or 0
Cuda_glDevice(void)
{
return mgr.glDevice;
}

intptr_t
CudaDevice_freeMemory(const CudaDevice *device)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
CUDA_CALL(cuCtxSetCurrent,(device->context));
size_t freeMem,totalMem;
CUDA_CALL(cuMemGetInfo,(&freeMem,&totalMem));
return freeMem;
}

void
CudaDevice_sync(const CudaDevice *device)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
CUDA_CALL(cuCtxSetCurrent,(device->context));
CUDA_CALL(cuStreamSynchronize,(device->stream));
}

void
CudaMarker_init(CudaMarker *marker,
                const CudaDevice *device)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
CUDA_CALL(cuCtxSetCurrent,(device->context));
marker->device=device;
CUDA_CALL(cuEventCreate,(&marker->event,CU_EVENT_DEFAULT));
}

void
CudaMarker_destroy(CudaMarker *marker)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
CUDA_CALL(cuCtxSetCurrent,(marker->device->context));
CUDA_CALL(cuEventDestroy,(marker->event));
}

void
CudaMarker_set(CudaMarker *marker)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
CUDA_CALL(cuCtxSetCurrent,(marker->device->context));
CUDA_CALL(cuEventRecord,(marker->event,marker->device->stream));
}

bool // previous work is done
CudaMarker_test(const CudaMarker *marker)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
// CUDA_CALL(cuCtxSetCurrent,(marker->device->context)); // not necessary
CUresult cuErr=cuEventQuery(marker->event);
switch(cuErr)
  {
  case CUDA_SUCCESS: return true;
  case CUDA_ERROR_NOT_READY: return false;
  default: CUDA_DROP_DEAD("cuEventQuery() failure",cuErr); return false;
  }
}

void
CudaMarker_sync(CudaMarker *marker)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
// CUDA_CALL(cuCtxSetCurrent,(marker->device->context)); // not necessary
CUDA_CALL(cuEventSynchronize,(marker->event));
}

int64_t // microseconds
CudaMarker_duration(const CudaMarker *marker,
                    const CudaMarker *previous)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
// CUDA_CALL(cuCtxSetCurrent,(marker->device->context)); // not necessary
float milliseconds;
CUDA_CALL(cuEventElapsedTime,(&milliseconds,previous->event,marker->event));
return (int64_t)(1.0e3*(real64_t)milliseconds);
}

static
void
CudaProgram_initCommon(CudaProgram *program,
                       const CudaDevice *device,
                       const char *name,
                       const char *options,
                       const char *sourceCode,
                       const void *binaryCode,
                       intptr_t binaryCodeSize,
                       bool prefersCacheToShared)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
CUDA_CALL(cuCtxSetCurrent,(device->context));
program->device=device;
program->name=Memory_allocStr(name);
if(options&&*options)
  {
  program->options=Memory_allocStr(options);
  }
else
  {
  program->options=(char *)0;
  }
if(sourceCode)
  {
  program->sourceCode=Memory_allocStr(sourceCode);
  }
else
  {
  program->sourceCode=(char *)0;
  }
if(binaryCode)
  {
  program->binaryCode=Memory_alloc(binaryCodeSize+1,uint8_t);
  memcpy(program->binaryCode,binaryCode,binaryCodeSize);
  program->binaryCode[binaryCodeSize]=0;
  program->binaryCodeSize=binaryCodeSize;
  }
else
  {
  program->binaryCode=(uint8_t *)0;
  program->binaryCodeSize=0;
  }
program->prefersCacheToShared=prefersCacheToShared;
Text_init(&program->buildLog);
program->module=(CUmodule)0;
program->kernel=(CUfunction)0;
CUresult cuErr;
if(program->sourceCode)
  {
  char inFile[0x100]="";
  char outFile[0x100]="";
  char logFile[0x100]="";
  intptr_t sz=(int)strlen(program->sourceCode);
  if(!Path_createTmpFile(inFile,0x100,".","cuda_??????.cu"))
    {
    char msg[0x100];
    Platform_lastErrorMsg(msg,0x100);
    Text_append(&program->buildLog,
                "creating source file: %s\n",msg);
    }
  else if(!Path_createTmpFile(outFile,0x100,".","cuda_??????." OUT_EXT))
    {
    char msg[0x100];
    Platform_lastErrorMsg(msg,0x100);
    Text_append(&program->buildLog,
                "creating output file: %s\n",msg);
    }
  else if(!Path_createTmpFile(logFile,0x100,".","cuda_??????.log"))
    {
    char msg[0x100];
    Platform_lastErrorMsg(msg,0x100);
    Text_append(&program->buildLog,
                "creating log file: %s\n",msg);
    }
  else if(!Path_writeFile(inFile,program->sourceCode,sz,false)==sz)
    {
    char msg[0x100];
    Platform_lastErrorMsg(msg,0x100);
    Text_append(&program->buildLog,
                "writing to source file %s: %s\n",inFile,msg);
    }
  else
    {
    char buffer[0x400];
    snprintf(buffer,0x400,
             "nvcc -arch=sm_%d%d %s --%s -o %s %s >%s 2>&1",
             CudaDevice_computeCapabilityMajor(program->device),
             CudaDevice_computeCapabilityMinor(program->device),
             program->options ? program->options : "",
             OUT_EXT,outFile,inFile,logFile);
    buffer[0x3FF]='\0';
    int status=system(buffer);
#if defined _WIN32
    if(status==-1)
      {
      int errorCode=errno;
      Text_append(&program->buildLog,
                  "system(\"%s\") failure --- %s (%d)\n",
                  buffer,strerror(errorCode),errorCode);
      }
    else if(status)
      {
      Text_append(&program->buildLog,
                  "system(\"%s\") returned %d\n",
                  buffer,status);
      }
#else
    if(WIFSIGNALED(status))
      {
      Text_append(&program->buildLog,
                  "system(\"%s\") terminated by signal %d\n",
                  buffer,WTERMSIG(status));
      }
    else if(WEXITSTATUS(status))
      {
      Text_append(&program->buildLog,
                  "system(\"%s\") returned %d\n",
                  buffer,WEXITSTATUS(status));
      }
#endif
    sz=Path_fileSize(logFile);
    if(sz>0)
      {
      char *buildLog=Memory_alloc(sz+1,char);
      if(Path_readFile(logFile,buildLog,sz,0)==sz)
        {
        buildLog[sz]='\0';
        Text_appendStr(&program->buildLog,buildLog);
        }
      Memory_free(buildLog);
      }
    sz=Path_fileSize(outFile);
    if(sz>0)
      {
      uint8_t *binaryCode=Memory_alloc(sz+1,uint8_t);
      if(Path_readFile(outFile,binaryCode,sz,0)==sz)
        {
        binaryCode[sz]='\0';
        program->binaryCode=binaryCode;
        program->binaryCodeSize=sz;
        }
      else
        {
        Memory_free(binaryCode);
        }
      }
    }
  if(inFile[0])
    {
    Path_remove(inFile);
    }
  if(outFile[0])
    {
    Path_remove(outFile);
    }
  if(logFile[0])
    {
    Path_remove(logFile);
    }
  }
if(program->binaryCode)
  {
  cuErr=cuModuleLoadData(&program->module,program->binaryCode);
  if(cuErr)
    {
    char msg[0x100];
    Cuda_errorMsg(cuErr,msg,0x100);
    Text_append(&program->buildLog,
                "cuModuleLoadData() failure: %s\n",msg);
    }
  }
if(program->module)
  {
  cuErr=cuModuleGetFunction(&program->kernel,program->module,program->name);
  if(cuErr)
    {
    char msg[0x100];
    Cuda_errorMsg(cuErr,msg,0x100);
    Text_append(&program->buildLog,
                "cuModuleGetFunction(%s) failure: %s\n",program->name,msg);
    }
  }
if(!strcmp(Text_get(&program->buildLog),"\n")||
   !strcmp(Text_get(&program->buildLog),"\r\n")||
   !strcmp(Text_get(&program->buildLog),"\r"))
  {
  Text_clear(&program->buildLog);
  }
if(program->kernel)
  {
  if(program->prefersCacheToShared)
    {
    CUDA_CALL(cuFuncSetCacheConfig,(program->kernel,CU_FUNC_CACHE_PREFER_L1));
    }
  FUNC_ATTR(max_threads_per_block,
            CU_FUNC_ATTRIBUTE_MAX_THREADS_PER_BLOCK);
  FUNC_ATTR(shared_size_bytes,
            CU_FUNC_ATTRIBUTE_SHARED_SIZE_BYTES);
  FUNC_ATTR(const_size_bytes,
            CU_FUNC_ATTRIBUTE_CONST_SIZE_BYTES);
  FUNC_ATTR(local_size_bytes,
            CU_FUNC_ATTRIBUTE_LOCAL_SIZE_BYTES);
  FUNC_ATTR(num_regs,
            CU_FUNC_ATTRIBUTE_NUM_REGS);
  FUNC_ATTR(ptx_version,
            CU_FUNC_ATTRIBUTE_PTX_VERSION);
  FUNC_ATTR(binary_version,
            CU_FUNC_ATTRIBUTE_BINARY_VERSION);
  }
}

void
CudaProgram_init(CudaProgram *program,
                 const CudaDevice *device,
                 const char *name,
                 const char *options,
                 const char *sourceCode,
                 bool prefersCacheToShared)
{
CudaProgram_initCommon(program,device,name,
                       options,sourceCode,
                       (const char *)0,0,
                       prefersCacheToShared);
}

void
CudaProgram_initFromBinary(CudaProgram *program,
                           const CudaDevice *device,
                           const char *name,
                           const void *binaryCode,
                           intptr_t binaryCodeSize,
                           bool prefersCacheToShared)
{
CudaProgram_initCommon(program,device,name,
                       (const char *)0,(const char *)0,
                       binaryCode,binaryCodeSize,
                       prefersCacheToShared);
}

void
CudaProgram_destroy(CudaProgram *program)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
CUDA_CALL(cuCtxSetCurrent,(program->device->context));
if(program->module)
  {
  CUDA_CALL(cuModuleUnload,(program->module));
  }
Text_destroy(&program->buildLog);
if(program->binaryCode)
  {
  Memory_free(program->binaryCode);
  }
if(program->sourceCode)
  {
  Memory_free(program->sourceCode);
  }
if(program->options)
  {
  Memory_free(program->options);
  }
Memory_free(program->name);
}

void
CudaProgram_launch(const CudaProgram *program,
                   int gridDimX,
                   int gridDimY,
                   int gridDimZ,
                   int blockDimX,
                   int blockDimY,
                   int blockDimZ,
                   int sharedMemorySize,
                   const void * const *args)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
CUDA_CALL(cuCtxSetCurrent,(program->device->context));
CUDA_CALL(cuLaunchKernel,(program->kernel,
                          gridDimX,gridDimY,gridDimZ,
                          blockDimX,blockDimY,blockDimZ,
                          sharedMemorySize,
                          program->device->stream,
                          (void **)args,(void **)0));
}

void
CudaBuffer_init(CudaBuffer *buffer,
                const CudaDevice *device,
                intptr_t size)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
CUDA_CALL(cuCtxSetCurrent,(device->context));
buffer->device=device;
buffer->size=size;
buffer->devptr=(CUdeviceptr)0;
buffer->gl=0;
CUDA_CALL(cuMemAlloc,(&buffer->devptr,buffer->size));
}

void
CudaBuffer_initFromGl(CudaBuffer *buffer,
                      const CudaDevice *device,
                      int glBuffer)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
CUDA_CALL(cuCtxSetCurrent,(device->context));
buffer->device=device;
buffer->size=0;
buffer->devptr=(CUdeviceptr)0;
buffer->gfxResource=(CUgraphicsResource)0;
buffer->gl=glBuffer;
CUDA_CALL(cuGraphicsGLRegisterBuffer,(&buffer->gfxResource,
                                      glBuffer,
                                      CU_GRAPHICS_REGISTER_FLAGS_NONE));
CUDA_CALL(cuGraphicsMapResources,(1,&buffer->gfxResource,
                                  buffer->device->stream));
// ensure mapping is done
CUDA_CALL(cuStreamSynchronize,(buffer->device->stream));
CUDA_CALL(cuGraphicsResourceGetMappedPointer,(&buffer->devptr,
                                              &buffer->size,
                                              buffer->gfxResource));
CUDA_CALL(cuGraphicsUnmapResources,(1,&buffer->gfxResource,
                                    buffer->device->stream));
}

void
CudaBuffer_destroy(CudaBuffer *buffer)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
CUDA_CALL(cuCtxSetCurrent,(buffer->device->context));
if(buffer->gl)
  {
  CUDA_CALL(cuGraphicsUnregisterResource,(buffer->gfxResource));
  }
else
  {
  CUDA_CALL(cuMemFree,(buffer->devptr));
  }
}

void
CudaBuffer_acquireGl(CudaBuffer *buffer)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
CUDA_CALL(cuCtxSetCurrent,(buffer->device->context));
CUDA_CALL(cuGraphicsMapResources,(1,&buffer->gfxResource,
                                  buffer->device->stream));
}

void
CudaBuffer_releaseGl(CudaBuffer *buffer)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
CUDA_CALL(cuCtxSetCurrent,(buffer->device->context));
CUDA_CALL(cuGraphicsUnmapResources,(1,&buffer->gfxResource,
                                    buffer->device->stream));
}

void
CudaBuffer_send(CudaBuffer *buffer,
                intptr_t offset,
                intptr_t size,
                const void *data)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
CUDA_CALL(cuCtxSetCurrent,(buffer->device->context));
CUDA_CALL(cuMemcpyHtoDAsync,(buffer->devptr+offset,data,size,
                             buffer->device->stream));
}

void
CudaBuffer_receive(CudaBuffer *buffer,
                   intptr_t offset,
                   intptr_t size,
                   void *data)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
CUDA_CALL(cuCtxSetCurrent,(buffer->device->context));
CUDA_CALL(cuMemcpyDtoHAsync,(data,buffer->devptr+offset,size,
                             buffer->device->stream));
}

void
CudaBuffer_copy(CudaBuffer *destBuffer,
                const CudaBuffer *srcBuffer,
                intptr_t destOffset,
                intptr_t srcOffset,
                intptr_t size)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
CUDA_CALL(cuCtxSetCurrent,(destBuffer->device->context));
if(destBuffer->device==srcBuffer->device)
  {
  CUDA_CALL(cuMemcpyDtoDAsync,(destBuffer->devptr+destOffset,
                               srcBuffer->devptr+srcOffset,
                               size,
                               destBuffer->device->stream));
  }
else
  {
  CUDA_CALL(cuMemcpyPeerAsync,(destBuffer->devptr+destOffset,
                               destBuffer->device->context,
                               srcBuffer->devptr+srcOffset,
                               srcBuffer->device->context,
                               size,
                               destBuffer->device->stream));
  }
}

void
CudaLockedMem_init(CudaLockedMem *mem,
                   bool writeOnly,
                   intptr_t size)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
// no need to select a specific context with CU_MEMHOSTALLOC_PORTABLE 
mem->data=(void *)0;
mem->size=size;
mem->writeOnly=writeOnly;
unsigned int flags=CU_MEMHOSTALLOC_PORTABLE|CU_MEMHOSTALLOC_DEVICEMAP;
if(mem->writeOnly)
  {
  flags|=CU_MEMHOSTALLOC_WRITECOMBINED;
  }
CUDA_CALL(cuMemHostAlloc,(&mem->data,mem->size,flags));
// FIXME: is this really not specific to any device?
CUDA_CALL(cuMemHostGetDevicePointer,(&mem->devptr,mem->data,0));
}

void
CudaLockedMem_destroy(CudaLockedMem *mem)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
CUDA_CALL(cuMemFreeHost,(mem->data));
}

//----------------------------------------------------------------------------
