//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransGl/glDraw.h"

#include "TransProg/text.h"

#define INTEGER_GL_TYPE(integerType,sizeOfInteger)             \
        switch(sizeOfInteger)                                  \
          {                                                    \
          case 1:  integerType=GL_UNSIGNED_BYTE;  break;       \
          case 2:  integerType=GL_UNSIGNED_SHORT; break;       \
          case 4:  integerType=GL_UNSIGNED_INT;   break;       \
          default: integerType=0;                              \
                   DROP_DEAD("bad value for " #sizeOfInteger); \
          }

#define REAL_GL_TYPE(realType,sizeOfReal)                   \
        switch(sizeOfReal)                                  \
          {                                                 \
          case 2:  if(!CoreGl_has_ARB_half_float_vertex())  \
                     {                                      \
                     DROP_DEAD("no OpenGL half-float");     \
                     }                                      \
                   realType=GL_HALF_FLOAT; break;           \
          case 4:  realType=GL_FLOAT;      break;           \
          case 8:  realType=GL_DOUBLE;     break;           \
          default: realType=0;                              \
                   DROP_DEAD("bad value for " #sizeOfReal); \
          }

typedef struct _ShaderInfo
{
Text vertexSource;
Text geometrySource;
Text fragmentSource;
GlShader shader;
int v_Vertex;
int v_Normal;
int v_TexCoord;
int v_Color;
int v_Radius;
int u_ProjectionMatrix;
int u_ModelViewMatrix;
int u_TexUnit;
int u_Rgba;
int u_AmbientSpecularShininess;
int u_AmbientSpecularShininessRadius;
int u_HeadLight;
int u_Fog;
int u_ScreenWidth;
} ShaderInfo;

typedef struct _GlDrawManager
{
bool initDone;
int initCpuCtx;
ShaderInfo shapeShader[16];
ShaderInfo pointSphereShader[4];
} GlDrawManager;

static GlDrawManager mgr;

static
void
GlDraw_generateShapeShader(Text *vertexSrc,
                           Text *fragmentSrc,
                           GlShapeKind kind)
{
Text_append(vertexSrc,
  " #define HAS_NORMALS %d   \n"
  " #define HAS_TEXCOORDS %d \n"
  " #define HAS_COLORS %d    \n"
  " #define HAS_FOG %d       \n",
  kind&GLSHAPE_HAS_NORMALS   ? 1 : 0,
  kind&GLSHAPE_HAS_TEXCOORDS ? 1 : 0,
  kind&GLSHAPE_HAS_COLORS    ? 1 : 0,
  kind&GLSHAPE_HAS_FOG       ? 1 : 0);
Text_appendStr(fragmentSrc,Text_get(vertexSrc));
Text_appendStr(vertexSrc,
  " uniform mat4 u_ProjectionMatrix;                                     \n"
  " uniform mat4 u_ModelViewMatrix;                                      \n"
  " in vec3 v_Vertex;                                                    \n"
  " #if HAS_NORMALS || HAS_FOG                                           \n"
  "   out vec3 f_CamPos;                                                 \n"
  " #endif                                                               \n"
  " #if HAS_NORMALS                                                      \n"
  "   in vec3 v_Normal;                                                  \n"
  "   out vec3 f_Normal;                                                 \n"
  " #endif                                                               \n"
  " #if HAS_TEXCOORDS                                                    \n"
  "   in vec2 v_TexCoord;                                                \n"
  "   out vec2 f_TexCoord;                                               \n"
  " #endif                                                               \n"
  " #if HAS_COLORS                                                       \n"
  "   uniform vec4 u_Rgba;                                               \n"
  "   in vec3 v_Color;                                                   \n"
  "   out vec4 f_Color;                                                  \n"
  " #endif                                                               \n");
Text_appendStr(fragmentSrc,
  " uniform vec3 u_AmbientSpecularShininess;                             \n"
  " #define u_Ambient   (u_AmbientSpecularShininess.x)                   \n"
  " #define u_Specular  (u_AmbientSpecularShininess.y)                   \n"
  " #define u_Shininess (u_AmbientSpecularShininess.z)                   \n"
  " #if HAS_NORMALS || HAS_FOG                                           \n"
  "   in vec3 f_CamPos;                                                  \n"
  " #endif                                                               \n"
  " #if HAS_NORMALS                                                      \n"
  "   uniform vec3 u_HeadLight;                                          \n"
  "   in vec3 f_Normal;                                                  \n"
  " #endif                                                               \n"
  " #if HAS_TEXCOORDS                                                    \n"
  "   in vec2 f_TexCoord;                                                \n"
  "   uniform sampler2D u_TexUnit;                                       \n"
  " #endif                                                               \n"
  " #if HAS_COLORS                                                       \n"
  "   in vec4 f_Color;                                                   \n"
  " #else                                                                \n"
  "   uniform vec4 u_Rgba;                                               \n"
  " # define f_Color u_Rgba                                              \n"
  " #endif                                                               \n"
  " #if HAS_FOG                                                          \n"
  "   uniform vec4 u_Fog;                                                \n"
  " #endif                                                               \n");
Text_appendStr(vertexSrc,
  " void \n"
  " main(void)                                                           \n"
  " {                                                                    \n"
  " vec4 camPos=u_ModelViewMatrix*vec4(v_Vertex,1.0);                    \n"
  " #if HAS_NORMALS || HAS_FOG                                           \n"
  "   f_CamPos=camPos.xyz;                                               \n"
  " #endif                                                               \n"
  " #if HAS_NORMALS                                                      \n"
  "   f_Normal=normalize((u_ModelViewMatrix*vec4(v_Normal,0.0)).xyz);    \n"
  " #endif                                                               \n"
  " #if HAS_TEXCOORDS                                                    \n"
  "   f_TexCoord=v_TexCoord;                                             \n"
  " #endif                                                               \n"
  " #if HAS_COLORS                                                       \n"
  "   f_Color=vec4(v_Color,1.0)*u_Rgba;                                  \n"
  " #endif                                                               \n"
  " gl_Position=u_ProjectionMatrix*camPos;                               \n"
  " }                                                                    \n");
Text_appendStr(fragmentSrc,
  " void \n"
  " main(void)                                                           \n"
  " {                                                                    \n"
  " #if HAS_NORMALS                                                      \n"
  "   vec3 normal=normalize(f_Normal);                                   \n"
  "   vec3 lightDir=normalize(u_HeadLight-f_CamPos);                     \n"
  "   vec3 viewDir=normalize(-f_CamPos);                                 \n"
  "   float diffuse=max(0.0,dot(normal,lightDir));                       \n"
  "   vec3 halfVector=normalize(viewDir+lightDir);                       \n"
  "   float specular=u_Specular*max(0.0,pow(dot(normal,halfVector),      \n"
  "                                         u_Shininess));               \n"
  "   float intensity=max(u_Ambient,diffuse+specular);                   \n"
  "   vec4 fragColor=vec4(f_Color.rgb*intensity,f_Color.a);              \n"
  " #else                                                                \n"
  "   vec4 fragColor=f_Color;                                            \n"
  " #endif                                                               \n"
  " #if HAS_TEXCOORDS                                                    \n"
  "   fragColor=fragColor*texture2D(u_TexUnit,f_TexCoord);               \n"
  " #endif                                                               \n"
  " #if HAS_FOG                                                          \n"
  "   float squareDistance=dot(f_CamPos,f_CamPos);                       \n"
  "   float fogFactor=exp(-u_Fog.a*u_Fog.a*squareDistance);              \n"
  "   gl_FragColor=vec4(mix(u_Fog.rgb,fragColor.rgb,fogFactor),          \n"
  "                     fragColor.a);                                    \n"
  " #else                                                                \n"
  "   gl_FragColor=fragColor;                                            \n"
  " #endif                                                               \n"
  " }                                                                    \n");
}

static
void
GlDraw_generatePointSphereShader(Text *vertexSrc,
                                 Text *geometrySrc,
                                 Text *fragmentSrc,
                                 GlPointSphereKind kind)
{
Text_append(vertexSrc,
  " #define HAS_COLORS %d   \n"
  " #define HAS_RADIUSES %d \n"
  " #define HAS_FOG %d      \n",
  kind&GLPOINTSPHERE_HAS_COLORS   ? 1 : 0,
  kind&GLPOINTSPHERE_HAS_RADIUSES ? 1 : 0,
  kind&GLPOINTSPHERE_HAS_FOG      ? 1 : 0);
Text_appendStr(geometrySrc,Text_get(vertexSrc));
Text_appendStr(fragmentSrc,Text_get(vertexSrc));
Text_appendStr(vertexSrc,
  " in vec3 v_Vertex;                                                    \n"
  " #if HAS_COLORS                                                       \n"
  "   uniform vec4 u_Rgba;                                               \n"
  "   in vec3 v_Color;                                                   \n"
  "   out vec4 g_Color;                                                  \n"
  " #endif                                                               \n"
  " #if HAS_RADIUSES                                                     \n"
  "   uniform vec4 u_AmbientSpecularShininessRadius;                     \n"
  " # define u_Radius (u_AmbientSpecularShininessRadius.w)               \n"
  "   in float v_Radius;                                                 \n"
  "   out float g_Radius;                                                \n"
  " #endif                                                               \n");
Text_appendStr(geometrySrc,
  " #if __VERSION__>=150                                                 \n"
  "   layout(points) in;                                                 \n"
  "   // --> ProgramParameteri(GEOMETRY_INPUT_TYPE)=POINTS               \n"
  "   layout(triangle_strip,max_vertices=4) out;                         \n"
  "   // --> ProgramParameteri(GEOMETRY_OUTPUT_TYPE)=TRIANGLE_STRIP      \n"
  "   // --> ProgramParameteri(GEOMETRY_VERTICES_OUT)=4                  \n"
  " #endif                                                               \n"
  " uniform mat4 u_ProjectionMatrix;                                     \n"
  " uniform mat4 u_ModelViewMatrix;                                      \n"
  " out vec2 f_Offset;                                                   \n"
  " flat out float f_Radius;                                             \n"
  " flat out vec3 f_CamPos;                                              \n"
  " #if HAS_COLORS                                                       \n"
  "   in vec4 g_Color[];                                                 \n"
  "   flat out vec4 f_Color;                                             \n"
  " #endif                                                               \n"
  " #if HAS_RADIUSES                                                     \n"
  "   in float g_Radius[];                                               \n"
  " #else                                                                \n"
  "   uniform vec4 u_AmbientSpecularShininessRadius;                     \n"
  " # define u_Radius (u_AmbientSpecularShininessRadius.w)               \n"
  " #endif                                                               \n");
Text_appendStr(fragmentSrc,
  " uniform vec4 u_AmbientSpecularShininessRadius;                       \n"
  " #define u_Ambient   (u_AmbientSpecularShininessRadius.x)             \n"
  " #define u_Specular  (u_AmbientSpecularShininessRadius.y)             \n"
  " #define u_Shininess (u_AmbientSpecularShininessRadius.z)             \n"
  " uniform mat4 u_ProjectionMatrix;                                     \n"
  " uniform vec3 u_HeadLight;                                            \n"
  " in vec2 f_Offset;                                                    \n"
  " flat in float f_Radius;                                              \n"
  " flat in vec3 f_CamPos;                                               \n"
  " #if HAS_COLORS                                                       \n"
  "   flat in vec4 f_Color;                                              \n"
  " #else                                                                \n"
  "   uniform vec4 u_Rgba;                                               \n"
  " # define f_Color u_Rgba                                              \n"
  " #endif                                                               \n"
  " #if HAS_FOG                                                          \n"
  "   uniform vec4 u_Fog;                                                \n"
  " #endif                                                               \n");
Text_appendStr(vertexSrc,
  " void                                                                 \n"
  " main(void)                                                           \n"
  " {                                                                    \n"
  " #if HAS_COLORS                                                       \n"
  "   g_Color=vec4(v_Color,1.0)*u_Rgba;                                  \n"
  " #endif                                                               \n"
  " #if HAS_RADIUSES                                                     \n"
  "   g_Radius=v_Radius*u_Radius;                                        \n"
  " #endif                                                               \n"
  " gl_Position=vec4(v_Vertex,1.0);                                      \n"
  " }                                                                    \n");
Text_appendStr(geometrySrc,
  " void                                                                 \n"
  " main(void)                                                           \n"
  " {                                                                    \n"
  " #if HAS_COLORS                                                       \n"
  "   f_Color=g_Color[0]; // common to the four vertices                 \n"
  " #endif                                                               \n"
  " #if HAS_RADIUSES                                                     \n"
  "   float radius=g_Radius[0];                                          \n"
  " #else                                                                \n"
  " # define radius u_Radius                                             \n"
  " #endif                                                               \n"
  " vec4 camPos=u_ModelViewMatrix*gl_PosIn(0);                           \n"
  " f_CamPos=camPos.xyz; // common to the four vertices                  \n"
  " f_Radius=radius; // common to the four vertices                      \n"
  " f_Offset=vec2(-1.0,-1.0);                                            \n"
  " gl_Position=u_ProjectionMatrix*                                      \n"
  "             (camPos+vec4(-radius,-radius,0.0,0.0));                  \n"
  " EmitVertex();                                                        \n"
  " f_Offset=vec2(+1.0,-1.0);                                            \n"
  " gl_Position=u_ProjectionMatrix*                                      \n"
  "             (camPos+vec4(+radius,-radius,0.0,0.0));                  \n"
  " EmitVertex();                                                        \n"
  " f_Offset=vec2(-1.0,+1.0);                                            \n"
  " gl_Position=u_ProjectionMatrix*                                      \n"
  "             (camPos+vec4(-radius,+radius,0.0,0.0));                  \n"
  " EmitVertex();                                                        \n"
  " f_Offset=vec2(+1.0,+1.0);                                            \n"
  " gl_Position=u_ProjectionMatrix*                                      \n"
  "             (camPos+vec4(+radius,+radius,0.0,0.0));                  \n"
  " EmitVertex();                                                        \n"
  " EndPrimitive();                                                      \n"
  " }                                                                    \n");
Text_appendStr(fragmentSrc,
  " void                                                                 \n"
  " main(void)                                                           \n"
  " {                                                                    \n"
  " vec3 normal=vec3(f_Offset,dot(f_Offset,f_Offset));                   \n"
  " if(normal.z>1.0) discard;                                            \n"
  " normal.z=sqrt(1.0-normal.z);                                         \n"
  " vec3 camPos=f_CamPos+f_Radius*normal;                                \n"
  " vec3 lightDir=normalize(u_HeadLight-camPos);                         \n"
  " vec3 viewDir=normalize(-camPos);                                     \n"
  " float diffuse=max(0.0,dot(normal,lightDir));                         \n"
  " vec3 halfVector=normalize(viewDir+lightDir);                         \n"
  " float specular=u_Specular*max(0.0,pow(dot(normal,halfVector),        \n"
  "                                       u_Shininess));                 \n"
  " float intensity=max(u_Ambient,diffuse+specular);                     \n"
  " #if HAS_FOG                                                          \n"
  "   float squareDistance=dot(f_CamPos,f_CamPos);                       \n"
  "   float fogFactor=exp(-u_Fog.a*u_Fog.a*squareDistance);              \n"
  "   gl_FragColor=vec4(mix(u_Fog.rgb,f_Color.rgb*intensity,fogFactor),  \n"
  "                     f_Color.a);                                      \n"
  " #else                                                                \n"
  "   gl_FragColor=vec4(f_Color.rgb*intensity,f_Color.a);                \n"
  " #endif                                                               \n"
  " #if 1                                                                \n"
  "   vec4 clipPos=u_ProjectionMatrix*vec4(camPos,1.0);                  \n"
  "   gl_FragDepth=((gl_DepthRange.diff*(clipPos.z/clipPos.w))+          \n"
  "                 gl_DepthRange.near+gl_DepthRange.far)*0.5;           \n"
  " #endif                                                               \n"
  " }                                                                    \n");
}

static
void
GlDraw_generatePointSphereShaderFallback(Text *vertexSrc,
                                         Text *fragmentSrc,
                                         GlPointSphereKind kind)
{
Text_append(vertexSrc,
  " #define HAS_COLORS %d   \n"
  " #define HAS_RADIUSES %d \n"
  " #define HAS_FOG %d      \n",
  kind&GLPOINTSPHERE_HAS_COLORS   ? 1 : 0,
  kind&GLPOINTSPHERE_HAS_RADIUSES ? 1 : 0,
  kind&GLPOINTSPHERE_HAS_FOG      ? 1 : 0);
Text_appendStr(fragmentSrc,Text_get(vertexSrc));
Text_appendStr(vertexSrc,
  " uniform mat4 u_ProjectionMatrix;                                     \n"
  " uniform mat4 u_ModelViewMatrix;                                      \n"
  " uniform vec4 u_AmbientSpecularShininessRadius;                       \n"
  " #define u_Radius (u_AmbientSpecularShininessRadius.w)                \n"
  " uniform float u_ScreenWidth;                                         \n"
  " in vec3 v_Vertex;                                                    \n"
  " out vec3 f_CamPos;                                                   \n"
  " #if HAS_COLORS                                                       \n"
  "   uniform vec4 u_Rgba;                                               \n"
  "   in vec3 v_Color;                                                   \n"
  "   out vec4 f_Color;                                                  \n"
  " #endif                                                               \n"
  " #if HAS_RADIUSES                                                     \n"
  "   in float v_Radius;                                                 \n"
  "   out float f_Radius;                                                \n"
  " #endif                                                               \n");
Text_appendStr(fragmentSrc,
  " uniform vec4 u_AmbientSpecularShininessRadius;                       \n"
  " #define u_Ambient   (u_AmbientSpecularShininessRadius.x)             \n"
  " #define u_Specular  (u_AmbientSpecularShininessRadius.y)             \n"
  " #define u_Shininess (u_AmbientSpecularShininessRadius.z)             \n"
  " uniform mat4 u_ProjectionMatrix;                                     \n"
  " uniform vec3 u_HeadLight;                                            \n"
  " uniform vec4 u_Fog;                                                  \n"
  " in vec3 f_CamPos;                                                    \n"
  " #if HAS_COLORS                                                       \n"
  "   in vec4 f_Color;                                                   \n"
  " #else                                                                \n"
  "   uniform vec4 u_Rgba;                                               \n"
  " # define f_Color u_Rgba                                              \n"
  " #endif                                                               \n"
  " #if HAS_RADIUSES                                                     \n"
  "   in float f_Radius;                                                 \n"
  " #else                                                                \n"
  " # define f_Radius (u_AmbientSpecularShininessRadius.w)               \n"
  " #endif                                                               \n");
Text_appendStr(vertexSrc,
  " void                                                                 \n"
  " main(void)                                                           \n"
  " {                                                                    \n"
  " #if HAS_COLORS                                                       \n"
  "   f_Color=vec4(v_Color,1.0)*u_Rgba;                                  \n"
  " #endif                                                               \n"
  " #if HAS_RADIUSES                                                     \n"
  "   float radius=v_Radius*u_Radius;                                    \n"
  "   f_Radius=radius;                                                   \n"
  " #else                                                                \n"
  " # define radius u_Radius                                             \n"
  " #endif                                                               \n"
  " vec4 camPos=u_ModelViewMatrix*vec4(v_Vertex,1.0);                    \n"
  " gl_Position=u_ProjectionMatrix*camPos;                               \n"
  " gl_PointSize=-radius*u_ScreenWidth*                                  \n"
  "              u_ProjectionMatrix[0].x/camPos.z;                       \n"
  " f_CamPos=camPos.xyz;                                                 \n"
  " }                                                                    \n");
Text_appendStr(fragmentSrc,
  " void                                                                 \n"
  " main(void)                                                           \n"
  " {                                                                    \n"
  " vec3 normal;                                                         \n"
  " // GL_POINT_SPRITE_COORD_ORIGIN=GL_UPPER_LEFT                        \n"
  " normal.x=gl_PointCoord.s*2.0-1.0;                                    \n"
  " normal.y=1.0-gl_PointCoord.t*2.0;                                    \n"
  " normal.z=dot(normal.xy,normal.xy);                                   \n"
  " if(normal.z>1.0) discard;                                            \n"
  " normal.z=sqrt(1.0-normal.z);                                         \n"
  " vec3 camPos=f_CamPos+f_Radius*normal;                                \n"
  " vec3 lightDir=normalize(u_HeadLight-camPos);                         \n"
  " vec3 viewDir=normalize(-camPos);                                     \n"
  " float diffuse=max(0.0,dot(normal,lightDir));                         \n"
  " vec3 halfVector=normalize(viewDir+lightDir);                         \n"
  " float specular=u_Specular*max(0.0,pow(dot(normal,halfVector),        \n"
  "                                       u_Shininess));                 \n"
  " float intensity=max(u_Ambient,diffuse+specular);                     \n"
  " #if HAS_FOG                                                          \n"
  "   float squareDistance=dot(f_CamPos,f_CamPos);                       \n"
  "   float fogFactor=exp(-u_Fog.a*u_Fog.a*squareDistance);              \n"
  "   gl_FragColor=vec4(mix(u_Fog.rgb,f_Color.rgb*intensity,fogFactor),  \n"
  "                     f_Color.a);                                      \n"
  " #else                                                                \n"
  "   gl_FragColor=vec4(f_Color.rgb*intensity,f_Color.a);                \n"
  " #endif                                                               \n"
  " #if 1                                                                \n"
  "   vec4 clipPos=u_ProjectionMatrix*vec4(camPos,1.0);                  \n"
  "   gl_FragDepth=((gl_DepthRange.diff*(clipPos.z/clipPos.w))+          \n"
  "                 gl_DepthRange.near+gl_DepthRange.far)*0.5;           \n"
  " #endif                                                               \n"
  " }                                                                    \n");
}

static
void
GlDraw_buildShader(ShaderInfo *si)
{
int program=0;
if(Text_length(&si->geometrySource)&&!CoreGl_hasGeometryShaders())
  {
  memset(&si->shader,0,sizeof(si->shader));
  }
else
  {
  GlShader_init(&si->shader,
                Text_get(&si->vertexSource),
                Text_get(&si->geometrySource),
                Text_get(&si->fragmentSource),true);
  program=GlShader_program(&si->shader);
  if(!program)
    {
    const char *log=GlShader_log(&si->shader);
    if(log)
      {
      Platform_msg("%s",log);
      }
    }
  }
if(program)
  {
  si->v_Vertex=glGetAttribLocation(program,"v_Vertex");
  si->v_Normal=glGetAttribLocation(program,"v_Normal");
  si->v_TexCoord=glGetAttribLocation(program,"v_TexCoord");
  si->v_Color=glGetAttribLocation(program,"v_Color");
  si->v_Radius=glGetAttribLocation(program,"v_Radius");
  si->u_ProjectionMatrix=glGetUniformLocation(program,"u_ProjectionMatrix");
  si->u_ModelViewMatrix=glGetUniformLocation(program,"u_ModelViewMatrix");
  si->u_TexUnit=glGetUniformLocation(program,"u_TexUnit");
  si->u_Rgba=glGetUniformLocation(program,"u_Rgba");
  si->u_AmbientSpecularShininess=glGetUniformLocation(program,
                                 "u_AmbientSpecularShininess");
  si->u_AmbientSpecularShininessRadius=glGetUniformLocation(program,
                                       "u_AmbientSpecularShininessRadius");
  si->u_HeadLight=glGetUniformLocation(program,"u_HeadLight");
  si->u_Fog=glGetUniformLocation(program,"u_Fog");
  si->u_ScreenWidth=glGetUniformLocation(program,"u_ScreenWidth");
  }
else
  {
  si->v_Vertex=-1;
  si->v_Normal=-1;
  si->v_TexCoord=-1;
  si->v_Color=-1;
  si->v_Radius=-1;
  si->u_ProjectionMatrix=-1;
  si->u_ModelViewMatrix=-1;
  si->u_TexUnit=-1;
  si->u_Rgba=-1;
  si->u_AmbientSpecularShininess=-1;
  si->u_AmbientSpecularShininessRadius=-1;
  si->u_HeadLight=-1;
  si->u_Fog=-1;
  si->u_ScreenWidth=-1;
  }
}

void
GlDraw_init(void)
{
if(!mgr.initDone)
  {
  mgr.initDone=true;
  mgr.initCpuCtx=Platform_cpuCtx();
  for(int i=0;i<16;++i)
    {
    ShaderInfo *si=mgr.shapeShader+i;
    Text_init(&si->vertexSource);
    Text_init(&si->geometrySource);
    Text_init(&si->fragmentSource);
    GlDraw_generateShapeShader(&si->vertexSource,
                               &si->fragmentSource,
                               (GlShapeKind)i);
    GlDraw_buildShader(si);
    }
  for(int i=0;i<4;++i)
    {
    ShaderInfo *si=mgr.pointSphereShader+i;
    Text_init(&si->vertexSource);
    Text_init(&si->geometrySource);
    Text_init(&si->fragmentSource);
    if(CoreGl_hasGeometryShaders())
      {
      GlDraw_generatePointSphereShader(&si->vertexSource,
                                       &si->geometrySource,
                                       &si->fragmentSource,
                                       (GlPointSphereKind)i);
      }
    else
      {
      GlDraw_generatePointSphereShaderFallback(&si->vertexSource,
                                               &si->fragmentSource,
                                               (GlPointSphereKind)i);
      }
    GlDraw_buildShader(si);
    }
  }
}

static
int // size rounded up to 2^N
GlDraw_roundSize(int size,
                 bool up)
{
int newSize=1;
while((newSize<<1)<=size) // (newSize=2^N) <= size
  {
  newSize<<=1;
  }
if(up&&(newSize<size)) // (newSize=2^N) >= size
  {
  newSize<<=1;
  }
return newSize;
}

static
void
GlDraw_downScale(int inWidth,
                 int inHeight,
                 int components,
                 const uint8_t *inData,
                 int outWidth,
                 int outHeight,
                 uint8_t *outData)
{
real64_t sx=inWidth/(real64_t)outWidth;
real64_t sy=inHeight/(real64_t)outHeight;
for(int y=0;y<outHeight;++y)
  {
  for(int x=0;x<outWidth;++x)
    {
    int outIndex=(x+(outWidth*y))*components;
    int x1=(int)(sx*x);
    int x2=(int)(sx*(x+1));
    if(x2==x1)
      {
      ++x2;
      }
    if(x2>inWidth)
      {
      x2=inWidth;
      }
    int y1=(int)(sy*y);
    int y2=(int)(sy*(y+1));
    if(y2==y1)
      {
      ++y2;
      }
    if(y2>inHeight)
      {
      y2=inHeight;
      }
    real64_t comp[4]={0.0,0.0,0.0,0.0};
    for(int inY=y1;inY<y2;++inY)
      {
      for(int inX=x1;inX<x2;++inX)
        {
        int inIndex=(inX+(inWidth*inY))*components;
        for(int c=0;c<components;++c)
          {
          comp[c]+=inData[inIndex+c];
          }
        }
      }
    real64_t div=1.0/((y2-y1)*(x2-x1));
    for(int c=0;c<components;++c)
      {
      outData[outIndex+c]=(uint8_t)(comp[c]*div);
      }
    }
  }
}

static
void
GlDraw_upScale(int inWidth,
               int inHeight,
               int components,
               const uint8_t *inData,
               int outWidth,
               int outHeight,
               uint8_t *outData)
{
real64_t sx=inWidth/(real64_t)outWidth;
real64_t sy=inHeight/(real64_t)outHeight;
for(int y=0;y<outHeight;++y)
  {
  for(int x=0;x<outWidth;++x)
    {
    int outIndex=(x+(outWidth*y))*components;
    int inX=(int)(sx*x);
    int inY=(int)(sy*y);
    int inIndex=(inX+(inWidth*inY))*components;
    for(int c=0;c<components;++c)
      {
      outData[outIndex+c]=inData[inIndex+c];
      }
    }
  }
}

static
void
GlDraw_makeTexture(int textureTarget,
                   int width,
                   int height,
                   int components,
                   const uint8_t *storage,
                   bool roundUp)
{
int format,internalFormat;
switch(components)
  {
#if 0 // FIXME: luminance is not supported anymore
  case 1: // grey
    {
    format=GL_LUMINANCE;
    internalFormat=GL_COMPRESSED_LUMINANCE;
    break;
    }
  case 2: // grey-alpha
    {
    format=GL_LUMINANCE_ALPHA;
    internalFormat=GL_COMPRESSED_LUMINANCE_ALPHA;
    break;
    }
#else
  case 1: // grey
    {
    format=GL_RED;
    internalFormat=GL_COMPRESSED_RED;
    break;
    }
  case 2: // grey-alpha
    {
    format=GL_RG;
    internalFormat=GL_COMPRESSED_RG;
    break;
    }
#endif
  case 3: // RGB
    {
    format=GL_RGB;
    internalFormat=GL_COMPRESSED_RGB;
    break;
    }
  case 4: // RGB-alpha
    {
    format=GL_RGBA;
    internalFormat=GL_COMPRESSED_RGBA;
    break;
    }
  default:
    {
    format=internalFormat=0;
    DROP_DEAD("bad number of texture components");
    }
  }
int newWidth=GlDraw_roundSize(width,roundUp);
int newHeight=GlDraw_roundSize(height,roundUp);
if((width!=newWidth)||(height!=newHeight))
  {
  uint8_t *data=Memory_alloc(newWidth*newHeight*components,uint8_t);
  if(roundUp)
    {
    GlDraw_upScale(width,height,components,storage,
                   newWidth,newHeight,data);
    }
  else
    {
    GlDraw_downScale(width,height,components,storage,
                     newWidth,newHeight,data);
    }
  glTexImage2D(textureTarget,0,internalFormat,newWidth,newHeight,
               0,format,GL_UNSIGNED_BYTE,data);
  OPENGL_CHECK;
  Memory_free(data);
  }
else
  {
  glTexImage2D(textureTarget,0,internalFormat,width,height,
               0,format,GL_UNSIGNED_BYTE,storage);
  OPENGL_CHECK;
  }
}

int // texture id
GlDraw_loadTexture(int width,
                   int height,
                   int components,
                   const uint8_t *storage,
                   bool roundUp)
{
int textureId;
glGenTextures(1,(unsigned int *)&textureId);
if(!textureId)
  {
  DROP_DEAD("glGenTextures() failure");
  }
glBindTexture(GL_TEXTURE_2D,textureId);
OPENGL_CHECK;
GlDraw_makeTexture(GL_TEXTURE_2D,width,height,components,storage,roundUp);
int minFilter=GL_LINEAR;
if(glGenerateMipmap)
  {
  glGenerateMipmap(GL_TEXTURE_2D);
  OPENGL_CHECK;
  minFilter=GL_LINEAR_MIPMAP_LINEAR;
  }
glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,minFilter);
OPENGL_CHECK;
glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
OPENGL_CHECK;
glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
OPENGL_CHECK;
glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
OPENGL_CHECK;
glBindTexture(GL_TEXTURE_2D,0);
OPENGL_CHECK;
return textureId;
}

int // cube-map texture id
GlDraw_loadCubeMapTextures(int size,
                           int components,
                           const uint8_t *storage[6],
                           bool roundUp)
{
int textureId;
glGenTextures(1,(unsigned int *)&textureId);
if(!textureId)
  {
  DROP_DEAD("glGenTextures() failure");
  }
glBindTexture(GL_TEXTURE_CUBE_MAP,textureId);
OPENGL_CHECK;
for(int i=0;i<6;++i)
  {
  GlDraw_makeTexture(GL_TEXTURE_CUBE_MAP_POSITIVE_X+i,
                     size,size,components,storage[i],roundUp);
  }
int minFilter=GL_LINEAR;
if(glGenerateMipmap)
  {
  glGenerateMipmap(GL_TEXTURE_CUBE_MAP);
  OPENGL_CHECK;
  minFilter=GL_LINEAR_MIPMAP_LINEAR;
  }
glTexParameteri(GL_TEXTURE_CUBE_MAP,GL_TEXTURE_MIN_FILTER,minFilter);
OPENGL_CHECK;
glTexParameteri(GL_TEXTURE_CUBE_MAP,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
OPENGL_CHECK;
glTexParameteri(GL_TEXTURE_CUBE_MAP,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
OPENGL_CHECK;
glTexParameteri(GL_TEXTURE_CUBE_MAP,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
OPENGL_CHECK;
glBindTexture(GL_TEXTURE_CUBE_MAP,0);
OPENGL_CHECK;
return textureId;
}

void
GlDraw_releaseTexture(int textureId)
{
glDeleteTextures(1,(const unsigned int *)&textureId);
OPENGL_CHECK;
}

void
GlDraw_snapshot(int x,
                int y,
                int width,
                int height,
                uint8_t *rgbStorage)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
glReadBuffer(GL_FRONT);
OPENGL_CHECK;
glPixelStorei(GL_UNPACK_SWAP_BYTES,GL_FALSE);
OPENGL_CHECK;
glPixelStorei(GL_UNPACK_LSB_FIRST,GL_FALSE);
OPENGL_CHECK;
glPixelStorei(GL_UNPACK_ROW_LENGTH,0);
OPENGL_CHECK;
glPixelStorei(GL_UNPACK_SKIP_ROWS,0);
OPENGL_CHECK;
glPixelStorei(GL_UNPACK_SKIP_PIXELS,0);
OPENGL_CHECK;
glPixelStorei(GL_UNPACK_ALIGNMENT,1);
OPENGL_CHECK;
glReadPixels(x,y,width,height,GL_RGB,GL_UNSIGNED_BYTE,rgbStorage);
OPENGL_CHECK;
}

void
GlVertexArray_init(GlVertexArray *vertexArray)
{
vertexArray->vao=0;
if(CoreGl_has_ARB_vertex_array_object())
  {
  glGenVertexArrays(1,(unsigned int *)&vertexArray->vao);
  if(!vertexArray->vao)
    {
    DROP_DEAD("glGenVertexArrays() failure");
    }
  }
IPtrVector_init(&vertexArray->info);
}

void
GlVertexArray_destroy(GlVertexArray *vertexArray)
{
IPtrVector_destroy(&vertexArray->info);
if(vertexArray->vao)
  {
  glDeleteVertexArrays(1,(const unsigned int *)&vertexArray->vao);
  OPENGL_CHECK;
  }
}

static
void
GlVertexArray_bindBuffer(const GlVertexArray *vertexArray,
                         int bufferIndex)
{
int attribute=GlVertexArray_attribute(vertexArray,bufferIndex);
int buffer=GlVertexArray_buffer(vertexArray,bufferIndex);
if(attribute!=-1)
  {
  glEnableVertexAttribArray(attribute);
  OPENGL_CHECK;
  glBindBuffer(GL_ARRAY_BUFFER,buffer);
  OPENGL_CHECK;
  int components=GlVertexArray_components(vertexArray,bufferIndex);
  int glType=GlVertexArray_glType(vertexArray,bufferIndex);
  int byteStride=GlVertexArray_byteStride(vertexArray,bufferIndex);
  intptr_t byteOffset=GlVertexArray_byteOffset(vertexArray,bufferIndex);
  glVertexAttribPointer(attribute,components,glType,GL_FALSE,byteStride,
                        (const void *)byteOffset);
  OPENGL_CHECK;
  int divisor=GlVertexArray_divisor(vertexArray,bufferIndex);
  if(divisor)
    {
    glVertexBindingDivisor(attribute,divisor);
    OPENGL_CHECK;
    }
  }
else
  {
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,buffer);
  OPENGL_CHECK;
  }
}

static
void
GlVertexArray_unbindBuffer(const GlVertexArray *vertexArray,
                           int bufferIndex)
{
int attribute=GlVertexArray_attribute(vertexArray,bufferIndex);
if(attribute!=-1)
  {
  int divisor=GlVertexArray_divisor(vertexArray,bufferIndex);
  if(divisor)
    {
    glVertexBindingDivisor(attribute,0);
    OPENGL_CHECK;
    }
  glDisableVertexAttribArray(attribute);
  OPENGL_CHECK;
  glBindBuffer(GL_ARRAY_BUFFER,0);
  OPENGL_CHECK;
  }
else
  {
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,0);
  OPENGL_CHECK;
  }
}

int // buffer index
GlVertexArray_manageBuffer(GlVertexArray *vertexArray,
                           int attribute,
                           int components,
                           int glType,
                           int byteStride,
                           int divisor,
                           int buffer,
                           intptr_t byteOffset)
{
int bufferIndex=GlVertexArray_count(vertexArray);
IPtrVector_push(&vertexArray->info,attribute);
IPtrVector_push(&vertexArray->info,components);
IPtrVector_push(&vertexArray->info,glType);
IPtrVector_push(&vertexArray->info,byteStride);
IPtrVector_push(&vertexArray->info,divisor);
IPtrVector_push(&vertexArray->info,buffer);
IPtrVector_push(&vertexArray->info,byteOffset);
if(vertexArray->vao)
  {
  glBindVertexArray(vertexArray->vao);
  OPENGL_CHECK;
  GlVertexArray_bindBuffer(vertexArray,bufferIndex);
  glBindVertexArray(0);
  OPENGL_CHECK;
  }
return bufferIndex;
}

int // buffer index
GlVertexArray_createBufferGeneric(GlVertexArray *vertexArray,
                                  int attribute,
                                  int components,
                                  int divisor,
                                  intptr_t count,
                                  const void *data,
                                  int sizeOfData)
{
int buffer;
glGenBuffers(1,(unsigned int *)&buffer);
if(!buffer)
  {
  DROP_DEAD("glGenBuffers() failure");
  }
int glType;
int bufferKind;
intptr_t bufferSize;
if(attribute!=-1)
  {
  REAL_GL_TYPE(glType,sizeOfData)
  bufferKind=GL_ARRAY_BUFFER;
  bufferSize=components*count*sizeOfData;
  }
else
  {
  INTEGER_GL_TYPE(glType,sizeOfData)
  bufferKind=GL_ELEMENT_ARRAY_BUFFER;
  bufferSize=count*sizeOfData;
  }
glBindBuffer(bufferKind,buffer);
OPENGL_CHECK;
glBufferData(bufferKind,bufferSize,data,
             data ? GL_STATIC_DRAW : GL_STREAM_DRAW);
OPENGL_CHECK;
return GlVertexArray_manageBuffer(vertexArray,attribute,components,
                                  glType,0,divisor,buffer,0);
}

void
GlVertexArray_bind(const GlVertexArray *vertexArray)
{
if(vertexArray->vao)
  {
  glBindVertexArray(vertexArray->vao);
  OPENGL_CHECK;
  }
else
  {
  for(int i=0,i_end=GlVertexArray_count(vertexArray);i<i_end;++i)
    {
    GlVertexArray_bindBuffer(vertexArray,i);
    }
  }
}

void
GlVertexArray_unbind(const GlVertexArray *vertexArray)
{
if(vertexArray->vao)
  {
  glBindVertexArray(0);
  OPENGL_CHECK;
  }
else
  {
  for(int i=0,i_end=GlVertexArray_count(vertexArray);i<i_end;++i)
    {
    GlVertexArray_unbindBuffer(vertexArray,i);
    }
  }
}

void
GlShader_init(GlShader *shader,
              const char *vertexSource,
              const char *geometrySource,
              const char *fragmentSource,
              bool addPreamble)
{
shader->vertex=0;
shader->geometry=0;
shader->fragment=0;
shader->program=0;
Text_init(&shader->log);
bool tryToLink=true;
char glslVersion[0x100];
sprintf(glslVersion," #version %d \n",CoreGl_shadingLanguageVersion());
struct
  {
  int *shader;
  int kind;
  const char *kindStr;
  const char *src;
  const char *preamble;
  } iter[3]=
  { { &shader->vertex, GL_VERTEX_SHADER, "VERTEX", vertexSource,
      " #if __VERSION__<130 \n"
      " # define in attribute \n"
      " # define out varying \n"
      " # define smooth \n"
      " # define flat \n"
      " # define noperspective \n"
      " #endif \n" },
    { &shader->geometry, GL_GEOMETRY_SHADER, "GEOMETRY", geometrySource,
      " #if __VERSION__<130 \n"
      " # define in varying in \n"
      " # define out varying out \n"
      " # define smooth \n"
      " # define flat \n"
      " # define noperspective \n"
      " #endif \n"
      " #if __VERSION__>=150 \n"
      " # define gl_PosIn(i) (gl_in[i].gl_Position) \n"
      " #else \n"
      " # extension GL_EXT_geometry_shader4 : enable \n"
      " # define gl_PosIn(i) (gl_PositionIn[i]) \n"
      " #endif \n" },
    { &shader->fragment, GL_FRAGMENT_SHADER, "FRAGMENT", fragmentSource,
      " #if __VERSION__<130 \n"
      " # define in varying \n"
      " # define smooth \n"
      " # define flat \n"
      " # define noperspective \n"
      " #else \n"
      " # define texture2D texture \n"
      " # define shadow2D texture  \n"
      " # define texture3D texture \n"
      " # define textureCube texture \n"
      " # if __VERSION__>=330 \n"
      "     layout(location=0) \n"
      " # endif \n"
      "   out vec4 defaultFragColor; \n"
      " # define gl_FragColor defaultFragColor \n"
      " #endif \n" } };
for(int i=0;i<3;++i)
  {
  if(iter[i].src&&*iter[i].src)
    {
    const char *strings[3]={ "", "", iter[i].src };
    if(addPreamble)
      {
      strings[0]=glslVersion;
      strings[1]=iter[i].preamble;
      }
    *iter[i].shader=glCreateShader(iter[i].kind);
    if(!*iter[i].shader)
      {
      Text_append(&shader->log,
                  "glCreateShader(GL_%s_SHADER) failure: %s\n",
                  iter[i].kindStr,
                  CoreGl_lastError());
      tryToLink=false;
      }
    else
      {
      glShaderSource(*iter[i].shader,3,strings,(const int *)0);
      OPENGL_CHECK;
      glCompileShader(*iter[i].shader);
      OPENGL_CHECK;
      bool compilationError;
      char *log=CoreGl_shaderStatus(*iter[i].shader,&compilationError);
      if(log)
        {
        Text_append(&shader->log,
                    "%s SHADER SOURCE:\n%s%s%s\n",
                    iter[i].kindStr,strings[0],strings[1],strings[2]);
        Text_append(&shader->log,
                    "%s SHADER LOG:\n%s\n",iter[i].kindStr,log);
        Memory_free(log);
        }
      if(compilationError)
        {
        glDeleteShader(*iter[i].shader);
        OPENGL_CHECK;
        *iter[i].shader=0;
        tryToLink=false;
        }
      }
    }
  }
if(tryToLink)
  {
  shader->program=glCreateProgram();
  if(!shader->program)
    {
    Text_append(&shader->log,
                "glCreateProgram() failure: %s\n",
                CoreGl_lastError());
    }
  else
    {
    if(shader->vertex)
      {
      glAttachShader(shader->program,shader->vertex);
      OPENGL_CHECK;
      }
    if(shader->geometry)
      {
      glAttachShader(shader->program,shader->geometry);
      OPENGL_CHECK;
      if(CoreGl_shadingLanguageVersion()<150)
        {
        // ugly workaround: parse source to detect layout parameters
        int inputType=-1;
        int outputType=-1;
        int verticesOut=-1;
        char line[0x400];
        const char *begin=geometrySource;
        const char *end=geometrySource+strlen(geometrySource);
        while(begin<end)
          {
          const char *cr=strchr(begin,'\n');
          if(!cr) cr=end;
          int len=(int)(cr-begin);
          memcpy(line,begin,len);
          line[len]='\0';
          char txt[0x100];
          int value;
          int charCount=0;
          if((sscanf(line," layout ( %[^)] ) in ; %n",
                     txt,&charCount)==1)&&charCount)
            {
            if(!strcmp(txt,"points"))
              {
              inputType=GL_POINTS;
              }
            else if(!strcmp(txt,"lines"))
              {
              inputType=GL_LINES;
              }
            else if(!strcmp(txt,"lines_adjacency"))
              {
              inputType=GL_LINES_ADJACENCY;
              }
            else if(!strcmp(txt,"triangles"))
              {
              inputType=GL_TRIANGLES;
              }
            else if(!strcmp(txt,"triangles_adjacency"))
              {
              inputType=GL_TRIANGLES_ADJACENCY;
              }
            }
          charCount=0;
          if((sscanf(line," layout ( %[^),] , max_vertices = %d ) out ; %n",
                     txt,&value,&charCount)==2)&&charCount)
            {
            if(!strcmp(txt,"points"))
              {
              outputType=GL_POINTS;
              }
            else if(!strcmp(txt,"line_strip"))
              {
              outputType=GL_LINE_STRIP;
              }
            else if(!strcmp(txt,"triangle_strip"))
              {
              outputType=GL_TRIANGLE_STRIP;
              }
            verticesOut=value;
            }
          begin=cr+1;
          }
        // use GL_EXT_geometry_shader4 layout parameters
        if(inputType>=0)
          {
          glProgramParameteri(shader->program,
                              0x8DDB, // GL_GEOMETRY_INPUT_TYPE_EXT
                              inputType);
          OPENGL_CHECK;
          }
        if(outputType>=0)
          {
          glProgramParameteri(shader->program,
                              0x8DDC, // GL_GEOMETRY_OUTPUT_TYPE_EXT
                              outputType);
          OPENGL_CHECK;
          }
        if(verticesOut>0)
          {
          glProgramParameteri(shader->program,
                              0x8DDA, // GL_GEOMETRY_VERTICES_OUT_EXT
                              verticesOut);
          OPENGL_CHECK;
          }
        }
      }
    if(shader->fragment)
      {
      glAttachShader(shader->program,shader->fragment);
      OPENGL_CHECK;
      }
    glLinkProgram(shader->program);
    OPENGL_CHECK;
    bool linkError;
    char *log=CoreGl_shaderStatus(shader->program,&linkError);
    if(log)
      {
      Text_append(&shader->log,
                  "SHADER PROGRAM LOG:\n%s\n",log);
      Memory_free(log);
      }
    if(linkError)
      {
      glDeleteProgram(shader->program);
      OPENGL_CHECK;
      shader->program=0;
      }
    }
  }
}

void
GlShader_destroy(GlShader *shader)
{
Text_destroy(&shader->log);
if(shader->program)
  {
  glDeleteProgram(shader->program);
  OPENGL_CHECK;
  }
if(shader->fragment)
  {
  glDeleteShader(shader->fragment);
  OPENGL_CHECK;
  }
if(shader->geometry)
  {
  glDeleteShader(shader->geometry);
  OPENGL_CHECK;
  }
if(shader->vertex)
  {
  glDeleteShader(shader->vertex);
  OPENGL_CHECK;
  }
}

void
GlShape_initGeneric(GlShape *shape,
                    GlShapeKind kind,
                    int glPrimitive,
                    intptr_t indexCount,
                    const void *indices,
                    int sizeOfIndex,
                    intptr_t vertexCount,
                    const void *vertices,
                    int sizeOfVertex,
                    const void *normals,
                    int sizeOfNormal,
                    const void *texCoords,
                    int sizeOfTexCoord,
                    const void *colors,
                    int sizeOfColor)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
shape->kind=kind;
shape->glPrimitive=glPrimitive;
shape->indexCount=indexCount;
shape->vertexCount=vertexCount;
GlVertexArray_init(&shape->vertexArray);
const ShaderInfo *si=mgr.shapeShader+shape->kind;
if(shape->kind&GLSHAPE_HAS_COLORS)
  {
  shape->colors=GlVertexArray_createBufferGeneric(&shape->vertexArray,
                                                  si->v_Color,3,0,
                                                  shape->vertexCount,
                                                  colors,
                                                  sizeOfColor);
  }
else
  {
  shape->colors=-1;
  }
if(shape->kind&GLSHAPE_HAS_TEXCOORDS)
  {
  shape->texCoords=GlVertexArray_createBufferGeneric(&shape->vertexArray,
                                                     si->v_TexCoord,2,0,
                                                     shape->vertexCount,
                                                     texCoords,
                                                     sizeOfTexCoord);
  }
else
  {
  shape->texCoords=-1;
  }
if(shape->kind&GLSHAPE_HAS_NORMALS)
  {
  shape->normals=GlVertexArray_createBufferGeneric(&shape->vertexArray,
                                                   si->v_Normal,3,0,
                                                   shape->vertexCount,
                                                   normals,
                                                   sizeOfNormal);
  }
else
  {
  shape->normals=-1;
  }
shape->vertices=GlVertexArray_createBufferGeneric(&shape->vertexArray,
                                                  si->v_Vertex,3,0,
                                                  shape->vertexCount,
                                                  vertices,
                                                  sizeOfVertex);
if(shape->indexCount)
  {
  shape->indices=GlVertexArray_createBufferGeneric(&shape->vertexArray,
                                                   -1,-1,0,
                                                   shape->indexCount,
                                                   indices,
                                                   sizeOfIndex);
  }
else
  {
  shape->indices=-1;
  }
}

void
GlShape_destroy(GlShape *shape)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
for(int i=0,i_end=GlVertexArray_count(&shape->vertexArray);i<i_end;++i)
  {
  int buffer=GlVertexArray_buffer(&shape->vertexArray,i);
  glDeleteBuffers(1,(const unsigned int *)&buffer);
  OPENGL_CHECK;
  }
GlVertexArray_destroy(&shape->vertexArray);
}

void
GlShape_draw(const GlShape *shape,
             const real32_t rgba[4],
             real32_t ambient,
             real32_t specular,
             real32_t shininess,
             const real32_t headLight[3],
             const real32_t fog[4],
             int textureId,
             const real32_t *projectionMatrix,
             const real32_t *modelViewMatrix)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
const ShaderInfo *si=mgr.shapeShader+
                     shape->kind+(fog ? GLSHAPE_HAS_FOG : 0);
glUseProgram(GlShader_program(&si->shader));
OPENGL_CHECK;
glUniformMatrix4fv(si->u_ProjectionMatrix,1,GL_FALSE,
                   (real32_t *)projectionMatrix);
OPENGL_CHECK;
glUniformMatrix4fv(si->u_ModelViewMatrix,1,GL_FALSE,
                   (real32_t *)modelViewMatrix);
OPENGL_CHECK;
glUniform4fv(si->u_Rgba,1,(real32_t *)rgba);
OPENGL_CHECK;
glUniform3f(si->u_AmbientSpecularShininess,ambient,specular,shininess);
OPENGL_CHECK;
if(si->u_HeadLight!=-1)
  {
  glUniform3fv(si->u_HeadLight,1,(real32_t *)headLight);
  OPENGL_CHECK;
  }
if(si->u_Fog!=-1)
  {
  glUniform4fv(si->u_Fog,1,(real32_t *)fog);
  OPENGL_CHECK;
  }
if(si->u_TexUnit!=-1)
  {
  glUniform1i(si->u_TexUnit,0); // GL_TEXTURE0
  OPENGL_CHECK;
  glActiveTexture(GL_TEXTURE0);
  OPENGL_CHECK;
  glBindTexture(GL_TEXTURE_2D,textureId);
  OPENGL_CHECK;
  }
GlVertexArray_bind(&shape->vertexArray);
if(shape->indexCount)
  {
  int indexType=GlVertexArray_glType(&shape->vertexArray,shape->indices);
  if(glDrawRangeElements)
    {
    glDrawRangeElements(shape->glPrimitive,0,(int)shape->vertexCount-1,
                        (int)shape->indexCount,indexType,
                        (const void *)0);
    OPENGL_CHECK;
    }
  else
    {
    glDrawElements(shape->glPrimitive,(int)shape->indexCount,indexType,
                   (const void *)0);
    OPENGL_CHECK;
    }
  }
else
  {
  glDrawArrays(shape->glPrimitive,0,(int)shape->vertexCount);
  OPENGL_CHECK;
  }
GlVertexArray_unbind(&shape->vertexArray);
glUseProgram(0);
OPENGL_CHECK;
}

void
GlPointSphere_initGeneric(GlPointSphere *pointSphere,
                          GlPointSphereKind kind,
                          intptr_t vertexCount,
                          const void *vertices,
                          int sizeOfVertex,
                          const void *colors,
                          int sizeOfColor,
                          const void *radiuses,
                          int sizeOfRadius)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
pointSphere->kind=kind;
pointSphere->vertexCount=vertexCount;
GlVertexArray_init(&pointSphere->vertexArray);
const ShaderInfo *si=mgr.pointSphereShader+pointSphere->kind;
if(pointSphere->kind&GLPOINTSPHERE_HAS_RADIUSES)
  {
  pointSphere->radiuses=GlVertexArray_createBufferGeneric(
                        &pointSphere->vertexArray,
                        si->v_Radius,1,0,
                        pointSphere->vertexCount,
                        radiuses,
                        sizeOfRadius);
  }
else
  {
  pointSphere->radiuses=-1;
  }
if(pointSphere->kind&GLPOINTSPHERE_HAS_COLORS)
  {
  pointSphere->colors=GlVertexArray_createBufferGeneric(
                      &pointSphere->vertexArray,
                      si->v_Color,3,0,
                      pointSphere->vertexCount,
                      colors,
                      sizeOfColor);
  }
else
  {
  pointSphere->colors=-1;
  }
pointSphere->vertices=GlVertexArray_createBufferGeneric(
                      &pointSphere->vertexArray,
                      si->v_Vertex,3,0,
                      pointSphere->vertexCount,
                      vertices,
                      sizeOfVertex);
}

void
GlPointSphere_destroy(GlPointSphere *pointSphere)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
for(int i=0,i_end=GlVertexArray_count(&pointSphere->vertexArray);i<i_end;++i)
  {
  int buffer=GlVertexArray_buffer(&pointSphere->vertexArray,i);
  glDeleteBuffers(1,(const unsigned int *)&buffer);
  OPENGL_CHECK;
  }
GlVertexArray_destroy(&pointSphere->vertexArray);
}

void
GlPointSphere_draw(const GlPointSphere *pointSphere,
                   const real32_t rgba[4],
                   real32_t ambient,
                   real32_t specular,
                   real32_t shininess,
                   real32_t radius,
                   const real32_t headLight[3],
                   const real32_t fog[4],
                   const real32_t *projectionMatrix,
                   const real32_t *modelViewMatrix)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
const ShaderInfo *si=mgr.pointSphereShader+
                     pointSphere->kind+(fog ? GLSHAPE_HAS_FOG : 0);
glUseProgram(GlShader_program(&si->shader));
OPENGL_CHECK;
glUniformMatrix4fv(si->u_ProjectionMatrix,1,GL_FALSE,
                   (real32_t *)projectionMatrix);
OPENGL_CHECK;
glUniformMatrix4fv(si->u_ModelViewMatrix,1,GL_FALSE,
                   (real32_t *)modelViewMatrix);
OPENGL_CHECK;
glUniform4fv(si->u_Rgba,1,(real32_t *)rgba);
OPENGL_CHECK;
glUniform4f(si->u_AmbientSpecularShininessRadius,
            ambient,specular,shininess,radius);
OPENGL_CHECK;
glUniform3fv(si->u_HeadLight,1,(real32_t *)headLight);
OPENGL_CHECK;
if(si->u_Fog!=-1)
  {
  glUniform4fv(si->u_Fog,1,(real32_t *)fog);
  OPENGL_CHECK;
  }
GlVertexArray_bind(&pointSphere->vertexArray);
if(si->u_ScreenWidth!=-1)
  {
  real32_t vp[4];
  glGetFloatv(GL_VIEWPORT,vp);
  OPENGL_CHECK;
  // FIXME: need to double or quadruplicate width on some graphics cards!?!?
  // vp[2]=*4.0f;
  glUniform1f(si->u_ScreenWidth,vp[2]);
  OPENGL_CHECK;
  glEnable(GL_PROGRAM_POINT_SIZE);
  OPENGL_CHECK;
  if(!CoreGl_coreProfile())
    { glEnable(0x8861); OPENGL_CHECK; } // GL_POINT_SPRITE
  glDrawArrays(GL_POINTS,0,(int)pointSphere->vertexCount);
  OPENGL_CHECK;
  if(!CoreGl_coreProfile())
    { glDisable(0x8861); OPENGL_CHECK; } // GL_POINT_SPRITE
  glDisable(GL_PROGRAM_POINT_SIZE);
  OPENGL_CHECK;
  }
else
  {
  glDrawArrays(GL_POINTS,0,(int)pointSphere->vertexCount);
  OPENGL_CHECK;
  }
GlVertexArray_unbind(&pointSphere->vertexArray);
glUseProgram(0);
OPENGL_CHECK;
}

//----------------------------------------------------------------------------
