//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransGl/vrRoom.h"

#include "TransProg/path.h"

#define KEYSTONE_MIPMAP 1

static
const char * VrRoom_keystoneVertexSrc=
" in vec2 v_Vertex;                                                       \n"
" uniform mat4 u_TexMatrix;                                               \n"
" out vec4 f_TexCoord;                                                    \n"
" void                                                                    \n"
" main(void)                                                              \n"
" {                                                                       \n"
" f_TexCoord=u_TexMatrix*vec4(v_Vertex,0.0,1.0);                          \n"
" gl_Position=vec4(v_Vertex*2.0-1.0,0.0,1.0);                             \n"
" }                                                                       \n";

static
const char *VrRoom_keystoneFragmentSrc=
" uniform int u_Warning;                                                  \n"
" uniform sampler2D u_TexUnit;                                            \n"
" in vec4 f_TexCoord;                                                     \n"
" void                                                                    \n"
" main(void)                                                              \n"
" {                                                                       \n"
" vec4 texColor=texture2D(u_TexUnit,f_TexCoord.xy/f_TexCoord.w);          \n"
" if(u_Warning!=0)                                                        \n"
"   {                                                                     \n"
"   gl_FragColor=vec4(texColor.rgb*0.5+vec3(0.5,0.0,0.0),1.0);            \n"
"   }                                                                     \n"
" else                                                                    \n"
"   {                                                                     \n"
"   gl_FragColor=texColor;                                                \n"
"   }                                                                     \n"
" }                                                                       \n";

static
bool // keep view
VrRoom_drawViewports(VrRoom *room,
                     GlView *view)
{
for(int i=0,i_end=room->screenCount;i<i_end;++i)
  {
  struct _VrScreen *screen=room->screens+i;
  bool permutation=room->hasRightEye&&screen->stereoPermutation;
  glBindFramebuffer(GL_FRAMEBUFFER,room->fbo);
  glViewport(0,0,room->fboWidth,room->fboHeight);
  room->currentScreen=i;
  if(permutation) room->currentRightEye=!room->currentRightEye;
  (*room->callback)(room);
  if(!GlView_data(view))
    {
    return false; // destroyed during callback
    }
  if(permutation) room->currentRightEye=!room->currentRightEye;
  glBindFramebuffer(GL_FRAMEBUFFER,0);
  if(!i)
    {
    glClearColor(room->externalRgba[0],
                 room->externalRgba[1],
                 room->externalRgba[2],
                 room->externalRgba[3]);
    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
    }
  glViewport(screen->x,screen->y,screen->w,screen->h);
  glUseProgram(GlShader_program(&room->shader));
  glUniform1i(room->u_Warning,screen->warning ? 1 : 0);
  glUniformMatrix4fv(room->u_TexMatrix,1,GL_FALSE,screen->texMatrix);
  glUniform1i(room->u_TexUnit,0); // GL_TEXTURE0
  glActiveTexture(GL_TEXTURE0);
  glBindTexture(GL_TEXTURE_2D,room->fboTexture);
#if KEYSTONE_MIPMAP
  glGenerateMipmap(GL_TEXTURE_2D);
#endif
  GlVertexArray_bind(&room->vertexArray);
  glDrawArrays(GL_TRIANGLE_FAN,4*i,4);
  }
return true;
}

static
bool // keep view
VrRoom_callback(GlView *view)
{
const GfxState *state=GlView_state(view);
// GfxState_displayEvent(state);
VrRoom *room=(VrRoom *)GlView_data(view);
if(!room)
  {
  return false; // destroyed
  }
bool useCallback=true;
switch(GfxState_event(state))
  {
  case GFX_DRAW_EVT:
    {
    if(GfxState_visible(state))
      {
      if(room->hasRightEye)
        {
        glDrawBuffer(GL_BACK_RIGHT);
        room->currentRightEye=true;
        if(!VrRoom_drawViewports(room,view))
          {
          return false; // destroyed during callback
          }
        glDrawBuffer(GL_BACK_LEFT);
        }
      room->currentRightEye=false;
      if(!VrRoom_drawViewports(room,view))
        {
        return false; // destroyed during callback
        }
      }
    useCallback=false;
    break;
    }
  default:
    {
    break;
    }
  }
if(useCallback)
  {
  (*room->callback)(room);
  }
return GlView_data(view) ? true : false; // not destroyed during callback
}

static
bool // success
VrRoom_nextLine(FILE *input,
                char *line,
                int capacity)
{
for(;;)
  {
  if(!fgets(line,capacity,input)) return false;
  const char *pos=strchr(line,'#');
  if(pos)
    {
    line[pos-line]='\0'; // cut line to ignore comment
    }
  if(line[0]&&strcmp(line,"\n")&&strcmp(line,"\r\n"))
    {
    return true; // ignore empty lines
    }
  }
}

void
VrRoom_init(VrRoom *room,
            const char *fileName,
            VrRoomCallback callback,
            void *data)
{
room->callback=callback;
room->data=data;
FILE *f=Path_fopen(fileName,"r");
if(!f)
  { SYS_DROP_DEAD("Path_fopen() failure"); }
char line[0x100];
int charCount=0;
if(!VrRoom_nextLine(f,line,0x100)||
   (sscanf(line," %d viewports %n",&room->screenCount,&charCount)!=1)||
   !charCount||
   (room->screenCount<=0))
  { DROP_DEAD("number of viewports expected"); }
room->screens=Memory_alloc(room->screenCount,struct _VrScreen);
for(int i=0,i_end=room->screenCount;i<i_end;++i)
  {
  struct _VrScreen *screen=room->screens+i;
  charCount=0;
  if(!VrRoom_nextLine(f,line,0x100)||
     (sscanf(line," viewport-x-y-w-h"
                  " %d %d %d %d %n",
                  &screen->x,&screen->y,&screen->w,&screen->h,
                  &charCount)!=4)||
     !charCount)
    { DROP_DEAD("viewport-x-y-w-h expected"); }
  charCount=0;
  if(!VrRoom_nextLine(f,line,0x100)||
     (sscanf(line," corners-bl-br-tr-tl"
                  " %lg %lg %lg %lg %lg %lg %lg %lg %n",
                  &screen->xbl,&screen->ybl,&screen->xbr,&screen->ybr,
                  &screen->xtr,&screen->ytr,&screen->xtl,&screen->ytl,
                  &charCount)!=8)||
     !charCount)
    { DROP_DEAD("corners-bl-br-tr-tl expected"); }
  charCount=0;
  if(!VrRoom_nextLine(f,line,0x100)||
     (sscanf(line," location-tx-ty-tz-yaw-pitch-roll"
                  " %lg %lg %lg %lg %lg %lg %n",
                  &screen->tx,&screen->ty,&screen->tz,
                  &screen->yaw,&screen->pitch,&screen->roll,
                  &charCount)!=6)||
     !charCount)
    { DROP_DEAD("location-tx-ty-tz-yaw-pitch-roll expected"); }
  charCount=0;
  if(!VrRoom_nextLine(f,line,0x100)||
     (sscanf(line," screen-w-h"
                  " %lg %lg %n",
                  &screen->width,&screen->height,&charCount)!=2)||
     !charCount)
    { DROP_DEAD("screen-w-h expected"); }
  int stereoPermutation=0;
  charCount=0;
  if(!VrRoom_nextLine(f,line,0x100)||
     (sscanf(line," stereo-permutation"
                  " %d %n",
                  &stereoPermutation,&charCount)!=1)||
     !charCount)
    { DROP_DEAD("stereo-permutation expected"); }
  screen->stereoPermutation=stereoPermutation ? true : false;
  }
fclose(f);
room->xOffset=room->screens[0].x;
room->yOffset=room->screens[0].y;
int xMax=room->xOffset;
int yMax=room->yOffset;
room->fboWidth=room->screens[0].w;
room->fboHeight=room->screens[0].h;
for(int i=0,i_end=room->screenCount;i<i_end;++i)
  {
  struct _VrScreen *screen=room->screens+i;
  if(screen->x<room->xOffset) room->xOffset=screen->x;
  if(screen->x+screen->w>xMax) xMax=screen->x+screen->w;
  if(screen->y<room->yOffset) room->yOffset=screen->y;
  if(screen->y+screen->h>yMax) yMax=screen->y+screen->h;
  if(screen->w>room->fboWidth) room->fboWidth=screen->w;
  if(screen->h>room->fboHeight) room->fboHeight=screen->h;
  }
int newSize=1;
while(newSize<room->fboWidth) { newSize<<=1; }
room->fboWidth=newSize;
newSize=1;
while(newSize<room->fboHeight) { newSize<<=1; }
room->fboHeight=newSize;
for(int i=0,i_end=room->screenCount;i<i_end;++i)
  {
  struct _VrScreen *screen=room->screens+i;
  screen->x-=room->xOffset;
  screen->y=yMax-(screen->y+screen->h);
  real64_t dx,dy,bottom,right,top,left,diag;
  dx=screen->xbr-screen->xbl; dy=screen->ybr-screen->ybl;
  bottom=sqrt(dx*dx+dy*dy);
  dx=screen->xtr-screen->xbr; dy=screen->ytr-screen->ybr;
  right=sqrt(dx*dx+dy*dy);
  dx=screen->xtr-screen->xtl; dy=screen->ytr-screen->ytl;
  top=sqrt(dx*dx+dy*dy);
  dx=screen->xtl-screen->xbl; dy=screen->ytl-screen->ybl;
  left=sqrt(dx*dx+dy*dy);
  dx=screen->xtr-screen->xbl; dy=screen->ytr-screen->ybl;
  diag=sqrt(dx*dx+dy*dy);
  real64_t sbr=(bottom+right+diag)*0.5;
  real64_t stl=(top+left+diag)*0.5;
  screen->usedSurface=sqrt(sbr*(sbr-bottom)*(sbr-right)*(sbr-diag))+
                      sqrt(stl*(stl-top)*(stl-left)*(stl-diag));
  screen->warning=false;
  }
room->view=GlView_open(room->xOffset,room->yOffset,
                       xMax-room->xOffset,yMax-room->yOffset,true,
                       &VrRoom_callback,room);
glGenTextures(1,(unsigned int *)&room->fboTexture);
glBindTexture(GL_TEXTURE_2D,room->fboTexture);
glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,room->fboWidth,room->fboHeight,
             0,GL_RGB,GL_UNSIGNED_BYTE,(uint8_t *)0);
#if KEYSTONE_MIPMAP
  glGenerateMipmap(GL_TEXTURE_2D);
  glTexParameteri(GL_TEXTURE_2D,
                  GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
#else
  glTexParameteri(GL_TEXTURE_2D,
                  GL_TEXTURE_MIN_FILTER,GL_LINEAR);
#endif
glTexParameteri(GL_TEXTURE_2D,
                GL_TEXTURE_MAG_FILTER,GL_LINEAR);
glTexParameteri(GL_TEXTURE_2D,
                GL_TEXTURE_WRAP_S,GL_REPEAT);
glTexParameteri(GL_TEXTURE_2D,
                GL_TEXTURE_WRAP_T,GL_REPEAT);
glBindTexture(GL_TEXTURE_2D,0);
glGenRenderbuffers(1,(unsigned int *)&room->fboDepth);
glBindRenderbuffer(GL_RENDERBUFFER,room->fboDepth);
glRenderbufferStorage(GL_RENDERBUFFER,GL_DEPTH_COMPONENT16,
                      room->fboWidth,room->fboHeight);
glBindRenderbuffer(GL_RENDERBUFFER,0);
glGenFramebuffers(1,(unsigned int *)&room->fbo);
glBindFramebuffer(GL_FRAMEBUFFER,room->fbo);
glFramebufferTexture2D(GL_FRAMEBUFFER,GL_COLOR_ATTACHMENT0,
                       GL_TEXTURE_2D,room->fboTexture,0);
glFramebufferRenderbuffer(GL_FRAMEBUFFER,GL_DEPTH_ATTACHMENT,
                          GL_RENDERBUFFER,room->fboDepth);
if(glCheckFramebufferStatus(GL_FRAMEBUFFER)!=GL_FRAMEBUFFER_COMPLETE)
  { DROP_DEAD("glCheckFramebufferStatus() failure\n"); }
glBindFramebuffer(GL_FRAMEBUFFER,0);
GlShader_init(&room->shader,VrRoom_keystoneVertexSrc,
              NULL,VrRoom_keystoneFragmentSrc,true);
int program=GlShader_program(&room->shader);
if(!program)
  {
  const char *log=GlShader_log(&room->shader);
  if(log)
    {
    Platform_msg("%s",log);
    }
  DROP_DEAD("shader failure\n");
  }
room->v_Vertex=glGetAttribLocation(program,"v_Vertex");
room->u_Warning=glGetUniformLocation(program,"u_Warning");
room->u_TexMatrix=glGetUniformLocation(program,"u_TexMatrix");
room->u_TexUnit=glGetUniformLocation(program,"u_TexUnit");
GlVertexArray_init(&room->vertexArray);
int idx=GlVertexArray_createBuffer(&room->vertexArray,
                                   room->v_Vertex,2,0,
                                   4*room->screenCount,
                                   (real32_t *)0);
room->vertexBuffer=GlVertexArray_buffer(&room->vertexArray,idx);
VrRoom_updateCorners(room);
room->externalRgba[0]=0.0f;
room->externalRgba[1]=0.0f;
room->externalRgba[2]=0.0f;
room->externalRgba[3]=1.0f;
room->currentScreen=0;
room->currentRightEye=false;
room->hasRightEye=(GlView_stereoMode()==GFX_ACTIVE_STEREO) ? true : false;
//-- change initial values of somme common settings --
// FIXME: these settings are global, not local to a specific viewer
glClearColor((real32_t)(0x60/255.0),
             (real32_t)(0x60/255.0),
             (real32_t)(0x60/255.0),1.0f);        // initial is 0,0,0,0
glEnable(GL_DEPTH_TEST);                          // initial is disabled
glEnable(GL_CULL_FACE);                           // initial is disabled
glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA); // initial is 1,0
glPixelStorei(GL_PACK_ALIGNMENT,1);               // initial is 4
glPixelStorei(GL_UNPACK_ALIGNMENT,1);             // initial is 4
}

void
VrRoom_destroy(VrRoom *room)
{
room->view->data=(void *)0; // mark as destroyed
GlView_postRedraw(room->view); // force callback invocation
glDeleteBuffers(1,(const unsigned int *)&room->vertexBuffer);
GlVertexArray_destroy(&room->vertexArray);
GlShader_destroy(&room->shader);
glDeleteFramebuffers(1,(const unsigned int *)&room->fbo);
glDeleteRenderbuffers(1,(const unsigned int *)&room->fboDepth);
glDeleteTextures(1,(const unsigned int *)&room->fboTexture);
Memory_free(room->screens);
}

void
VrRoom_applyPhysicalParams(VrRoom *room,
                           const SpatialBasis *origin,
                           const SpatialBasis *viewPoint,
                           real64_t eyeOffsetX,
                           real64_t eyeOffsetY,
                           real64_t eyeOffsetZ,
                           real64_t scale,
                           real64_t nearDistance,
                           real64_t farDistance)
{
if(!room->hasRightEye) eyeOffsetY=0.0;
nearDistance*=scale;
farDistance*=scale;
//-- determine eye position in world --
real64_t lEyeX=eyeOffsetX*scale,lEyeY=eyeOffsetY*scale,lEyeZ=eyeOffsetZ*scale;
real64_t rEyeX=lEyeX,rEyeY=-lEyeY,rEyeZ=lEyeZ;
SpatialBasis_localToGlobalPosition(viewPoint,&lEyeX,&lEyeY,&lEyeZ);
SpatialBasis_localToGlobalPosition(origin,&lEyeX,&lEyeY,&lEyeZ);
SpatialBasis_localToGlobalPosition(viewPoint,&rEyeX,&rEyeY,&rEyeZ);
SpatialBasis_localToGlobalPosition(origin,&rEyeX,&rEyeY,&rEyeZ);
for(int i=0,i_end=room->screenCount;i<i_end;++i)
  {
  struct _VrScreen *screen=room->screens+i;
  //-- set view-point to screen location --
  SpatialBasis_setMatrix(&screen->leftViewPoint,origin->matrix);
  SpatialBasis_translate(&screen->leftViewPoint,
                         screen->tx*scale,screen->ty*scale,screen->tz*scale);
  SpatialBasis_yaw(&screen->leftViewPoint,screen->yaw);
  SpatialBasis_pitch(&screen->leftViewPoint,screen->pitch);
  SpatialBasis_roll(&screen->leftViewPoint,screen->roll);
  SpatialBasis_setMatrix(&screen->rightViewPoint,
                         screen->leftViewPoint.matrix);
  //-- set view-point to eye position --
  real64_t ldx=lEyeX,ldy=lEyeY,ldz=lEyeZ;
  SpatialBasis_globalToLocalPosition(&screen->leftViewPoint,&ldx,&ldy,&ldz);
  SpatialBasis_setPosition(&screen->leftViewPoint,lEyeX,lEyeY,lEyeZ);
  Gfx_computeViewMatrix(screen->leftViewMtx,&screen->leftViewPoint);
  real64_t rdx=rEyeX,rdy=rEyeY,rdz=rEyeZ;
  SpatialBasis_globalToLocalPosition(&screen->rightViewPoint,&rdx,&rdy,&rdz);
  SpatialBasis_setPosition(&screen->rightViewPoint,rEyeX,rEyeY,rEyeZ);
  Gfx_computeViewMatrix(screen->rightViewMtx,&screen->rightViewPoint);
  //-- compute projection --
  real64_t lDiv;
  real64_t rDiv;
  if((-ldx<nearDistance)||(-rdx<nearDistance))
    {
    screen->warning=true;
    rDiv=lDiv=1.0/nearDistance;
    }
  else
    {
    screen->warning=false;
    lDiv=-1.0/ldx;
    rDiv=-1.0/rdx;
    }
  real64_t halfW=screen->width*0.5*scale;
  real64_t halfH=screen->height*0.5*scale;
  screen->leftFrustum.nearDistance=nearDistance;
  screen->leftFrustum.farDistance=farDistance;
  screen->leftFrustum.leftSlope=(+halfW-ldy)*lDiv;
  screen->leftFrustum.rightSlope=(-halfW-ldy)*lDiv;
  screen->leftFrustum.downSlope=(-halfH-ldz)*lDiv;
  screen->leftFrustum.upSlope=(+halfH-ldz)*lDiv;
  Gfx_computeFrustumMatrix(screen->leftProjMtx,&screen->leftFrustum);
  screen->rightFrustum.farDistance=farDistance;
  screen->rightFrustum.nearDistance=nearDistance;
  screen->rightFrustum.leftSlope=(+halfW-rdy)*rDiv;
  screen->rightFrustum.rightSlope=(-halfW-rdy)*rDiv;
  screen->rightFrustum.downSlope=(-halfH-rdz)*rDiv;
  screen->rightFrustum.upSlope=(+halfH-rdz)*rDiv;
  Gfx_computeFrustumMatrix(screen->rightProjMtx,&screen->rightFrustum);
  }
}

void
VrRoom_updateCorners(VrRoom *room)
{
glBindBuffer(GL_ARRAY_BUFFER,room->vertexBuffer);
for(int i=0,i_end=room->screenCount;i<i_end;++i)
  {
  struct _VrScreen *screen=room->screens+i;
  /***
  |s=[ m11 m12 0 m14 ]*|x --> use texCoord s_q=s/q, t_q=t/q
  |t [ m21 m22 0 m24 ] |y
  |0 [   0   0 1   0 ] |0
  |q [ m41 m42 0   1 ] |1
  s=s_q*q --> m11*x+m12*y+m14=s_q*(m41*x+*m42*y+1)
  t=t_q*q --> m21*x+m22*y+m24=t_q*(m41*x+*m42*y+1)
  x=xbl,y=ybl s_q=0,t_q=0 --> m11*xbl+m12*ybl+m14=0                 (eq.0)
                              m21*xbl+m22*ybl+m24=0                 (eq.1)
  x=xbr,y=ybr s_q=1,t_q=0 --> m11*xbr+m12*ybr+m14=m41*xbr+m42*ybr+1 (eq.2)
                              m21*xbr+m22*ybr+m24=0                 (eq.3)
  x=xtr,y=ytr s_q=1,t_q=1 --> m11*xtr+m12*ytr+m14=m41*xtr+m42*ytr+1 (eq.4)
                              m21*xtr+m22*ytr+m24=m41*xtr+m42*ytr+1 (eq.5)
  x=xtl,y=ytl s_q=0,t_q=1 --> m11*xtl+m12*ytl+m14=0                 (eq.6)
                              m21*xtl+m22*ytl+m24=m41*xtl+m42*ytl+1 (eq.7)
  ***/
  real64_t equationSystem[]={ +screen->xbl, +screen->ybl, 1.0,
                              0.0,          0.0,          0.0,
                              0.0,          0.0,          0.0,   // (eq.0)
                              0.0,          0.0,          0.0,
                              +screen->xbl, +screen->ybl, 1.0,
                              0.0,          0.0,          0.0,   // (eq.1)
                              +screen->xbr, +screen->ybr, 1.0,
                              0.0,          0.0,          0.0,
                              -screen->xbr, -screen->ybr, 1.0,   // (eq.2)
                              0.0,          0.0,          0.0,
                              +screen->xbr, +screen->ybr, 1.0,
                              0.0,          0.0,          0.0,   // (eq.3)
                              +screen->xtr, +screen->ytr, 1.0,
                              0.0,          0.0,          0.0,
                              -screen->xtr, -screen->ytr, 1.0,   // (eq.4)
                              0.0,          0.0,          0.0,
                              +screen->xtr, +screen->ytr, 1.0,
                              -screen->xtr, -screen->ytr, 1.0,   // (eq.5)
                              +screen->xtl, +screen->ytl, 1.0,
                              0.0,          0.0,          0.0,
                              0.0,          0.0,          0.0,   // (eq.6)
                              0.0,          0.0,          0.0,
                              +screen->xtl, +screen->ytl, 1.0,
                              -screen->xtl, -screen->ytl, 1.0 }; // (eq.7)
                  // m11  m12  m14  m21  m22  m24  m41  m42
  real64_t vars[8]={ 1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0 };
  if(!gaussianElimination64(8,equationSystem,vars))
    {
    Platform_msg("singular matrix for equation system\n");
    }
  real32_t *texMatrix=screen->texMatrix;
  texMatrix[0]=(real32_t)vars[0];  // m11
  texMatrix[1]=(real32_t)vars[3];  // m21
  texMatrix[2]=0.0f;
  texMatrix[3]=(real32_t)vars[6];  // m41
  texMatrix[4]=(real32_t)vars[1];  // m12
  texMatrix[5]=(real32_t)vars[4];  // m22
  texMatrix[6]=0.0f;
  texMatrix[7]=(real32_t)vars[7];  // m42
  texMatrix[8]=0.0f;
  texMatrix[9]=0.0f;
  texMatrix[10]=1.0f;
  texMatrix[11]=0.0f;
  texMatrix[12]=(real32_t)vars[2]; // m14
  texMatrix[13]=(real32_t)vars[5]; // m24
  texMatrix[14]=0.0f;
  texMatrix[15]=1.0f;
  real32_t corners[4*2]={ (real32_t)screen->xbl, (real32_t)screen->ybl,
                          (real32_t)screen->xbr, (real32_t)screen->ybr,
                          (real32_t)screen->xtr, (real32_t)screen->ytr,
                          (real32_t)screen->xtl, (real32_t)screen->ytl };
  glBufferSubData(GL_ARRAY_BUFFER,i*sizeof(corners),sizeof(corners),corners);
  }
glBindBuffer(GL_ARRAY_BUFFER,0);
}

//----------------------------------------------------------------------------
