//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/gfxUtils.h"

#include "TransProg/path.h"

void
GfxState_displayEvent(const GfxState *state)
{
char buffer[0x100];
sprintf(buffer,"%dx%d %s",
        GfxState_width(state),
        GfxState_height(state),
        GfxState_visible(state) ? "visible" : "unvisible");
GfxModifier mod=GfxState_modifiers(state);
if(mod)
  {
  strcat(buffer," mod=[");
  if(mod&GFX_CAPSLOCK_MDF) strcat(buffer," CAPSLOCK");
  if(mod&GFX_SHIFT_MDF)    strcat(buffer," SHIFT");
  if(mod&GFX_CONTROL_MDF)  strcat(buffer," CONTROL");
  if(mod&GFX_NUMLOCK_MDF)  strcat(buffer," NUMLOCK");
  if(mod&GFX_LEFT_MDF)     strcat(buffer," LEFT");
  if(mod&GFX_MIDDLE_MDF)   strcat(buffer," MIDDLE");
  if(mod&GFX_RIGHT_MDF)    strcat(buffer," RIGHT");
  strcat(buffer," ]");
  }
switch(GfxState_event(state))
  {
  case GFX_NO_EVT:
    {
    Platform_msg("%s NO_EVT\n",
                 buffer);
    break;
    }
  case GFX_BUTTON_EVT:
    {
    const char *btn;
    switch(GfxState_button(state))
      {
      case GFX_LEFT_BTN:   btn="LEFT";   break;
      case GFX_MIDDLE_BTN: btn="MIDDLE"; break;
      case GFX_RIGHT_BTN:  btn="RIGHT";  break;
      case GFX_UP_BTN:     btn="UP";     break;
      case GFX_DOWN_BTN:   btn="DOWN";   break;
      default:             btn="???";    break;
      }
    Platform_msg("%s BUTTON_EVT [%s] %s (%d,%d)\n",
                 buffer,
                 btn,
                 GfxState_pressed(state) ? "pressed" : "released",
                 GfxState_mouseX(state),
                 GfxState_mouseY(state));
    break;
    }
  case GFX_KEY_EVT:
    {
    char key[0x10]="";
    strcpy(key,GfxState_key(state));
    if((strlen(key)==1)&&!isprint((int)key[0]))
      {
      sprintf(key,"0x%.2x",GfxState_key(state)[0]);
      }
    Platform_msg("%s KEY_EVT [%s] %s (%d,%d)\n",
                 buffer,
                 key,
                 GfxState_pressed(state) ? "pressed" : "released",
                 GfxState_mouseX(state),
                 GfxState_mouseY(state));
    break;
    }
  case GFX_MOTION_EVT:
    {
    Platform_msg("%s MOTION_EVT (%d,%d)\n",
                 buffer,
                 GfxState_mouseX(state),
                 GfxState_mouseY(state));
    break;
    }
  case GFX_CONFIG_EVT:
    {
    Platform_msg("%s CONFIG_EVT\n",
                 buffer);
    break;
    }
  case GFX_CLOSE_EVT:
    {
    Platform_msg("%s CLOSE_EVT\n",
                 buffer);
    break;
    }
  case GFX_DRAW_EVT:
    {
    Platform_msg("%s DRAW_EVT\n",
                 buffer);
    break;
    }
  }
}

void
Gfx_computeFrustumMatrix(real32_t *out_frustumMatrix,
                         const Frustum *frustum)
{
real64_t RpL=-(frustum->rightSlope+frustum->leftSlope); // +y --> -x
real64_t _RmL=-1.0/(frustum->rightSlope-frustum->leftSlope); // +y --> -x
real64_t UpD=frustum->upSlope+frustum->downSlope;
real64_t _UmD=1.0/(frustum->upSlope-frustum->downSlope);
real64_t NpF=frustum->nearDistance+frustum->farDistance;
real64_t _NmF=1.0/(frustum->nearDistance-frustum->farDistance);
real64_t NtF=frustum->nearDistance*frustum->farDistance;
// standard glFrustum()
out_frustumMatrix[MTX_R00]=(real32_t)(2.0*_RmL);
out_frustumMatrix[MTX_R01]=0.0f;
out_frustumMatrix[MTX_R02]=(real32_t)(RpL*_RmL);
out_frustumMatrix[MTX_R10]=0.0f;
out_frustumMatrix[MTX_R11]=(real32_t)(2.0*_UmD);
out_frustumMatrix[MTX_R12]=(real32_t)(UpD*_UmD);
out_frustumMatrix[MTX_R20]=0.0f;
out_frustumMatrix[MTX_R21]=0.0f;
out_frustumMatrix[MTX_R22]=(real32_t)(NpF*_NmF);
out_frustumMatrix[MTX_TX]=0.0f;
out_frustumMatrix[MTX_TY]=0.0f;
out_frustumMatrix[MTX_TZ]=(real32_t)(2.0*NtF*_NmF);
out_frustumMatrix[MTX_Z0]=0.0f;
out_frustumMatrix[MTX_Z1]=0.0f;
out_frustumMatrix[MTX_Z2]=-1.0f;
out_frustumMatrix[MTX_ONE]=0.0f;
}

void
Gfx_computeProjectionMatrix(real32_t *out_projectionMatrix,
                            const Perspective *perspective)
{
Gfx_computeFrustumMatrix(out_projectionMatrix,
                         Perspective_frustum(perspective));
}

void
Gfx_computeStereoProjectionMatrices(real32_t *out_leftProjectionMatrix,
                                    real32_t *out_rightProjectionMatrix,
                                    const Perspective *perspective)
{
Frustum frustum=*Perspective_frustum(perspective);
real64_t dy=0.5*Perspective_stereoEyeSeparation(perspective);
if(frustum.nearDistance>=Perspective_stereoViewDistance(perspective))
  {
  dy/=frustum.nearDistance;
  }
else
  {
  dy/=Perspective_stereoViewDistance(perspective);
  }
frustum.leftSlope-=dy;
frustum.rightSlope-=dy;
Gfx_computeFrustumMatrix(out_leftProjectionMatrix,&frustum);
frustum.leftSlope+=2.0*dy;
frustum.rightSlope+=2.0*dy;
Gfx_computeFrustumMatrix(out_rightProjectionMatrix,&frustum);
}

static
void
Gfx_applyViewMatrix(real32_t *out_viewMatrix,
                    const real64_t *mtx,
                    real64_t camX,
                    real64_t camY,
                    real64_t camZ)
{
// inverse of view matrix left-multiplied by
//    0 -1  0  0        -y --> x
//    0  0  1  0         z --> y
//    1  0  0  0         x --> z
//    0  0  0  1
out_viewMatrix[MTX_R00]=(real32_t)-mtx[MTX_R01]; // -y --> x
out_viewMatrix[MTX_R01]=(real32_t)-mtx[MTX_R11];
out_viewMatrix[MTX_R02]=(real32_t)-mtx[MTX_R21];
out_viewMatrix[MTX_R10]=(real32_t)+mtx[MTX_R02]; // z --> y
out_viewMatrix[MTX_R11]=(real32_t)+mtx[MTX_R12];
out_viewMatrix[MTX_R12]=(real32_t)+mtx[MTX_R22];
out_viewMatrix[MTX_R20]=(real32_t)-mtx[MTX_R00]; // -x --> z
out_viewMatrix[MTX_R21]=(real32_t)-mtx[MTX_R10];
out_viewMatrix[MTX_R22]=(real32_t)-mtx[MTX_R20];
out_viewMatrix[MTX_TX]=(real32_t)(-camX*out_viewMatrix[MTX_R00]
                                  -camY*out_viewMatrix[MTX_R01]
                                  -camZ*out_viewMatrix[MTX_R02]);
out_viewMatrix[MTX_TY]=(real32_t)(-camX*out_viewMatrix[MTX_R10]
                                  -camY*out_viewMatrix[MTX_R11]
                                  -camZ*out_viewMatrix[MTX_R12]);
out_viewMatrix[MTX_TZ]=(real32_t)(-camX*out_viewMatrix[MTX_R20]
                                  -camY*out_viewMatrix[MTX_R21]
                                  -camZ*out_viewMatrix[MTX_R22]);
out_viewMatrix[MTX_Z0]=
out_viewMatrix[MTX_Z1]=
out_viewMatrix[MTX_Z2]=0.0f;
out_viewMatrix[MTX_ONE]=1.0f;
}

void
Gfx_computeViewMatrix(real32_t *out_viewMatrix,
                      const SpatialBasis *viewPoint)
{
const real64_t *mtx=viewPoint->matrix;
Gfx_applyViewMatrix(out_viewMatrix,mtx,
                    mtx[MTX_TX],mtx[MTX_TY],mtx[MTX_TZ]);
}

void
Gfx_computeStereoViewMatrices(real32_t *out_leftViewMatrix,
                              real32_t *out_rightViewMatrix,
                              const SpatialBasis *viewPoint,
                              const Perspective *perspective)
{
const real64_t *mtx=viewPoint->matrix;
real64_t dy=Perspective_stereoEyeSeparation(perspective)*0.5;
Gfx_applyViewMatrix(out_leftViewMatrix,mtx,
                    mtx[MTX_TX]+dy*mtx[MTX_R01],
                    mtx[MTX_TY]+dy*mtx[MTX_R11],
                    mtx[MTX_TZ]+dy*mtx[MTX_R21]);
Gfx_applyViewMatrix(out_rightViewMatrix,mtx,
                    mtx[MTX_TX]-dy*mtx[MTX_R01],
                    mtx[MTX_TY]-dy*mtx[MTX_R11],
                    mtx[MTX_TZ]-dy*mtx[MTX_R21]);
}

void
Gfx_computeModelViewMatrix(real32_t *out_modelViewMatrix,
                           const real32_t *viewMatrix,
                           const real64_t *modelMatrix)
{
out_modelViewMatrix[MTX_R00]=(real32_t)
                             (viewMatrix[MTX_R00]*modelMatrix[MTX_R00]+
                              viewMatrix[MTX_R01]*modelMatrix[MTX_R10]+
                              viewMatrix[MTX_R02]*modelMatrix[MTX_R20]);
out_modelViewMatrix[MTX_R01]=(real32_t)
                             (viewMatrix[MTX_R00]*modelMatrix[MTX_R01]+
                              viewMatrix[MTX_R01]*modelMatrix[MTX_R11]+
                              viewMatrix[MTX_R02]*modelMatrix[MTX_R21]);
out_modelViewMatrix[MTX_R02]=(real32_t)
                             (viewMatrix[MTX_R00]*modelMatrix[MTX_R02]+
                              viewMatrix[MTX_R01]*modelMatrix[MTX_R12]+
                              viewMatrix[MTX_R02]*modelMatrix[MTX_R22]);
out_modelViewMatrix[MTX_TX]=(real32_t)
                            (viewMatrix[MTX_R00]*modelMatrix[MTX_TX]+
                             viewMatrix[MTX_R01]*modelMatrix[MTX_TY]+
                             viewMatrix[MTX_R02]*modelMatrix[MTX_TZ]+
                             viewMatrix[MTX_TX]);
out_modelViewMatrix[MTX_R10]=(real32_t)
                             (viewMatrix[MTX_R10]*modelMatrix[MTX_R00]+
                              viewMatrix[MTX_R11]*modelMatrix[MTX_R10]+
                              viewMatrix[MTX_R12]*modelMatrix[MTX_R20]);
out_modelViewMatrix[MTX_R11]=(real32_t)
                             (viewMatrix[MTX_R10]*modelMatrix[MTX_R01]+
                              viewMatrix[MTX_R11]*modelMatrix[MTX_R11]+
                              viewMatrix[MTX_R12]*modelMatrix[MTX_R21]);
out_modelViewMatrix[MTX_R12]=(real32_t)
                             (viewMatrix[MTX_R10]*modelMatrix[MTX_R02]+
                              viewMatrix[MTX_R11]*modelMatrix[MTX_R12]+
                              viewMatrix[MTX_R12]*modelMatrix[MTX_R22]);
out_modelViewMatrix[MTX_TY]=(real32_t)
                            (viewMatrix[MTX_R10]*modelMatrix[MTX_TX]+
                             viewMatrix[MTX_R11]*modelMatrix[MTX_TY]+
                             viewMatrix[MTX_R12]*modelMatrix[MTX_TZ]+
                             viewMatrix[MTX_TY]);
out_modelViewMatrix[MTX_R20]=(real32_t)
                             (viewMatrix[MTX_R20]*modelMatrix[MTX_R00]+
                              viewMatrix[MTX_R21]*modelMatrix[MTX_R10]+
                              viewMatrix[MTX_R22]*modelMatrix[MTX_R20]);
out_modelViewMatrix[MTX_R21]=(real32_t)
                             (viewMatrix[MTX_R20]*modelMatrix[MTX_R01]+
                              viewMatrix[MTX_R21]*modelMatrix[MTX_R11]+
                              viewMatrix[MTX_R22]*modelMatrix[MTX_R21]);
out_modelViewMatrix[MTX_R22]=(real32_t)
                             (viewMatrix[MTX_R20]*modelMatrix[MTX_R02]+
                              viewMatrix[MTX_R21]*modelMatrix[MTX_R12]+
                              viewMatrix[MTX_R22]*modelMatrix[MTX_R22]);
out_modelViewMatrix[MTX_TZ]=(real32_t)
                            (viewMatrix[MTX_R20]*modelMatrix[MTX_TX]+
                             viewMatrix[MTX_R21]*modelMatrix[MTX_TY]+
                             viewMatrix[MTX_R22]*modelMatrix[MTX_TZ]+
                             viewMatrix[MTX_TZ]);
out_modelViewMatrix[MTX_Z0]=
out_modelViewMatrix[MTX_Z1]=
out_modelViewMatrix[MTX_Z2]=0.0f;
out_modelViewMatrix[MTX_ONE]=1.0f;
}

bool // success
Gfx_saveTgaImage(const char *fileName,
                 int width,
                 int height,
                 int components,
                 const uint8_t *rgbStorage)
{
if((components!=1)&&(components!=3)&&(components!=4))
  {
  return false;
  }
FILE *file=Path_fopen(fileName,"wb");
if(!file)
  {
  return false;
  }
uint8_t header[18];
memset(header,0,18);
header[2]=(uint8_t)(components==1 ? 3 : 2); // uncompressed grey or RGB
header[12]=(uint8_t)(width&0x000000FF);
header[13]=(uint8_t)((width>>8)&0x000000FF);
header[14]=(uint8_t)(height&0x000000FF);
header[15]=(uint8_t)((height>>8)&0x000000FF);
header[16]=(uint8_t)(components<<3);
if(fwrite(header,sizeof(uint8_t),18,file)!=18)
  {
  fclose(file);
  return false;
  }
for(int y=0;y<height;++y)
  {
  for(int x=0;x<width;++x)
    {
    int i=components*(y*width+x);
    if(components>=3)
      {
      if(fwrite(rgbStorage+i+2,sizeof(uint8_t),1,file)!=1) // blue
        {
        fclose(file);
        return false;
        }
      if(fwrite(rgbStorage+i+1,sizeof(uint8_t),1,file)!=1) // green
        {
        fclose(file);
        return false;
        }
      }
    if(fwrite(rgbStorage+i,sizeof(uint8_t),1,file)!=1) // red or grey
      {
      fclose(file);
      return false;
      }
    if((components==4)&&
       (fwrite(rgbStorage+i+3,sizeof(uint8_t),1,file)!=1)) // alpha
      {
      fclose(file);
      return false;
      }
    }
  }
fclose(file);
return true;
}

uint8_t * // allocated image storage (bottom to top) or 0
Gfx_loadTgaImage(const char *fileName,
                 int *out_width,
                 int *out_height,
                 int *out_components)
{
uint8_t header[18];
FILE *file=Path_fopen(fileName,"rb");
if(!file)
  {
  return (uint8_t *)0;
  }
if(fread(header,sizeof(uint8_t),18,file)!=18)
  {
  fclose(file);
  return (uint8_t *)0;
  }
int idLength=header[0];
int colormapType=header[1];
int dataTypeCode=header[2];
int colormapLength=header[6]*256+header[5];
*out_width=header[13]*256+header[12];
*out_height=header[15]*256+header[14];
int bitsPerPixel=header[16];
bool flipV=(header[17]&(uint8_t)0x20) ? true : false;
bool flipH=(header[17]&(uint8_t)0x10) ? true : false;
if((((dataTypeCode!=3)||(bitsPerPixel!=8))&&
    ((dataTypeCode!=2)||((bitsPerPixel!=24)&&(bitsPerPixel!=32))))||
   ((colormapType!=0)&&(colormapType!=1))||
   (!*out_width)||(!*out_height))
  {
  fclose(file);
  return (uint8_t *)0;
  }
if(fseek(file,idLength+colormapType*colormapLength,SEEK_CUR)==-1)
  {
  fclose(file);
  return (uint8_t *)0;
  }
int i;
*out_components=bitsPerPixel>>3;
int lineSize=(*out_width)*(*out_components);
int dataSize=(*out_height)*lineSize;
uint8_t *data=Memory_alloc(dataSize,uint8_t);
if(flipV)
  {
  if(flipH)
    {
    for(uint8_t *dest=data+dataSize-*out_components;
        dest>=data;dest-=*out_components)
      {
      if(fread(dest,sizeof(uint8_t),*out_components,file)!=
         (size_t)*out_components)
        {
        fclose(file);
        Memory_free(data);
        return (uint8_t *)0;
        }
      }
    }
  else
    {
    for(uint8_t *line=data+dataSize-lineSize;line>=data;line-=lineSize)
      {
      if(fread(line,sizeof(uint8_t),lineSize,file)!=(size_t)lineSize)
        {
        fclose(file);
        Memory_free(data);
        return (uint8_t *)0;
        }
      }
    }
  }
else
  {
  if(flipH)
    {
    for(i=0;i<*out_height;++i)
      {
      uint8_t *line=data+i*lineSize;
      for(uint8_t *row=line+lineSize-*out_components;
          row>=line;row-=*out_components)
        {
        if(fread(row,sizeof(uint8_t),*out_components,file)!=
           (size_t)*out_components)
          {
          fclose(file);
          Memory_free(data);
          return (uint8_t *)0;
          }
        }
      }
    }
  else
    {
    if(fread(data,sizeof(uint8_t),dataSize,file)!=(size_t)dataSize)
      {
      fclose(file);
      Memory_free(data);
      return (uint8_t *)0;
      }
    }
  }
fclose(file);
if(*out_components>=3)
  {
  for(i=0;i<dataSize;i+=*out_components)
    {
    uint8_t tmp=data[i]; // swap red and blue
    data[i]=data[i+2];
    data[i+2]=tmp;
    }
  }
return data;
}

//----------------------------------------------------------------------------
