//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/memory.h"
#include "TransProg/synchro.h"

// #define DISABLE_CUSTOM_ALLOC 1
#define ALIGNED_DISABLED_CUSTOM_ALLOC 1

struct _MemoryBlock
{
CACHE_ALIGNED
intptr_t size;
struct _MemoryBlock *nextFree;
int pageIndex;
int cpuCtx;
bool free;
};

struct _MemoryPage
{
uint8_t *addr;
intptr_t size;
struct _MemoryBlock *firstFree;
};

struct _MemoryAllocator
{
CACHE_ALIGNED
SpinSynchro sync;
struct _MemoryPage *pageTable;
int pageTableCapacity;
int pageTableCount;
};

struct _MemoryCache
{
struct _MemoryBlock *entries[1];
};

#define BLKSZ (sizeof(struct _MemoryBlock))

static struct _MemoryAllocator _Memory_allocators[ASSUMED_MAX_NUMA_COUNT];

static uint8_t *_Memory_cacheAddr;
static intptr_t _Memory_cacheLimit;
static int _Memory_cacheEntries;

TRANSPROG_INLINE
struct _MemoryCache *
_Memory_cache(int cpuCtx)
{
return (struct _MemoryCache *)(_Memory_cacheAddr+
                               Platform_standardPageSize()*cpuCtx);
}

TRANSPROG_INLINE
struct _MemoryBlock *
_Memory_ptrToBlock(void *ptr)
{
return (struct _MemoryBlock *)(((uint8_t *)ptr)-BLKSZ);
}

TRANSPROG_INLINE
void *
_Memory_blockToPtr(struct _MemoryBlock *block)
{
return ((uint8_t *)block)+BLKSZ;
}

TRANSPROG_INLINE
struct _MemoryBlock *
_Memory_nextBlock(struct _MemoryBlock *block)
{
return (struct _MemoryBlock *)(((uint8_t *)block)+block->size);
}

TRANSPROG_INLINE
bool
_Memory_validBlock(struct _MemoryPage *page,
                   struct _MemoryBlock *block)
{
return (((uintptr_t)block)+BLKSZ)<=((uintptr_t)page->addr)+page->size;
}

TRANSPROG_INLINE
void
_Memory_agregateFree(struct _MemoryPage *page,
                     struct _MemoryBlock *block)
{
for(struct _MemoryBlock *next=_Memory_nextBlock(block);
    _Memory_validBlock(page,next)&&next->free;
    next=_Memory_nextBlock(block))
  {
  block->size+=next->size;
  block->nextFree=next->nextFree;
  }
}

TRANSPROG_INLINE
void
_Memory_freeListAdd(struct _MemoryPage *page,
                    struct _MemoryBlock *newFree)
{
struct _MemoryBlock **nextFree=&page->firstFree;
while(*nextFree<newFree)
  {
  nextFree=&((*nextFree)->nextFree);
  }
newFree->nextFree=*nextFree;
*nextFree=newFree;
newFree->free=true;
}

TRANSPROG_INLINE
void
_Memory_freeListRemove(struct _MemoryPage *page,
                       struct _MemoryBlock *oldFree)
{
struct _MemoryBlock **nextFree=&page->firstFree;
while(*nextFree!=oldFree)
  {
  nextFree=&((*nextFree)->nextFree);
  }
*nextFree=oldFree->nextFree;
oldFree->free=false;
}

TRANSPROG_INLINE
struct _MemoryBlock *
_Memory_allocFromPage(int cpuCtx,
                      struct _MemoryPage *page,
                      intptr_t size)
{
for(struct _MemoryBlock *block=page->firstFree;
    _Memory_validBlock(page,block);
    block=block->nextFree)
  {
  _Memory_agregateFree(page,block);
  intptr_t remaining=block->size-size;
  if(remaining>=0)
    {
    _Memory_freeListRemove(page,block);
    if(remaining)
      {
      block->size=size;
      struct _MemoryBlock *next=_Memory_nextBlock(block);
      next->size=remaining;
      next->pageIndex=block->pageIndex;
      // next->cpuCtx left uninitialised
      _Memory_freeListAdd(page,next);
      }
    block->cpuCtx=cpuCtx;
    return block;
    }
  }
return (struct _MemoryBlock *)0;
}

TRANSPROG_INLINE
struct _MemoryBlock *
_Memory_allocBlock(int cpuCtx,
                   struct _MemoryAllocator *alloc,
                   intptr_t size)
{
struct _MemoryPage *pageTable=alloc->pageTable;
for(int i=0,i_end=alloc->pageTableCount;i<i_end;++i)
  {
  struct _MemoryBlock *block=_Memory_allocFromPage(cpuCtx,pageTable+i,size);
  if(block)
    {
    return block;
    }
  }
intptr_t newPageSize=0;
uint8_t *newPageAddr=(uint8_t *)0;
bool preferHuge=true;
intptr_t unitSize=Platform_pageSize(preferHuge);
int64_t memSize=Platform_memorySize();
if(memSize)
  {
  if(Platform_maxNumaId()>=0)
    {
    memSize/=Platform_maxNumaId()+1; // share between numa nodes
    }
  memSize/=128; // avoid consuming too much memory at startup
  }
else
  {
  memSize=8LL*1024LL*1024LL; // arbitrary amount
  }
intptr_t minPageSize=Platform_alignSize(size,unitSize);
if(memSize<minPageSize)
  {
  memSize=minPageSize;
  }
for(newPageSize=Platform_alignSize((intptr_t)memSize,unitSize);
    newPageSize>=minPageSize;
    newPageSize-=unitSize)
  {
  newPageAddr=(uint8_t *)Memory_pageAlloc(&newPageSize,preferHuge);
  if(newPageAddr)
    {
    break;
    }
  }
if(!newPageAddr)
  {
  DROP_DEAD("cannot allocate memory page");
  }
intptr_t required=(alloc->pageTableCount+1)*sizeof(struct _MemoryPage);
if(required>alloc->pageTableCapacity)
  {
  struct _MemoryPage *newTable=(struct _MemoryPage *)
                               Memory_pageAlloc(&required,false);
  if(!newTable)
    {
    DROP_DEAD("cannot allocate page table");
    }
  if(alloc->pageTable)
    {
    memcpy(newTable,alloc->pageTable,
           alloc->pageTableCount*sizeof(struct _MemoryPage));
    Memory_pageFree(alloc->pageTable,alloc->pageTableCapacity);
    }
  alloc->pageTable=newTable;
  alloc->pageTableCapacity=(int)required;
  }
struct _MemoryBlock *block=(struct _MemoryBlock *)newPageAddr;
block->size=newPageSize-(newPageSize%BLKSZ);
block->nextFree=_Memory_nextBlock(block);
block->pageIndex=alloc->pageTableCount;
// block->cpuCtx left uninitialised
block->free=true;
struct _MemoryPage *page=alloc->pageTable+alloc->pageTableCount++;
page->addr=newPageAddr;
page->size=newPageSize;
page->firstFree=block;
return _Memory_allocFromPage(cpuCtx,page,size);
}

TRANSPROG_INLINE
struct _MemoryBlock *
_Memory_tryReAllocBlock(int cpuCtx,
                        struct _MemoryAllocator *alloc,
                        struct _MemoryBlock *oldBlock,
                        intptr_t newSize)
{
struct _MemoryPage *page=alloc->pageTable+oldBlock->pageIndex;
intptr_t remaining=oldBlock->size-newSize;
if(remaining>=0)
  {
  if(remaining)
    {
    oldBlock->size=newSize;
    struct _MemoryBlock *next=_Memory_nextBlock(oldBlock);
    next->size=remaining;
    next->pageIndex=oldBlock->pageIndex;
    // next->cpuCtx left uninitialised
    _Memory_freeListAdd(page,next);
    }
  oldBlock->cpuCtx=cpuCtx;
  return oldBlock;
  }
struct _MemoryBlock *next=_Memory_nextBlock(oldBlock);
if(_Memory_validBlock(page,next)&&next->free)
  {
  _Memory_agregateFree(page,next);
  intptr_t remaining=oldBlock->size+next->size-newSize;
  if(remaining>=0)
    {
    oldBlock->size=newSize;
    _Memory_freeListRemove(page,next);
    if(remaining)
      {
      next=_Memory_nextBlock(oldBlock);
      next->size=remaining;
      next->pageIndex=oldBlock->pageIndex;
      // next->cpuCtx left uninitialised
      _Memory_freeListAdd(page,next);
      }
    oldBlock->cpuCtx=cpuCtx;
    return oldBlock;
    }
  }
return (struct _MemoryBlock *)0;
}

void * // allocated memory
Memory_allocVoid(intptr_t size)
{
#if defined DISABLE_CUSTOM_ALLOC
# if defined ALIGNED_DISABLED_CUSTOM_ALLOC
    size+=Platform_maxCacheLineSize()+1;
    uint8_t *ptr=(uint8_t *)malloc(size);
    uint8_t *p=(uint8_t *)Platform_alignAddress(ptr+1,
                                                Platform_maxCacheLineSize());
    p[-1]=(uint8_t)(p-ptr);
    return p;
# else
    return malloc(size);
# endif
#else
  int cpuCtx=Platform_cpuCtx();
  struct _MemoryBlock *block=(struct _MemoryBlock *)0;
  size=Platform_alignSize(size+BLKSZ,BLKSZ);
  if(size<=_Memory_cacheLimit)
    {
    intptr_t index=(size-1)/BLKSZ;
    struct _MemoryCache *cache=_Memory_cache(cpuCtx);
    block=cache->entries[index];
    if(block)
      {
      cache->entries[index]=block->nextFree;
      }
    }
  if(!block)
    {
    int numaCtx=Platform_memNumaCtx(cpuCtx);
    struct _MemoryAllocator *alloc=_Memory_allocators+numaCtx;
    SpinSynchro_lockW(&alloc->sync);
    block=_Memory_allocBlock(cpuCtx,alloc,size);
    SpinSynchro_unlockW(&alloc->sync);
    }
  return _Memory_blockToPtr(block);
#endif
}

void
Memory_free(void *ptr)
{
if(ptr)
  {
#if defined DISABLE_CUSTOM_ALLOC
# if defined ALIGNED_DISABLED_CUSTOM_ALLOC
    uint8_t *p=(uint8_t *)ptr;
    free(p-p[-1]);
# else
    free(ptr);
# endif
#else
  struct _MemoryBlock *oldBlock=_Memory_ptrToBlock(ptr);
  int cpuCtx=Platform_cpuCtx();
  int oldCpuCtx=oldBlock->cpuCtx;
  intptr_t oldSize=oldBlock->size;
  if((oldSize<=_Memory_cacheLimit)&&(oldCpuCtx==cpuCtx))
    {
    intptr_t index=(oldSize-1)/BLKSZ;
    struct _MemoryCache *cache=_Memory_cache(cpuCtx);
    oldBlock->nextFree=cache->entries[index];
    cache->entries[index]=oldBlock;
    }
  else
    {
    int oldNumaCtx=Platform_memNumaCtx(oldCpuCtx);
    struct _MemoryAllocator *oldAlloc=_Memory_allocators+oldNumaCtx;
    SpinSynchro_lockW(&oldAlloc->sync);
    _Memory_freeListAdd(oldAlloc->pageTable+oldBlock->pageIndex,oldBlock);
    SpinSynchro_unlockW(&oldAlloc->sync);
    }
#endif
  }
}

void * // reallocated memory
Memory_reAllocVoid(void *oldPtr,
                   intptr_t newSize)
{
#if defined DISABLE_CUSTOM_ALLOC
# if defined ALIGNED_DISABLED_CUSTOM_ALLOC
    int oldOffset=0;
    if(oldPtr)
      {
      uint8_t *p=(uint8_t *)oldPtr;
      oldOffset=p[-1];
      oldPtr=p-oldOffset;
      }
    if(newSize)
      {
      newSize+=Platform_maxCacheLineSize()+1;
      }
    uint8_t *newPtr=(uint8_t *)realloc(oldPtr,newSize);
    if(newPtr)
      {
      uint8_t *p=(uint8_t *)Platform_alignAddress(newPtr+1,
                                                  Platform_maxCacheLineSize());
      int newOffset=p-newPtr;
      if(newOffset!=oldOffset)
        {
	memmove(newPtr+newOffset,newPtr+oldOffset,
                newSize-(Platform_maxCacheLineSize()+1));
        }
      p[-1]=(uint8_t)newOffset;
      newPtr=p;
      }
    return newPtr;
# else
    return realloc(oldPtr,newSize);
# endif
#else
  if(newSize&&oldPtr)
    {
    struct _MemoryBlock *oldBlock=_Memory_ptrToBlock(oldPtr);
    int cpuCtx=Platform_cpuCtx();
    int numaCtx=Platform_memNumaCtx(cpuCtx);
    int oldCpuCtx=oldBlock->cpuCtx;
    int oldNumaCtx=Platform_memNumaCtx(oldCpuCtx);
    if(oldNumaCtx==numaCtx)
      {
      struct _MemoryAllocator *alloc=_Memory_allocators+numaCtx;
      intptr_t size=Platform_alignSize(newSize+BLKSZ,BLKSZ);
      SpinSynchro_lockW(&alloc->sync);
      struct _MemoryBlock *newBlock=_Memory_tryReAllocBlock(cpuCtx,alloc,
                                                            oldBlock,size);
      SpinSynchro_unlockW(&alloc->sync);
      if(newBlock)
        {
        return _Memory_blockToPtr(newBlock);
        }
      }
    intptr_t oldSize=oldBlock->size-BLKSZ;
    void *newPtr=Memory_allocVoid(newSize);
    memcpy(newPtr,oldPtr,oldSize<newSize ? oldSize : newSize);
    Memory_free(oldPtr);
    return newPtr;
    }
  else if(newSize)
    {
    return Memory_allocVoid(newSize);
    }
  else if(oldPtr)
    {
    Memory_free(oldPtr);
    }
  return (void *)0;
#endif
}

void
Memory_numaUsage(const void *ptr,
                 intptr_t *numaUsage)
{
#if defined DISABLE_CUSTOM_ALLOC
  (void)ptr;
  (void)numaUsage;
#else
  const struct _MemoryBlock *block=_Memory_ptrToBlock((void *)ptr);
  numaUsage[Platform_memNumaCtx(block->cpuCtx)]+=block->size;
#endif
}

void * // allocated pages or 0
Memory_pageAlloc(intptr_t *inout_size,
                 bool preferHuge)
{
for(int huge=preferHuge ? 1 : 0;huge>=0;--huge)
  {
  intptr_t pageSize=huge ? Platform_hugePageSize()
                         : Platform_standardPageSize();
  if(!pageSize)
    {
    continue;
    }
  intptr_t size=Platform_alignSize(*inout_size,pageSize);
#if defined _WIN32
  DWORD flags=MEM_COMMIT|MEM_RESERVE;
# if defined MEM_LARGE_PAGES
  if(huge)
    {
    flags|=MEM_LARGE_PAGES;
    }
# endif
  void *pages=VirtualAlloc((void *)0,(SIZE_T)size,flags,PAGE_READWRITE);
#else
  int flags=MAP_PRIVATE|MAP_ANONYMOUS;
# if defined MAP_HUGETLB
  if(huge)
    {
    flags|=MAP_HUGETLB;
    }
# endif
  void *pages=mmap((void *)0,size,PROT_READ|PROT_WRITE,flags,-1,0);
  if(pages==MAP_FAILED)
    {
    pages=(void *)0;
    }
#endif
  if(pages) // nb: try first-touch strategy to bind pages to current numa node
    {
    uint8_t *p=(uint8_t *)pages;
    for(intptr_t i=0;i<size;i+=pageSize)
      {
      p[i]=0;
      }
    *inout_size=size;
    return pages;
    }
  }
return (void *)0;
}

void
Memory_pageFree(void *pages,
                intptr_t size)
{
#if defined _WIN32
  (void)size;
  if(!VirtualFree(pages,0,MEM_RELEASE))
    {
    SYS_DROP_DEAD("VirtualFree() failure");
    }
#else
  if(munmap(pages,size)==-1)
    {
    SYS_DROP_DEAD("munmap() failure");
    }
#endif
}

void
_Memory_init(void)
{
for(int numa=0,numa_end=Platform_numaCtxCount();numa<numa_end;++numa)
  {
  struct _MemoryAllocator *alloc=_Memory_allocators+numa;
  SpinSynchro_init(&alloc->sync);
  alloc->pageTable=(struct _MemoryPage *)0;
  alloc->pageTableCapacity=0;
  alloc->pageTableCount=0;
  }
int cpuCount=Platform_cpuCtxCount();
intptr_t size=Platform_standardPageSize()*cpuCount;
_Memory_cacheAddr=(uint8_t *)Memory_pageAlloc(&size,false);
if(!_Memory_cacheAddr)
  {
  DROP_DEAD("cannot allocate cache");
  }
_Memory_cacheEntries=(int)(Platform_standardPageSize()/
                           sizeof(struct _MemoryBlock *));
_Memory_cacheLimit=_Memory_cacheEntries*BLKSZ;
for(int cpu=0;cpu<cpuCount;++cpu)
  {
  struct _MemoryCache *cache=_Memory_cache(cpu);
  for(int i=0;i<_Memory_cacheEntries;++i)
    {
    cache->entries[i]=(struct _MemoryBlock *)0;
    }
  }
}

//----------------------------------------------------------------------------
